/*
   *Copyright:  

   ACEMod (2018--2021), The University of Sydney's invention disclosure CDIP Ref. 2019-123.
   Copyright ©  Oliver Cliff, Cameron Zachreson, Nathan Harding, Kristopher Fair, Mikhail Prokopenko, Mahendarajah Piraveenan

   AMTraC-19 (2020-2021), The University of Sydney's invention disclosure CDIP 2020-018 (2020).
   Copyright ©  Sheryl Chang, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   AMTraC-19 (2022-2023), The University of Sydney.
   Copyright ©  Sheryl Chang, Quang Dang Nguyen, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   AMTraC-19 (2024), The University of Sydney.
   Copyright ©  Quang Dang Nguyen, Sheryl Chang, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   *EndCopyright:
*/

#include "agent.h"
#include "debug.h"
#include "enums.h"

#include <iostream>
#include <cstdlib>
#include <math.h>
#include <algorithm>

#include <omp.h>

using namespace std;

#define P_CHILD    0
#define P_ADULT    1

extern float WDpro[3][3];

constexpr float Agent::DAYS_PER_STEP_;

float Agent::computeViralLoad(int t)
{
   float viral_titer;

   if (t < this->getLATENT_PERIOD_())
    {
       viral_titer = 0;
    }

   else if (t < this->getPEAK_VIRAL_TITER_())
    {
       float x = float( t - this->getLATENT_PERIOD_()) / float( this->getPEAK_VIRAL_TITER_() - this->getLATENT_PERIOD_());
       viral_titer = (pow(Agent::exponent_, x) - 1) / (Agent::exponent_ - 1);
    }

   else
    {
       viral_titer = float( this->getINFECTION_PERIOD_() - t ) / float( this->getINFECTION_PERIOD_() - this->getPEAK_VIRAL_TITER_());
    }

   return(viral_titer);
}


void Agent::printTiterVsTime(void)

{
   std::vector<float> viral_load_;

   for (unsigned t = 0; t < this->getLATENT_PERIOD_(); ++t)
    {
       viral_load_.push_back(0);
    }

   float viral_titer;
   for (unsigned t = this->getLATENT_PERIOD_(); t <= this->getINFECTION_PERIOD_(); ++t)
    {
       if (t < this->getPEAK_VIRAL_TITER_())
        {
           float x = float( t - this->getLATENT_PERIOD_()) / float( this->getPEAK_VIRAL_TITER_() - this->getLATENT_PERIOD_());
           viral_titer = (pow(Agent::exponent_, x) - 1) / (Agent::exponent_ - 1);
        }
       else
        {
           viral_titer = float( this->getINFECTION_PERIOD_() - t ) / float( this->getINFECTION_PERIOD_() - this->getPEAK_VIRAL_TITER_());
        }
       viral_load_.push_back(viral_titer);
    }


   for (unsigned i = 0; i < viral_load_.size(); ++i)
    {
       std::cout << "titer[" << i << "]: " << viral_load_.at(i) << std::endl;
    }
}


void Agent::init(void)
{

   float symp_rval = rand_r(&seed_thread[omp_get_thread_num()]) / (float)RAND_MAX;
   float symptomatic_fraction = PERCENT_SYMPTOMATIC;

   if (!this->isAdult()) {
      symptomatic_fraction = CHILD_PERCENT_SYMPTOMATIC;
   }

   if (symp_rval < symptomatic_fraction) {
      incubating_period_ = this->getPEAK_VIRAL_TITER_();
   } else {
      incubating_period_ = -1;
   }

   state_                   = DiseaseStates::StateSusceptible;
   infected_prior_          = false;
   new_illness_             = false;
   newly_detected_          = false; // new label to differentiate detections from illnesses
   illness_detected_        = false; // resets illness detection flag, this resets CI status too ( or at least it should )
   detected_recovered_      = false;

   SDstate_                    = SD_States::SD_noncompliant;
   CIstate_                    = CI_States::CI_noncompliant;
   HQstate_                    = HQ_States::HQ_noncompliant;
   SCstate_                    = SC_States::SC_nil;
   Compliance_state_           = Compliance_States::nil;
   Isolation_state_            = Isolation_States::nil;
   isolation_factor_home_      = 1;
   isolation_factor_work_      = 1;
   isolation_factor_community_ = 1;
   HQclock_                    = 0;

   steps_since_infection_ = 0;
   steps_since_recovery_ = 0;

   rational_opinion_on_compliance = 0;

   // clear all immunisation items save in the history.
   this->clear_immunisation_history();
}

// Getter and setter functions for Opinion Dynamics
float Agent::get_rational_opinion_on_compliance() {
   return rational_opinion_on_compliance;
}

void Agent::set_rational_opinion_on_compliance(float new_rational_opinion_value) {
   rational_opinion_on_compliance = new_rational_opinion_value;
}

void Agent::step(unsigned day) {
   if ((this->getState() > DiseaseStates::StateSusceptible) && (this->getState() < DiseaseStates::StateRemoved)) {
      this->steps_since_infection_++;

      // If this agent isn't showing symptoms but should at some point (i.e., inc period > 0), check if should start showing symptoms
      if ((this->getState() == DiseaseStates::StateAsymptomatic) &&
         (this->incubating_period_ > 0) &&
         (this->steps_since_infection_ > incubating_period_))
      {
         this->setState(DiseaseStates::StateSymptomatic);
         this->new_illness_ = true;
      }

      // if agent is detected, make sure households are updated accordingly, as their status is re-zeroed each round
      if (this->isIllnessDetected()) {
         if (!this->getHH()->getSymptomaticInfection()) {
            #pragma omp critical(step_HH_addSymptomaticAgent)
            {
               this->getHH()->addSymptomaticAgent();
            }
         }
      }

      unsigned Recovery_steps = this->getINFECTION_PERIOD_();
      if (this->steps_since_infection_ > Recovery_steps) {
         this->setState(DiseaseStates::StateRemoved);

         steps_since_recovery_ = 0;
         
         if (this->illness_detected_) {
            this->detected_recovered_ = true; // precludes vaccination
         }

         this->illness_detected_ = false; // CZ COV19 - making sure detected individuals are counted correctly

         #pragma omp critical(step_CD_uninfect_CD_)
         {
            this->getCD()->remove_n_infecteds_tot();
            if (this->getCD()->get_n_infecteds_tot() == 0) {
               this->getCD()->uninfect_CD();
            }
         }

      }
   } else if (this->getState() == DiseaseStates::StateRemoved) {
      steps_since_recovery_++;

      if (steps_since_recovery_ > this->getAFTER_RECOVERY_PERIOD_()) {

         // Reset and prepare this agent to join Susceptible population again.
         this->setState(DiseaseStates::StateSusceptible);
         this->infected_prior_         = false;
         this->new_illness_            = false;
         this->newly_detected_         = false;
         this->illness_detected_       = false;
         this->detected_recovered_     = false;
         this->steps_since_infection_  = 0;
         this->steps_since_recovery_   = 0;

         // Add an immunisation to this agent's immunisation history after this agent is recovered from an infection
         this->add_an_immunisation_to_history(
            day,
            this->VEc_natural_after_recovered, this->VEt_natural_after_recovered, 
            this->daily_waning_VEc_natural_after_recovered, this->daily_waning_VEt_natural_after_recovered, 
            Immunisation_Types::recovered
         );
      }
   }

}


void Agent::step_HQ(int duration_HQ)
{
   if (this->getIsolationState() == Isolation_States::HQ)
    {
       this->HQclock_ += 1;
    }

   if (this->HQclock_ > duration_HQ)
    {
       this->setIsolationState(Isolation_States::nil);
       this->HQclock_ = 0;
    }
}


void Agent::eval_Compliance_state(void)
{
   if ((this->getCIState() == CI_States::CI_compliant) &&
       (this->getHQState() == HQ_State::HQ_compliant) &&
       (this->getSDState() == SD_State::SD_compliant))
    {
       this->setComplianceState(Compliance_States::SD_CI_HQ);
    }

   else if ((this->getCIState() == CI_States::CI_compliant) &&
            (this->getHQState() == HQ_State::HQ_compliant) &&
            (this->getSDState() == SD_State::SD_noncompliant))
    {
       this->setComplianceState(Compliance_States::CI_HQ);
    }

   else if ((this->getCIState() == CI_States::CI_compliant) &&
            (this->getHQState() == HQ_State::HQ_noncompliant) &&
            (this->getSDState() == SD_State::SD_compliant))
    {
       this->setComplianceState(Compliance_States::SD_CI);
    }

   else if ((this->getCIState() == CI_States::CI_compliant) &&
            (this->getHQState() == HQ_State::HQ_noncompliant) &&
            (this->getSDState() == SD_State::SD_noncompliant))
    {
       this->setComplianceState(Compliance_States::CI);
    }

   else if ((this->getCIState() == CI_States::CI_noncompliant) &&
            (this->getHQState() == HQ_State::HQ_noncompliant) &&
            (this->getSDState() == SD_State::SD_compliant))
    {
       this->setComplianceState(Compliance_States::SD);
    }

   else if ((this->getCIState() == CI_States::CI_noncompliant) &&
            (this->getHQState() == HQ_State::HQ_noncompliant) &&
            (this->getSDState() == SD_State::SD_noncompliant))
    {
       this->setComplianceState(Compliance_States::nil);
    }
}


void Agent::eval_Compliance_state_Qmod(void)
{
   if ((this->getCIState() == CI_States::CI_compliant) &&
       (this->getHQState() == HQ_State::HQ_compliant) &&
       (this->getSDState() == SD_State::SD_compliant))
   {
       this->setComplianceState(Compliance_States::SD_CI_HQ);
   } else if ((this->getCIState() == CI_States::CI_compliant) &&
            (this->getHQState() == HQ_State::HQ_compliant) &&
            (this->getSDState() == SD_State::SD_noncompliant)) {
       this->setComplianceState(Compliance_States::CI_HQ);
   } else if ((this->getCIState() == CI_States::CI_compliant) &&
         (this->getHQState() == HQ_State::HQ_noncompliant) &&
         (this->getSDState() == SD_State::SD_compliant)) {
      this->setComplianceState(Compliance_States::SD_CI);
   } else if ((this->getCIState() == CI_States::CI_compliant) &&
         (this->getHQState() == HQ_State::HQ_noncompliant) &&
         (this->getSDState() == SD_State::SD_noncompliant)) {
      this->setComplianceState(Compliance_States::CI);
   } else if ((this->getCIState() == CI_States::CI_noncompliant) &&
         (this->getHQState() == HQ_State::HQ_noncompliant) &&
         (this->getSDState() == SD_State::SD_compliant)) {
      this->setComplianceState(Compliance_States::SD);
   } else { // simplify the last if branch to cover all other cases. As the initialised state is nil, this brach and the rest are all set to nil.
      this->setComplianceState(Compliance_States::nil);
   }
}

int Agent::findParent_ID(agent_map_type& agents)
{
   int parent_ID = -2;
   int agent_ID;

   if (!this->isStudent())
    {
       std::cout << "WARNING: trying to find parent of non-student" << std::endl;
       return(-1);
    }
   else
    {
       std::vector<unsigned long> IDs_HH = this->getHH()->getAgentIdxs();

       std::random_shuffle(IDs_HH.begin(), IDs_HH.end());

       for (int i = 0; i < IDs_HH.size(); i++)
        {
           agent_ID = this->getHH()->getAgentIdxs().at(i);

           if (agents.at(agent_ID)->isAdult())
            {
               parent_ID = agent_ID;

               if ((agents.at(parent_ID)->getSCState() == SC_States::SC_parent) ||
                   (agents.at(parent_ID)->getSCState() == SC_States::SC_teacher))
                {
                   return(parent_ID); // if a parent is already staying home, just choose them
                }
            }
        }
    }

   if (parent_ID < 0)
    {
       std::cout << "WARNING: unable to find parent in student's household" << std::endl;
    }
   return(parent_ID);
}


void Agent::setIsolationFactor_home(float FoI_SD_home, float FoI_CI_home, float FoI_HQ_home, float FoI_SC_home)
{
   if (this->getIsolationState() == Isolation_States::CI)
    {
       this->isolation_factor_home_ = FoI_CI_home;
    }
   else if (this->getIsolationState() == Isolation_States::HQ)
    {
       this->isolation_factor_home_ = FoI_HQ_home;
    }
   else if (this->getIsolationState() == Isolation_States::SD)
    {
       this->isolation_factor_home_ = FoI_SD_home;
    }
   else if (this->getIsolationState() == Isolation_States::SC)
    {
       this->isolation_factor_home_ = FoI_SC_home;
    }
   else
    {
       this->isolation_factor_home_ = 1;
    }
}


void Agent::setIsolationFactor_work(float FoI_SD_work, float FoI_CI_work, float FoI_HQ_work, float FoI_SC_work)
{
   if (this->getIsolationState() == Isolation_States::CI)
    {
       this->isolation_factor_work_ = FoI_CI_work;
    }
   else if (this->getIsolationState() == Isolation_States::HQ)
    {
       this->isolation_factor_work_ = FoI_HQ_work;
    }
   else if (this->getIsolationState() == Isolation_States::SD)
    {
       this->isolation_factor_work_ = FoI_SD_work;
    }
   else if (this->getIsolationState() == Isolation_States::SC)
    {
       this->isolation_factor_work_ = FoI_SC_work;
    }
   else
    {
       this->isolation_factor_work_ = 1;
    }
}


void Agent::setIsolationFactor_community(float FoI_SD_community, float FoI_CI_community, float FoI_HQ_community, float FoI_SC_community)
{
   if (this->getIsolationState() == Isolation_States::CI)
    {
       this->isolation_factor_community_ = FoI_CI_community;
    }
   else if (this->getIsolationState() == Isolation_States::HQ)
    {
       this->isolation_factor_community_ = FoI_HQ_community;
    }
   else if (this->getIsolationState() == Isolation_States::SD)
    {
       this->isolation_factor_community_ = FoI_SD_community;
    }
   else if (this->getIsolationState() == Isolation_States::SC)
    {
       this->isolation_factor_community_ = FoI_SC_community;
    }
   else
    {
       this->isolation_factor_community_ = 1;
    }
}


// This function checks if the symptomatic agents are detected.
bool Agent::Illness_detection(float& Detection)
{
   bool  agent_detected = false;
   // If the sick agent is detected then add them to the symptomatic area
   float detectP = rand_r(&seed_thread[omp_get_thread_num()]) / (float)RAND_MAX;

   if (detectP < Detection) {
      agent_detected = true;
      this->illness_detected_ = true;

      this->newly_detected_ = true;

      // Adding symptomatic agents to the detection pools
      #pragma omp critical(Illness_detection_) 
      {
         this->getHH()->addSymptomaticAgent();
         this->getCD()->addSymptomaticAgent();
         
         for (auto c_wg : working_groups_) {
            c_wg->addSymptomaticAgent();
         }
      }

   }

   return(agent_detected);
}

void Agent::addToNighttimeProbabilityOfInfection(unsigned day)
{
   float factor = this->computeFactor(day);
   #pragma omp critical(addToNighttimeProbabilityOfInfection_)
   {
      this->addToHouseholdProbability(factor);
      this->addToClusterProbability(factor);
      this->addToCommunityProbability(factor);
      this->addToNeighbourhoodProbability(factor);
   }
}


void Agent::addToNighttimeProbabilityOfInfection_SD(unsigned day)
{
   float factor = this->computeFactor(day);
   
   #pragma omp critical(addToNighttimeProbabilityOfInfection_SD_)
   {
      this->addToHouseholdProbability(factor * this->getIsolationFactor_home());
      this->addToClusterProbability(factor * this->getIsolationFactor_community());
      this->addToCommunityProbability(factor * this->getIsolationFactor_community());
      this->addToNeighbourhoodProbability(factor * this->getIsolationFactor_community());
   }

}


float Agent::getNighttimeProbabilityOfInfection(void)
{
   /*
    * Get the joint probability of infection, given where the agent interacts.
    *
    * Household: 5 element vector specifying Agegroup (5 groups)
    * Household Cluster: 2 element vector specifying child or adult (2 groups)
    * Census District: 5 element vector of age groups (5 groups)
    * Statistical local area: 5 element vector of age groups (5 groups)
    */
   return(1.0 - this->getHH()->getProbabilityOfTransmission()
          * this->getHC()->getProbabilityOfTransmission(this->isAdult())
          * this->getCD()->getProbabilityOfTransmission()
          * this->getSLA()->getProbabilityOfTransmission());
}

// Add potential sources of infection to the vector pointers of potential_source_IDs and potential_source_weights
void Agent::getNighttimePotentialSourcesOfInfection(std::vector<unsigned>& potential_source_IDs, std::vector<float>& potential_source_weights) {
   this->getHH()->getPotentialSourcesOfInfection(potential_source_IDs, potential_source_weights);
   this->getHC()->getPotentialSourcesOfInfection(this->isAdult(), potential_source_IDs, potential_source_weights);
   this->getCD()->getPotentialSourcesOfInfection(potential_source_IDs, potential_source_weights);
   this->getSLA()->getPotentialSourcesOfInfection(potential_source_IDs, potential_source_weights);
}

float Agent::getHouseholdProbabilityOfInfection(void)
{
   return(1.0 - this->getHH()->getProbabilityOfTransmission());
}


float Agent::getHouseholdClusterProbabilityOfInfection(void)
{
   return(1.0 - this->getHC()->getProbabilityOfTransmission(this->isAdult()));
}


float Agent::getNeighbourhoodProbabilityOfInfection(void)
{
   float P_not_trans = this->getCD()->getProbabilityOfTransmission();
   return(1.0 - P_not_trans);
}


float Agent::getCommunityProbabilityOfInfection(void)
{
   float P_not_trans = this->getSLA()->getProbabilityOfTransmission();
   return(1.0 - P_not_trans);
}

void Agent::addToDaytimeProbabilityOfInfection(unsigned day) {
   float factor = this->computeFactor(day);
   
   #pragma omp critical(addToDaytimeProbabilityOfInfection_)
   {
      for (auto c_wg : working_groups_) {
         c_wg->addInfectedAgent(this->get_trans_scaler(), factor, this->getID());
      }
   }
}

void Agent::addToDaytimeProbabilityOfInfection_SD(unsigned day) {
   float factor = this->computeFactor(day);
   
   #pragma omp critical(addToDaytimeProbabilityOfInfection_SD_)
   {
      for (auto c_wg : working_groups_) {
         c_wg->addInfectedAgent(this->get_trans_scaler(), factor * this->getIsolationFactor_work(), this->getID());
      }
   }
   
}

float Agent::getDaytimeProbabilityOfInfection(void) {
   float P_not_infection_prod = 1.0;
   for (auto c_wg : working_groups_) {
      P_not_infection_prod *= c_wg->getProbabilityOfTransmission();
   }
   return(1.0 - P_not_infection_prod);
}

float Agent::getDaytimeProbabilityOfInfection(unsigned type) {
   if (working_groups_.size() < type) {
      std::cerr << "Working group vector length: " << working_groups_.size() << " (requested " << type << ")" << std::endl;
   }
   return(1.0 - working_groups_.at(type)->getProbabilityOfTransmission());
}

// Add potential sources of infection to the vector pointers of potential_source_IDs and potential_source_weights
void Agent::getDaytimePotentialSourcesOfInfection(std::vector<unsigned>& potential_source_IDs, std::vector<float>& potential_source_weights) {
   for (auto c_wg : working_groups_) {
      c_wg->getPotentialSourcesOfInfection(potential_source_IDs, potential_source_weights);
   }
}

void Agent::clear_immunisation_history() {
   this->immunisation_history.clear();
}

float Agent::get_VEc_from_immunisation_history(unsigned day) {
   float squared_VEc_total = 0;
   for (unsigned immunisation_idx=0; immunisation_idx < immunisation_history.size(); immunisation_idx++) {
      float VEc_i = immunisation_history[immunisation_idx]->get_VEc() * (1.0 - (day - immunisation_history[immunisation_idx]->get_date()) * immunisation_history[immunisation_idx]->get_daily_waning_VEc());
      squared_VEc_total += (float) pow(VEc_i,2);
   }
   float VEc_total = std::min(sqrt(squared_VEc_total), (float) 1.0);
   return VEc_total;
}

float Agent::get_VEs_from_immunisation_history(unsigned day) {
   float squared_VEc_total = 0;
   for (unsigned immunisation_idx=0; immunisation_idx < immunisation_history.size(); immunisation_idx++) {
      float VEc_i = immunisation_history[immunisation_idx]->get_VEc() * (1.0 - (day - immunisation_history[immunisation_idx]->get_date()) * immunisation_history[immunisation_idx]->get_daily_waning_VEc());
      squared_VEc_total += (float) pow(VEc_i,2);
   }
   float VEc_total = std::min(sqrt(squared_VEc_total), (float) 1.0);
   return 1 - sqrt(1 - VEc_total);
}

float Agent::get_VEd_from_immunisation_history(unsigned day) {
   float squared_VEc_total = 0;
   for (unsigned immunisation_idx=0; immunisation_idx < immunisation_history.size(); immunisation_idx++) {
      float VEc_i = immunisation_history[immunisation_idx]->get_VEc() * (1.0 - (day - immunisation_history[immunisation_idx]->get_date()) * immunisation_history[immunisation_idx]->get_daily_waning_VEc());
      squared_VEc_total += (float) pow(VEc_i,2);
   }
   float VEc_total = std::min(sqrt(squared_VEc_total), (float) 1.0);
   return 1 - sqrt(1 - VEc_total);
}

float Agent::get_VEt_from_immunisation_history(unsigned day) {
   float squared_VEt_total = 0;
   for (unsigned immunisation_idx=0; immunisation_idx < immunisation_history.size(); immunisation_idx++) {
      float VEt_i = immunisation_history[immunisation_idx]->get_VEt() * (1.0 - (day - immunisation_history[immunisation_idx]->get_date()) * immunisation_history[immunisation_idx]->get_daily_waning_VEt());
      squared_VEt_total += (float) pow(VEt_i,2) ;
      
   }
   float VEt_total = std::min(sqrt(squared_VEt_total), (float) 1.0);
   return VEt_total;
}

void Agent::add_an_immunisation_to_history(unsigned day, float VEc, float VEt, float daily_waning_VEc, float daily_waning_VEt, Immunisation_Type immunisation) {
   unsigned next_idx = this->immunisation_history.size();
   this->immunisation_history.emplace(next_idx, std::make_shared <Immunity>(day, VEc, VEt, daily_waning_VEc, daily_waning_VEt, (Immunisation_Type) immunisation));
}

// summarise the agent's immunisation history. It is an easy setup for further extension if more detailed summary is required.
Immunisation_Summary_Type Agent::get_immunisation_summary_type() {
   unsigned vaccination_times = 0;
   unsigned recovering_times  = 0;
   for (unsigned immunisation_idx=0; immunisation_idx < immunisation_history.size(); immunisation_idx++) {
      if ((immunisation_history[immunisation_idx]->get_immunisation_type() == Immunisation_Types::vaccinated_PV) || (immunisation_history[immunisation_idx]->get_immunisation_type() == Immunisation_Types::vaccinated_GV)) {
         vaccination_times++;
      } else if (immunisation_history[immunisation_idx]->get_immunisation_type() == Immunisation_Types::recovered) {
         recovering_times++;
      }
   }

   if (vaccination_times == 0) {
      if (recovering_times == 0) {
         return Immunisation_Summary_Type::no_immunisation;
      } else if (recovering_times == 1) {
         return Immunisation_Summary_Type::recovered_once;
      } else if (recovering_times > 1) {
         return Immunisation_Summary_Type::recovered_twice_or_more;
      } else {
         std::cout << "ERROR: Unexpected error in summarising Immmunisation History of an agent!";
         exit(1);
      }
   } else if (vaccination_times == 1) {
      if (recovering_times == 0) {
         return Immunisation_Summary_Type::vaccinated_once;
      } else if (recovering_times == 1) {
         return Immunisation_Summary_Type::vaccinated_once_and_recovered_once;
      } else if (recovering_times > 1) {
         return Immunisation_Summary_Type::vaccinated_once_and_recovered_twice_or_more;
      } else {
         std::cout << "ERROR: Unexpected error in summarising Immmunisation History of an agent!";
         exit(1);
      }
   } else if (vaccination_times > 1) {
      if (recovering_times == 0) {
         return Immunisation_Summary_Type::vaccinated_twice_or_more;
      } else if (recovering_times == 1) {
         return Immunisation_Summary_Type::vaccinated_twice_or_more_and_recovered_once;
      } else if (recovering_times > 1) {
         return Immunisation_Summary_Type::vaccinated_twice_or_more_and_recovered_twice_or_more;
      } else {
         std::cout << "ERROR: Unexpected error in summarising Immmunisation History of an agent!";
         exit(1);
      }
   } else {
      std::cout << "ERROR: Unexpected error in summarising Immmunisation History of an agent!";
      exit(1);
   }
}


void Agent::infect(unsigned day, float VEc_natural_after_recovered, float VEt_natural_after_recovered, float daily_waning_VEc_NR, float daily_waning_VEt_NR)
{  
   // agent is now infected, then the set up for natural immunity after this infection will be prepared in advanced 
   // and this immunisation will be added to the immunisation history after this agent is recovered or deceased.
   this->VEc_natural_after_recovered = VEc_natural_after_recovered;
   this->VEt_natural_after_recovered = VEt_natural_after_recovered;
   this->daily_waning_VEc_natural_after_recovered = daily_waning_VEc_NR;
   this->daily_waning_VEt_natural_after_recovered = daily_waning_VEt_NR;

   // VEd is determined based on agents' immunisation history.
   float VEd_from_immunisation_history = this->get_VEd_from_immunisation_history(day);

   #pragma omp critical
   {
      this->getCD()->add_n_infecteds_tot();
      // if agent's CD is uninfected, increase newly infected CD counter
      if (this->getCD()->get_n_infecteds_tot() == 1) {
         this->getCD()->infect_CD(); // CD has at least one newly infected agent
      }
   }

   // This function is used to infect an individal;
   this->setState(DiseaseStates::StateAsymptomatic);
   this->steps_since_infection_ = 0;
   this->infected_prior_        = true;

   float symp_rval = rand_r(&seed_thread[omp_get_thread_num()]) / (float)RAND_MAX;
   float symptomatic_fraction = this->getPERCENT_SYMPTOMATIC();

   if (!this->isAdult()) {
      symptomatic_fraction = this->getCHILD_PERCENT_SYMPTOMATIC();
   }

   symptomatic_fraction *= (1 - VEd_from_immunisation_history);

   if (symp_rval < symptomatic_fraction) {
      float inc_rval = rand_r(&seed_thread[omp_get_thread_num()]) / (float)RAND_MAX;
      incubating_period_ = this->getPEAK_VIRAL_TITER_();
   } else {
      incubating_period_ = -1;
   }
}


float Agent::computeFactor(unsigned day)
{
   // VEt is determined based on the agent's immunisation history
   float VEt_from_immunisation_history = this->get_VEt_from_immunisation_history(day);
   
   if ((this->getState() > DiseaseStates::StateSusceptible) && (this->getState() < DiseaseStates::StateRemoved)) {
      float factor = computeViralLoad(this->steps_since_infection_ - 1);

      if (this->getState() == DiseaseStates::StateAsymptomatic && this->incubating_period_ < 0) {
         factor *= Agent::ASYMPTOMATIC_INFECTIVITY_;
      }

      factor *= (1 - VEt_from_immunisation_history);

      return(factor);
   }
   return(0.0);
}


// Identifying SA1 regions that have symptomatic infection in them
bool Agent::CD_Sympotmatic_present(void)
{
   bool Sympotmatic_infection = false;

   if (this->getCD()->getSymptomaticInfection())
    {
       Sympotmatic_infection = true;
    }
   return(Sympotmatic_infection);
}


// Initiate viral infection info. This setup characterises the virus.
void Agent::initiate_infection_info(Virus& viral_info) {
   this->infection_info.set_trans_scaler_dist(viral_info.get_trans_scaler_mean(), viral_info.get_trans_scaler_std());
   this->infection_info.set_latent_period_dist(viral_info.get_latent_period_mean(), viral_info.get_latent_period_std());
   this->infection_info.set_incubation_period_dist(viral_info.get_incubation_period_dist_mu(), viral_info.get_incubation_period_dist_sig());
   this->infection_info.set_recovery_period_dist(viral_info.get_recovery_period_dist_high(), viral_info.get_recovery_period_dist_low());
   this->infection_info.set_after_recovery_period_dist(viral_info.get_after_recovery_period_mean(), viral_info.get_after_recovery_period_std());
   this->infection_info.set_adult_symptomatic_fraction(viral_info.get_adult_symptomatic_fraction());
   this->infection_info.set_child_symptomatic_fraction(viral_info.get_child_symptomatic_fraction());
   this->infection_info.set_asymptomatic_infectivity(viral_info.get_asymptomatic_infectivity());
   this->infection_info.set_origin_code(viral_info.get_origin_code());
   this->set_and_sample_all_disease_params();
}

// Initiate viral infection info. This setup characterises the virus. This is the function for co-infection where there are two sources of infection. 
// TODO: NEED FURTHER IMPLEMENTATION IF WANT TO ENABLE CO-INFECTION.
void Agent::initiate_infection_info(Virus& viral_info_1, Virus& viral_info_2) {
   // Need to be changed to address two sources of infection
   this->infection_info.set_trans_scaler_dist(viral_info_1.get_trans_scaler_mean(), viral_info_1.get_trans_scaler_std());
   this->infection_info.set_latent_period_dist(viral_info_1.get_latent_period_mean(), viral_info_1.get_latent_period_std());
   this->infection_info.set_incubation_period_dist(viral_info_1.get_incubation_period_dist_mu(), viral_info_1.get_incubation_period_dist_sig());
   this->infection_info.set_recovery_period_dist(viral_info_1.get_recovery_period_dist_high(), viral_info_1.get_recovery_period_dist_low());
   this->infection_info.set_after_recovery_period_dist(viral_info_1.get_after_recovery_period_mean(), viral_info_1.get_after_recovery_period_std());
   this->infection_info.set_adult_symptomatic_fraction(viral_info_1.get_adult_symptomatic_fraction());
   this->infection_info.set_child_symptomatic_fraction(viral_info_1.get_child_symptomatic_fraction());
   this->infection_info.set_asymptomatic_infectivity(viral_info_1.get_asymptomatic_infectivity());
   this->infection_info.set_origin_code(viral_info_1.get_origin_code());
   this->set_and_sample_all_disease_params();
}

// Initiate natural disease history of the virus for this particular agent. It is specific for this agent. 
// Values of virus parameters are sampled from a distribution to show the unique interaction between the virus and this agent
void Agent::set_and_sample_all_disease_params(void) {
   float sampled_trans_scaler = this->infection_info.sample_trans_scaler();
   float sampled_latent_period = this->infection_info.sample_latent_period();
   float sampled_incubation_period = this->infection_info.sample_incubation_period();
   int   sampled_recovery_period = this->infection_info.sample_recovery_period();
   float sampled_after_recovery_period = this->infection_info.sample_after_recovery_period();

   this->setTrans_scaler(sampled_trans_scaler);
   this->setLATENT_PERIOD_((unsigned) (sampled_latent_period * 2));
   this->setPEAK_VIRAL_TITER_((unsigned) (sampled_incubation_period * 2 + sampled_latent_period * 2));
   this->setINFECTION_PERIOD_((unsigned) (sampled_incubation_period * 2 + sampled_recovery_period * 2 + sampled_latent_period * 2));
   this->setAFTER_RECOVERY_PERIOD_((unsigned) sampled_after_recovery_period * 2);
   this->setPERCENT_SYMPTOMATIC(this->infection_info.get_adult_symptomatic_fraction());
   this->setCHILD_PERCENT_SYMPTOMATIC(this->infection_info.get_child_symptomatic_fraction());
   this->setASYMPTOMATIC_INFECTIVITY_(this->infection_info.get_asymptomatic_infectivity());
}


