/*
   *Copyright:  

   ACEMod (2018--2021), The University of Sydney's invention disclosure CDIP Ref. 2019-123.
   Copyright ©  Oliver Cliff, Cameron Zachreson, Nathan Harding, Kristopher Fair, Mikhail Prokopenko, Mahendarajah Piraveenan

   AMTraC-19 (2020-2021), The University of Sydney's invention disclosure CDIP 2020-018 (2020).
   Copyright ©  Sheryl Chang, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   AMTraC-19 (2022-2023), The University of Sydney.
   Copyright ©  Sheryl Chang, Quang Dang Nguyen, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   AMTraC-19 (2024), The University of Sydney.
   Copyright ©  Quang Dang Nguyen, Sheryl Chang, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   *EndCopyright:
*/

#ifndef __AGENT_H__
#define __AGENT_H__

#include "enums.h"
#include "statistical_areas.h"
#include "debug.h"
#include "virus.h"
#include "immunity.h"

#include <iostream>
#include <vector>
#include <map>
#include <memory>

#include <omp.h>

/*
 * Person data structure
 */

class Agent
{
private:
   std::vector<unsigned> seed_thread;

   // Infection information for the infected virus
   float trans_scaler = 0;
   unsigned LATENT_PERIOD_ = 4;              // non-infectious period (in half-days) between exposure and incubation
   unsigned PEAK_VIRAL_TITER_ = 10;          // time from exposure to peak infectiousness (includes latent period)
   unsigned INFECTION_PERIOD_ = 34;          // total time from exposure to recovery
   unsigned AFTER_RECOVERY_PERIOD_ = 56;     // time for being able to be infected again after a recovery
   float ASYMPTOMATIC_INFECTIVITY_ = 0.3;    // proportion of symptomatic
   float PERCENT_SYMPTOMATIC       = 0.669;  // Percentage of infected agents that become symptomatic
   float CHILD_PERCENT_SYMPTOMATIC = 0.134;
   constexpr static float DAYS_PER_STEP_ = 0.5;
   constexpr static float exponent_  = 10.0;

   // Constant attributes associated to this agent
   const unsigned int ID_;
   const AgeGroup     age_group_;

   bool new_illness_;
   bool newly_detected_;

   std::vector<std::shared_ptr<WorkingGroup> > working_groups_;
   std::shared_ptr<Household> household_;

   // Disease states of the agent
   DiseaseState state_;
   bool         infected_prior_;
   bool         illness_detected_;
   bool         detected_recovered_;
   unsigned     steps_since_infection_;
   unsigned     steps_since_recovery_;

   // Immunity after an infection.
   float VEc_natural_after_recovered = 0;
   float VEt_natural_after_recovered = 0;
   float daily_waning_VEc_natural_after_recovered = 0;
   float daily_waning_VEt_natural_after_recovered = 0;

   SD_State SDstate_;
   CI_State CIstate_;
   HQ_State HQstate_;
   SC_State SCstate_;

   Compliance_State Compliance_state_;
   Isolation_State Isolation_state_;

   float isolation_factor_home_;
   float isolation_factor_work_;
   float isolation_factor_community_;

   int HQclock_;
   float VacGroup;

   // Opinion dynamics variables to store agent's opinion on compliance to an intervention (e.g., SD)
   float rational_opinion_on_compliance;

   int incubating_period_;

   // Add to individual nighttime mixing groups
   inline void addToHouseholdProbability(float factor) {  
      // Add infectious agent IDs to store in this context
      this->getHH()->addInfectedAgent(this->get_trans_scaler(), this->getAgeGroup(), factor, this->getID());
   }

   inline void addToClusterProbability(float factor) {
      // Add infectious agent IDs to store in this context
      this->getHC()->addInfectedAgent(this->get_trans_scaler(), this->isAdult(), factor, this->getID());
   }

   inline void addToCommunityProbability(float factor) {
      this->getSLA()->addInfectedAgent(this->get_trans_scaler(), this->getAgeGroup(), factor, this->getID());
   }

   inline void addToNeighbourhoodProbability(float factor) {  
      this->getCD()->addInfectedAgent(this->get_trans_scaler(), this->getAgeGroup(), factor, this->getID());
   }

public:

   /* Default constructor */
   Agent(unsigned long ID, AgeGroup ag)
      : ID_(ID),
        age_group_(ag),
        household_(NULL),
        state_(DiseaseStates::StateSusceptible),
        SDstate_(SD_States::SD_noncompliant),
        CIstate_(CI_States::CI_noncompliant),
        HQstate_(HQ_States::HQ_noncompliant),
        SCstate_(SC_States::SC_nil),
        Compliance_state_(Compliance_States::nil),
        Isolation_state_(Isolation_States::nil),
        isolation_factor_home_(1),
        isolation_factor_work_(1),
        isolation_factor_community_(1),
        HQclock_(0),
        infected_prior_(false),
        illness_detected_(false),
        detected_recovered_(false),
        steps_since_infection_(0),
        steps_since_recovery_(0),
        new_illness_(false),
        newly_detected_(false)
   {
      for (int seed_idx=0; seed_idx < omp_get_num_threads() + 128; seed_idx++) {
         seed_thread.emplace_back((unsigned) time(NULL) + ID % 42*seed_idx);
      }
   }

   Virus infection_info;
   void initiate_infection_info(Virus& viral_info);                        // For single-source infection
   void initiate_infection_info(Virus& viral_info_1, Virus& viral_info_2); // For co-infection
   void set_and_sample_all_disease_params(void);

   /*
    * Query methods
    */
   inline bool isInfected(void)
   {
      if ((state_ == DiseaseStates::StateSymptomatic) || (state_ == DiseaseStates::StateAsymptomatic))
      {
         return(true);
      }
      return(false);
   }

   inline bool isIll(void)
   {
      if (state_ == DiseaseStates::StateSymptomatic)
      {
         return(true);
      }
      return(false);
   }

   inline bool isInfectedPrior(void)
   {
      return(infected_prior_);
   }

   inline bool isIllnessDetected(void)
   {
      return(illness_detected_);
   }

   inline bool wasIllnessDetected(void)
   {
     return(detected_recovered_);
   }

   inline bool isSusceptible(void)
   {
      return(state_ == DiseaseStates::StateSusceptible);
   }

   inline bool isRemoved(void)
   {
      return(state_ == DiseaseStates::StateRemoved);
   }

   inline bool isWithdrawn(void)
   {
      return(state_ == DiseaseStates::StateWithdrawn);
   }

   inline bool isChild(void)
   {
      return(age_group_ == AgeGroups::Age0to4 || age_group_ == AgeGroups::Age5to18);
   }

   inline bool isStudent(void)
   {
      return(age_group_ == AgeGroups::Age5to18 && working_groups_.size() == 3);
   }

   inline bool isAdult(void)
   {
      return(!this->isChild());
   }

   inline bool isTeacher(void)
   {
      return(this->isAdult() && working_groups_.size() == 3);
   }

   inline bool willBeSymptomatic(void)
   {
      if (incubating_period_ > 0)
      {
         return(true);
      }
      return(false);
   }

   /*
    * Getter methods
    */

   float get_trans_scaler(void) {
      return trans_scaler;
   }

   unsigned getPEAK_VIRAL_TITER_(void)
   {
      return(PEAK_VIRAL_TITER_);
   }

   unsigned getLATENT_PERIOD_(void)
   {
      return(LATENT_PERIOD_);
   }

   float getASYMPTOMATIC_INFECTIVITY_(void)
   {
      return(ASYMPTOMATIC_INFECTIVITY_);
   }

   unsigned getINFECTION_PERIOD_(void)
   {
      return(INFECTION_PERIOD_);
   }

   unsigned getAFTER_RECOVERY_PERIOD_(void)
   {
      return(AFTER_RECOVERY_PERIOD_);
   }

   float getPERCENT_SYMPTOMATIC(void)
   {
      return(PERCENT_SYMPTOMATIC);
   }

   float getCHILD_PERCENT_SYMPTOMATIC(void)
   {
      return(CHILD_PERCENT_SYMPTOMATIC);
   }

   inline unsigned long getID(void)
   {
      return(ID_);
   }

   inline const std::shared_ptr<Household>& getHH(void)
   {
      return(household_);
   }

   inline const std::shared_ptr<HouseholdCluster>& getHC(void)
   {
      return(household_->getHC());
   }

   inline const std::shared_ptr<CensusDistrict>& getCD(void)
   {
      return(household_->getHC()->getCD());
   }

   inline const std::shared_ptr<StatisticalLocalArea>& getSLA(void)
   {
      return(household_->getHC()->getCD()->getSLA());
   }

   inline const std::vector<std::shared_ptr<WorkingGroup> >& getWGs(void)
   {
      return(working_groups_);
   }

   inline const std::shared_ptr<DestinationZone>& getDZ(void)
   {
      if (!working_groups_.empty())
      {
         return(working_groups_.front()->getDZ());
      }
      else
      {
         std::cout << "\ttesting... get DZ" << std::endl;
         // return(NULL);
      }
   }

   inline bool isNewIllness(void)
   {
      return(new_illness_);
   }

   inline bool isNewlyDetected(void)
   {
      return(newly_detected_);
   }

   inline void resetNewIllness(void)
   {
      new_illness_ = false;
   }

   inline void resetNewDetection(void)
   {
      newly_detected_ = false;
   }

   inline AgeGroup getAgeGroup(void)
   {
      return(age_group_);
   }

   inline DiseaseState getState(void)
   {
      return(state_);
   }

   inline SD_State getSDState(void)
   {
      return(SDstate_);
   }                                

   inline CI_State getCIState(void)
   {
      return(CIstate_);
   }

   inline HQ_State getHQState(void)
   {
      return(HQstate_);
   }

   inline SC_State getSCState(void)
   {
      return(SCstate_);
   }

   inline Isolation_State getIsolationState(void)
   {
      return(Isolation_state_);
   }

   inline Compliance_State getComplianceState(void)
   {
      return(Compliance_state_);
   }

   inline float getIsolationFactor_home(void)
   {
      return(isolation_factor_home_);
   }

   inline float getIsolationFactor_work(void)
   {
      return(isolation_factor_work_);
   }

   inline float getIsolationFactor_community(void)
   {
      return(isolation_factor_community_);
   }

   void eval_Compliance_state(void);

   void eval_Compliance_state_Qmod(void);

   int findParent_ID(std::map<unsigned long, std::shared_ptr<Agent>>&);

   float getDaytimeProbabilityOfInfection(void);
   float getDaytimeProbabilityOfInfection(unsigned);
   float getNighttimeProbabilityOfInfection(void);

   void getDaytimePotentialSourcesOfInfection(std::vector<unsigned>&, std::vector<float>&);
   void getNighttimePotentialSourcesOfInfection(std::vector<unsigned>&, std::vector<float>&);

   float getHouseholdProbabilityOfInfection(void);

   float getHouseholdClusterProbabilityOfInfection(void);

   float getNeighbourhoodProbabilityOfInfection(void);

   float getCommunityProbabilityOfInfection(void);

   // Getter and setter functions for Opinion Dynamics
   float get_rational_opinion_on_compliance();
   void  set_rational_opinion_on_compliance(float new_opinion_value);

   /*
    * Setter methods
    */

   inline void setTrans_scaler(float trans_scaler_new) {
      trans_scaler = trans_scaler_new;
   }

   inline void setPEAK_VIRAL_TITER_(unsigned new_peak_time)
   {
      PEAK_VIRAL_TITER_ = new_peak_time;
   }

   inline void setLATENT_PERIOD_(unsigned new_latent_period)
   {
      LATENT_PERIOD_ = new_latent_period;
   }

   inline void setINFECTION_PERIOD_(unsigned new_infection_period)
   {
      INFECTION_PERIOD_ = new_infection_period;
   }

   inline void setAFTER_RECOVERY_PERIOD_(unsigned new_after_recovery_period)
   {
      AFTER_RECOVERY_PERIOD_ = new_after_recovery_period;
   }

   inline void setPERCENT_SYMPTOMATIC(float new_factor)
   {
      PERCENT_SYMPTOMATIC = new_factor;
   }

   inline void setCHILD_PERCENT_SYMPTOMATIC(float new_factor)
   {
      CHILD_PERCENT_SYMPTOMATIC = new_factor;
   }

   inline void setASYMPTOMATIC_INFECTIVITY_(float new_factor)
   {
      ASYMPTOMATIC_INFECTIVITY_ = new_factor;
   }

   inline void addWG(std::shared_ptr<WorkingGroup> wg)
   {
      working_groups_.push_back(wg);
   }

   inline void setHH(std::shared_ptr<Household> hh)
   {
      household_ = hh;
   }

   inline void setState(DiseaseState new_state)
   {
      this->state_ = new_state;
   }                                                                               

   inline void setSDState(SD_State new_SDstate)
   {
      this->SDstate_ = new_SDstate;
   }                                            

   inline void setCIState(CI_State new_CIstate)
   {
      this->CIstate_ = new_CIstate;
   }

   inline void setIsolationState(Isolation_State new_state)
   {
      this->Isolation_state_ = new_state;
   }

   inline void setSCState(SC_State new_state)
   {
      this->SCstate_ = new_state;
   }

   inline void setHQState(HQ_State new_HQstate)
   {
      this->HQstate_ = new_HQstate;
   }

   inline void setComplianceState(Compliance_State new_state)
   {
      this->Compliance_state_ = new_state;
   }

   float computeFactor(unsigned);

   void init(void);

   void step(unsigned);

   void step_HQ(int duration_HQ);

   void setIsolationFactor_home(float FoI_SD_home, float FoI_CI_home, float FoI_HQ_home, float FoI_SC_home);

   void setIsolationFactor_work(float FoI_SD_work, float FoI_CI_work, float FoI_HQ_work, float FoI_SC_work);

   void setIsolationFactor_community(float FoI_SD_community, float FoI_CI_community, float FoI_HQ_community, float FoI_SC_community);

   bool Illness_detection(float& Detection);

   // Add to all mixing groups during daytime
   void addToDaytimeProbabilityOfInfection(unsigned);
   void addToDaytimeProbabilityOfInfection_SD(unsigned);

   // Add to all mixing groups during nighttime
   void addToNighttimeProbabilityOfInfection(unsigned);
   void addToNighttimeProbabilityOfInfection_SD(unsigned);

   void infect(unsigned, float, float, float, float);

   bool CD_Sympotmatic_present(void);
   float computeViralLoad(int t);
   void printTiterVsTime(void);

   // Immunisation history of past infection and vaccination
   std::map<unsigned, std::shared_ptr<Immunity>> immunisation_history;
   
   // Few functions related to the immplementation of immunisation history and its use.
   void clear_immunisation_history();
   void add_an_immunisation_to_history(unsigned, float, float, float, float, Immunisation_Type);
   Immunisation_Summary_Type get_immunisation_summary_type();
   float get_VEc_from_immunisation_history(unsigned);
   float get_VEs_from_immunisation_history(unsigned);
   float get_VEd_from_immunisation_history(unsigned);
   float get_VEt_from_immunisation_history(unsigned);

};

typedef std::pair<unsigned long, std::shared_ptr<Agent> >   agent_pair_type;
typedef std::map<unsigned long, std::shared_ptr<Agent> >    agent_map_type;

#endif
