/*
   *Copyright:  

   ACEMod (2018--2021), The University of Sydney's invention disclosure CDIP Ref. 2019-123.
   Copyright ©  Oliver Cliff, Cameron Zachreson, Nathan Harding, Kristopher Fair, Mikhail Prokopenko, Mahendarajah Piraveenan

   AMTraC-19 (2020-2021), The University of Sydney's invention disclosure CDIP 2020-018 (2020).
   Copyright ©  Sheryl Chang, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   AMTraC-19 (2022-2023), The University of Sydney.
   Copyright ©  Sheryl Chang, Quang Dang Nguyen, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   AMTraC-19 (2024), The University of Sydney.
   Copyright ©  Quang Dang Nguyen, Sheryl Chang, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   *EndCopyright:
*/

#include "epidemic.h"

#include <fstream>
#include <sstream>
#include <string>
#include <random>
#include <chrono>

#include <sys/time.h>

std::vector <std::pair <unsigned long, std::vector <unsigned long> > > Epidemic::n_commuters_to_airport_slas_;
std::vector <std::pair <unsigned long, std::vector <unsigned long> > > Epidemic::n_commuters_to_airport_agents_;
std::vector <unsigned long> Epidemic::agent_list_;

const float Epidemic::expected_number_infected_ = 0.00004;//0.00004;

std::string Epidemic::seeded_sla_fname { "SeedSLAs.dat" };

void Epidemic::init_transmission_rates(float p_trans)
{
   /* Initialise all transmission rates for all statistical areas */
   WorkingGroup::addTransmissionRate(p_trans * contact_rate_wg);
   for (auto c_tr : transmission_rate_schools) {
      float p = 1 - pow(1 - c_tr, 1.0 / 4.1);
      WorkingGroup::addTransmissionRate(p);
   }

   for (auto c_rate : contact_rate_sla) {
      StatisticalLocalArea::addTransmissionRate(p_trans * c_rate);
   }

   for (auto c_rate : contact_rate_cd) {
      CensusDistrict::addTransmissionRate(p_trans * c_rate);
   }

   for (unsigned i = 0; i < contact_rate_hc.size(); ++i) {
      auto contact_rate_child_or_adult = contact_rate_hc.at(i);
      for (unsigned j = 0; j < contact_rate_child_or_adult.size(); ++j) {
         /* Add transmission rates from child (0) or adult (1) */
         HouseholdCluster::addTransmissionRate(i, p_trans * contact_rate_child_or_adult.at(j));
      }
   }

   for (unsigned i = 0; i < transmission_rate_hh.size(); ++i) {
      auto c_hh_size = transmission_rate_hh.at(i);
      for (unsigned j = 0; j < c_hh_size.size(); ++j) {
         float p = 1 - pow(1 - c_hh_size.at(j), 1.0 / 4.1);

         /* Add transmission rate for household size i+1 */
         Household::addTransmissionRate(i, p);
      }
   }
}

void Epidemic::init_airport_seeded_slas(const std::string& database, const sla_map_type& sla_map)
{
   /* Open the file for reading */
   std::string seed_file(database + seeded_sla_fname);

   std::ifstream seed_stream(seed_file);

   std::string cline;
   unsigned    ap = 0;

   while (std::getline(seed_stream, cline))
   {
      std::stringstream i_ss;
      i_ss << cline;

      unsigned incoming_passengers;
      i_ss >> incoming_passengers;
      n_commuters_to_airport_slas_.emplace_back(incoming_passengers, std::vector <unsigned long>());
      n_commuters_to_airport_agents_.emplace_back(incoming_passengers, std::vector <unsigned long>());

      unsigned long c_sla;

      while (i_ss >> c_sla)
      {
         n_commuters_to_airport_slas_.back().second.push_back(c_sla);

         for (auto& agent_idx : sla_map.at(c_sla)->getAgentIdxs())
         {
            n_commuters_to_airport_agents_.back().second.push_back(agent_idx);
         }
      }
   }
   seed_stream.close();
}

unsigned Epidemic::get_one_airport_agent_idx(unsigned airport, agent_map_type& agents)
{
   auto com_to_ap = n_commuters_to_airport_agents_.begin();

   std::advance(com_to_ap, airport);

   /* Choose one agent in SLAs surrounding this airport */
   auto&         potential_agent_idxs = com_to_ap->second;
   unsigned int  rval             = rand() % potential_agent_idxs.size();
   unsigned long chosen_agent_idx = potential_agent_idxs.at(rval);

   std::cout << "Chosen agent " << chosen_agent_idx << " (out of "
             << potential_agent_idxs.size() << " options) in airport " << airport << std::endl;
   return(chosen_agent_idx);
}

// Initiate the infection info for each seeded cases.
void Epidemic::seed_epi_airports(agent_map_type& agents, std::vector <unsigned long>& infected_ids, std::vector<Virus>& virus_list, std::vector<float>& virus_weight_list, unsigned day, float VEc_NR, float VEt_NR, float daily_waning_VEc_NR, float daily_waning_VEt_NR) {
   std::default_random_engine generator(rand());        /* Kit's edit 06.07.18 - random seed */
   unsigned n_infected = 0;

   for (auto& com_to_agents : n_commuters_to_airport_agents_)
   {
      /* Draw number to infect from binomial */
      unsigned int max_to_infect = com_to_agents.first;
      std::binomial_distribution <int> bd(max_to_infect, expected_number_infected_);
      unsigned int n_to_infect = bd(generator);

      /* Infect random agents surrounding this airport SLA */
      auto& potential_agent_idxs = com_to_agents.second;
      for (unsigned i = 0; i < n_to_infect; ++i)
      {
         unsigned int  rval             = rand() % potential_agent_idxs.size();
         unsigned long chosen_agent_idx = potential_agent_idxs.at(rval);
         if (!agents.at(chosen_agent_idx)->isInfected())
         {            
            // Sample a viral infection information from a list for this newly seeded case.
            std::mt19937 gen(rand() + time(NULL));
            std::discrete_distribution<> dist(virus_weight_list.begin(), virus_weight_list.end());
            unsigned int sampled_viral_infection_info_idx = (unsigned int) dist(gen);

            agents.at(chosen_agent_idx)->initiate_infection_info(virus_list[sampled_viral_infection_info_idx]);
            agents.at(chosen_agent_idx)->infect(day, VEc_NR, VEt_NR, daily_waning_VEc_NR, daily_waning_VEt_NR);
            infected_ids.push_back(chosen_agent_idx);
            n_infected++;

            std::cout << "\t   - Seeded agent " << chosen_agent_idx << " with virus having code " << agents.at(chosen_agent_idx)->infection_info.get_origin_code() << std::endl;
         }
      }
   }
}

