/*
   *Copyright:  

   ACEMod (2018--2021), The University of Sydney's invention disclosure CDIP Ref. 2019-123.
   Copyright ©  Oliver Cliff, Cameron Zachreson, Nathan Harding, Kristopher Fair, Mikhail Prokopenko, Mahendarajah Piraveenan

   AMTraC-19 (2020-2021), The University of Sydney's invention disclosure CDIP 2020-018 (2020).
   Copyright ©  Sheryl Chang, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   AMTraC-19 (2022-2023), The University of Sydney.
   Copyright ©  Sheryl Chang, Quang Dang Nguyen, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   AMTraC-19 (2024), The University of Sydney.
   Copyright ©  Quang Dang Nguyen, Sheryl Chang, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   *EndCopyright:
*/

// Compile the code: g++ generate_population_2021.cpp -std=c++11 -O2
// Check our userguide for more information on how to run the code.

#ifndef GENERATE_POPULATION_H_
#define GENERATE_POPULATION_H_

#include "enums.h"

std::string current_dir = (char *)get_current_dir_name();
const std::string database { current_dir + "/AmTraC_inputs/2021_population/"};
const std::string output_directory { current_dir + "/AmTraC_inputs/2021_population/"};

/*
 * Sample spaces for CDFs
 */
const std::vector <unsigned> hh_comps { /* Single/lone household */
   HouseholdCompositions::Single,
   /* Couples without children */
   HouseholdCompositions::CouplesWithoutChildren,
   /* Couples with children */
   HouseholdCompositions::CouplesWithOneChild,
   HouseholdCompositions::CouplesWithTwoChildren,
   HouseholdCompositions::CouplesWithThreeChildren,
   HouseholdCompositions::CouplesWithFourChildren,
   /* Single parent families */
   HouseholdCompositions::SingleParentFamilyOneChild,
   HouseholdCompositions::SingleParentFamilyTwoChildren,
   HouseholdCompositions::SingleParentFamilyThreeChildren,
   HouseholdCompositions::SingleParentFamilyFourChildren,
   HouseholdCompositions::SingleParentFamilyFiveChildren,
   /* Group (share) households */
   HouseholdCompositions::GroupTwo,
   HouseholdCompositions::GroupThree,
   HouseholdCompositions::GroupFour,
   HouseholdCompositions::GroupFive,
   HouseholdCompositions::GroupSix
};

const std::vector <Sex> sexes { Sexes::Male, Sexes::Female };

const std::vector <AgeGroup> adult_agegroups { AgeGroups::Age19to29, AgeGroups::Age30to64, AgeGroups::Age65Plus };

const std::vector <AgeGroup> child_agegroups { AgeGroups::Age0to4, AgeGroups::Age5to18 };

/* Useful typedefs for statistical areas */
typedef unsigned long long                                       dz_type;
typedef unsigned long                                            cd_type;
typedef unsigned long                                            sla_type;
typedef unsigned long                                            wg_type;



//typedefs for school configuration function inputs
typedef unsigned long                                            postcode_type;
typedef std::map <dz_type, unsigned long>                        dz_numAgents_map_type;
typedef std::map <postcode_type, std::vector <sla_type> >        postcode_sla_map_type;
typedef std::map <sla_type, std::vector <dz_type> >              sla_dz_map_type;
typedef std::map <unsigned long, postcode_type>                  school_postcode_map_type;
typedef std::map <unsigned long, unsigned long>                  school_schoolsize_map_type;
typedef std::map <unsigned long, std::vector <unsigned long> >   schoolsize_school_map_type;
typedef std::map <sla_type, unsigned long>                       sla_numStudents_map_type;
typedef std::map <unsigned long, std::vector <sla_type> >        numStudents_sla_map_type;

// for student adjacency list
typedef std::map <std::pair <cd_type, dz_type>, unsigned long>   cd_dz_nagents_map_type;


//typedefs for school configuration function outputs
struct School
{
   unsigned long ID;
   unsigned long max_students;
   unsigned long num_students;
   unsigned long DZ;
   unsigned long SLA;
};
typedef std::map <unsigned long, struct School *>                                school_id_map_type;
typedef std::map <sla_type, std::vector <struct School *> >                      sla_school_map_type;



typedef std::map <cd_type, std::vector <std::pair <dz_type, unsigned long> > >   cd_dz_nagents_type;

/* Some constants for generating our population */
const float PERC_ADULTS_AT_HOME = 0.01;
//const float PERC_ADULTS_AT_HOME = 0.0;
const float SCHOOL_ATTENDANCE_LIKELIHOOD = 0.95; // may have to adjust this

const unsigned NUM_PEOPLE_IN_WG   = 20;
const unsigned HOUSES_PER_CLUSTER = 4;
//const float NUM_STUDENTS_TO_TEACHERS = 8.5;
const float NUM_STUDENTS_TO_TEACHERS = 8.5;  // may revise this by year since ratio is known or can be calculated from ACARA data
const float SCHOOL_ZONE_SIGMA        = 0.25; // in EpiCast units (1 unit ~= 100km, i.e., average commute distance is 15km)
// don't need this variable anymore
//const float MAX_STUDENT_SCALER = 0.4; // For SCHOOL-0.9.dat
const float    MAX_STUDENT_COMMUTE_DISTANCE = 1.5; // Max travel distance is 150km
const unsigned MAX_ALLOCATION_ATTEMPTS      = 5;
const unsigned SCHOOL_STEP_SIZE             = 5;

const unsigned long DZ_MULTIPLIER  = 100000000;
const unsigned long SLA_MULTIPLIER = 10000;

// now that school placement is pseudo-deterministic, I don't need these vectors anymore
//const std::vector< std::string > STATE_STRINGS{ "NSW", "VIC", "QLD", "SA", "WA", "TAS", "NT", "ACT" };
//const std::vector< unsigned > SCHOOL_BOUNDS{ 0, 35, 100, 200, 300, 400, 600, 800, 1000, 2500 }; /* Max number of students in school */

/* Simple agent structure */
struct SimpleAgent
{
   unsigned long my_id;
   AgeGroup      my_age;
   cd_type       my_cd;
   signed long   my_wg;
   unsigned long my_hh;
   unsigned long my_hc;

   inline bool   isAdult(void)
   {
      return(!this->isChild());
   }

   inline bool   isChild(void)
   {
      return(my_age == AgeGroups::Age0to4 || my_age == AgeGroups::Age5to18);
   }

   inline bool   isStudentAge(void)
   {
      return(my_age == AgeGroups::Age5to18);
   }

   inline bool   isStudent(void)
   {
      return(my_wg < 0);
   }

   // for validation maps
   inline bool   isBaby(void)
   {
      return(my_age == AgeGroups::Age0to4);
   }

   inline bool   isYoungAdult(void)
   {
      return(my_age == AgeGroups::Age19to29);
   }

   inline bool   isMiddleAged(void)
   {
      return(my_age == AgeGroups::Age30to64);
   }

   inline bool   isElderly(void)
   {
      return(my_age == AgeGroups::Age65Plus);
   }
};

struct WGLimitStruct
{
   unsigned long lower;
   unsigned long upper;
};


#endif // GENERATE_POPULATION_H_
