/*
   *Copyright:  

   ACEMod (2018--2021), The University of Sydney's invention disclosure CDIP Ref. 2019-123.
   Copyright ©  Oliver Cliff, Cameron Zachreson, Nathan Harding, Kristopher Fair, Mikhail Prokopenko, Mahendarajah Piraveenan

   AMTraC-19 (2020-2021), The University of Sydney's invention disclosure CDIP 2020-018 (2020).
   Copyright ©  Sheryl Chang, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   AMTraC-19 (2022-2023), The University of Sydney.
   Copyright ©  Sheryl Chang, Quang Dang Nguyen, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   AMTraC-19 (2024), The University of Sydney.
   Copyright ©  Quang Dang Nguyen, Sheryl Chang, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   *EndCopyright:
*/

// Compile the code: g++ -std=c++11 virus.cpp agent.cpp create_population.cpp epidemic.cpp statistical_areas.cpp main.cpp -fopenmp -o AMTRAC
// Check our userguide for more information on how to run the code.

#include <cstddef>
#include <fstream>
#include <iostream>
#include <sstream>
#include <map>
#include <string>
#include <vector>
#include <memory>
#include <time.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <getopt.h>
#include <unistd.h>
#include <random>
#include <stack>
#include <vector>
using namespace std;

#include "utils.hpp"
#include "agent.h"
#include "create_population.h"
#include "epidemic.h"
#include "debug.h"
#include "enums.h"

#include <omp.h>

class EpidemicSimulator
{
public:
    EpidemicSimulator(int argc, char *argv[]);
    void reset();
    void reset_run(unsigned run_id, std::vector<bool> &list_SD_compliant);
    void parse_arguments(int argc, char *argv[]);
    void parse_epidemic_params();
    void parse_pharm_intervention_params();
    void parse_non_pharm_intervention_params();
    void parse_daily_SD_level();
    void parse_daily_fatigue_rate();

    void print_summary();
    void print_params_to_files();

    void initiate_bookkeeping_everyday(unsigned run_id);
    void bookkeeping_everyday(unsigned run_id, unsigned day);
    void close_bookkeeping_everyday();
    int count_CD_incidence();
    int count_CD_prevalence();

    void initiate_a_random_seed(unsigned run_id);

    string basedir_generator();
    void simulate_a_single_run(unsigned run_id);
    void simulate_all_runs();

    void rezero_before_simulating();
    void rezero_everyday();
    void clearDayTimeInfectionLists();
    void clearNightTimeInfectionLists();

    void assign_SD_compliance();
    void assign_SD_compliance_Qmod(float SD_compliance_level, int days_since_last_SD_assignment, std::vector<bool> &list_SD_compliant);
    float assign_SD_compliance_opinion_Qmod(int days_since_last_SD_assignment, std::vector<unsigned>& detected_incidence_buffer, std::vector<bool>& list_SD_compliant, float fatigue_rate, unsigned day);

    void assign_CI_compliance();
    void assign_HQ_compliance();
    void assign_SC_compliance_students_teachers();
    void set_Compliance_states_all();
    void set_Compliance_states_all_Qmod();

    void pre_epidemic_vaccination(std::stack<unsigned long> &unvaccinated_senior_IDs_s,
                                  std::stack<unsigned long> &unvaccinated_adult_IDs_s,
                                  std::stack<unsigned long> &unvaccinated_child_IDs_s,
                                  unsigned day);

    unsigned long getNumAgentsInfected();
    unsigned long getNumAgentsIll();
    unsigned long getNumAgentsRemoved();

    void assign_SC_compliance_parents();
    void step_and_detect(unsigned);
    void step_without_detection(unsigned);
    void update_Isolation_states(bool SD_flag, bool CI_flag, bool HQ_flag, bool SC_flag);

    void set_Isolation_Factors_home(float FoI_SD, float FoI_CI, float FoI_HQ, float FoI_SC);
    void set_Isolation_Factors_work(float FoI_SD, float FoI_CI, float FoI_HQ, float FoI_SC);
    void set_Isolation_Factors_community(float FoI_SD, float FoI_CI, float FoI_HQ, float FoI_SC);

    void sampleSourcesForAnInfection(vector<unsigned> &, vector<float> &, vector<unsigned> &);
    void updateAgentStateAllContexts(std::shared_ptr<Agent>, float, bool, unsigned);

    void simulateDaytime(unsigned);
    void simulateNighttime(unsigned);
    void simulateDaytimeWeekend(unsigned);

    void simulateDaytime_SD(unsigned);
    void simulateNighttime_SD(unsigned);
    void simulateDaytimeWeekend_SD(unsigned);

    void update_HQ();

    void progressive_vaccination_general(std::stack<unsigned long> &, std::stack<unsigned long> &, std::stack<unsigned long> &, unsigned);
    void progressive_vaccination_priority(std::stack<unsigned long> &, std::stack<unsigned long> &, std::stack<unsigned long> &, unsigned);

private:
    // General parameters
    unsigned n_runs = 1;
    unsigned n_days = 196;
    unsigned n_agents = 0;
    unsigned n_recovered = 0;

    std::default_random_engine generator;

    struct timeval tv;
    char buffer[26];

    // Disease parameters
    float p_trans = 0.08;
    float trans_scaler_mean = 3.33;
    float trans_scaler_std = 0;
    float latent_period_mean = 0;
    float latent_period_std = 0;
    float incubation_period_mu = 1.013;
    float incubation_period_sig = 0.413;
    float recovery_period_hi = 11;
    float recovery_period_low = 7;
    float after_recovery_period_mean = 28;
    float after_recovery_period_std = 0;
    float adult_symptomatic_fraction = 0.67;
    float child_symptomatic_fraction = 0.268;
    float asymptomatic_infectivity = 0.3;
    float co_infection_rate = 0.0; // rate of co-infection. Temporarily it is not enabled in this code version.

    // -----------------------
    // Intervention variables
    // -----------------------

    // proportion of population to pre-vaccinate
    float perc_vac = 0.0;

    // progressive vaccination - some vaccines are delivered pre-epidemic, and some are delivered during epidemic
    int n_vac_per_week_PV = 0;
    float fraction_vac_pre_epidemic_PV = 0.0;
    int n_vac_pre_epidemic_PV = -1; // this is the number of immunisations delivered prior to epidemic outset
                                    // If n_vac_pre_epidemic_1 < 0, it will be calculated from actual population size and the expected fraction, not hardcoded.
                                    // Otherwise, fraction_vac_pre_epidemic_1 is calculated according to this number.

    int n_vac_per_week_GV = 0;
    float fraction_vac_pre_epidemic_GV = 0.0;
    int n_vac_pre_epidemic_GV = -1; // this is the number of immunisations delivered prior to epidemic outset
                                    // If n_vac_pre_epidemic_2 < 0, it will be calculated from actual population size and the expected fraction, not hardcoded.
                                    // Otherwise, fraction_vac_pre_epidemic_2 is calculated according to this number.

    // age-stratified vaccination
    float p_vac_children = 1;    // age groups 0 and 1
    float p_vac_18_to_64 = 10;   // age groups 2 and 3
    float p_vac_65_and_over = 1; // agegroup 4

    unsigned long num_vac = 0;
    unsigned long num_vac_65_plus = 0;
    unsigned long num_vac_18_to_64 = 0;
    unsigned long num_vac_children = 0;

    unsigned long num_vac_PV = 0;
    unsigned long num_vac_65_plus_PV = 0;
    unsigned long num_vac_18_to_64_PV = 0;
    unsigned long num_vac_children_PV = 0;

    unsigned long num_vac_GV = 0;
    unsigned long num_vac_65_plus_GV = 0;
    unsigned long num_vac_18_to_64_GV = 0;
    unsigned long num_vac_children_GV = 0;

    float VEc_NR = 0.7;                                     // natural immunity against symptomatic infection
    float VEt_NR = 0.4;                                     // natural immunity against infection transmission
    float per_cycle_waning_VEc_NR = 0.00034;                // per-cycle waning rate of natural immunity against symptomatic infection
    float daily_waning_VEc_NR = per_cycle_waning_VEc_NR * 2;// daily waning rate of natural immunity against symptomatic infection
    float per_cycle_waning_VEt_NR = 0.00034;                // per-cycle waning rate of natural immunity against infection transmission
    float daily_waning_VEt_NR = per_cycle_waning_VEt_NR * 2;// daily waning rate of natural immunity against infection transmission

    float VEc_PV = 0.7;                                     // vaccine efficacy against symptomatic infection
    float VEt_PV = 0.4;                                     // vaccine efficacy of infection transmission
    float per_cycle_waning_VEc_PV = 0.00034;                // per-cycle waning rate of priority vaccine efficacy against symptomatic infection
    float daily_waning_VEc_PV = per_cycle_waning_VEc_PV * 2;// daily waning rate of priority vaccine efficacy against symptomatic infection
    float per_cycle_waning_VEt_PV = 0.00034;                // per-cycle waning rate of priority vaccine efficacy against infection transmission
    float daily_waning_VEt_PV = per_cycle_waning_VEt_PV * 2;// daily waning rate of priority vaccine efficacy against infection transmission

    float VEc_GV = 0.5;                                     // vaccine efficacy against symptomatic infection
    float VEt_GV = 0.4;                                     // vaccine efficacy of infection transmission
    float per_cycle_waning_VEc_GV = 0.00034;                // per-cycle waning rate of general vaccine efficacy against symptomatic infection
    float daily_waning_VEc_GV = per_cycle_waning_VEc_GV * 2;// daily waning rate of general vaccine efficacy against symptomatic infection
    float per_cycle_waning_VEt_GV = 0.00034;                // per-cycle waning rate of general vaccine efficacy against infection transmission
    float daily_waning_VEt_GV = per_cycle_waning_VEt_GV * 2;// daily waning rate of general vaccine efficacy against infection transmission

    bool pharm_intervention = true;       // Boolean for enabling pharmaceutical intervention

    unsigned long n_airport_infections = 0;

    // -----------------------------------------------

    // symptomatic and pre-symptomatic detection rate
    float Detection_      = 0.12;         // probability of detection of illness
    float Detection_asymp = 0.01;         // probability of detection of pre- or asymptomatic infection

    // a sigmoid function to let detection rate vary, instead of staying at a fixed value. Detection is now calculated from a sigmoid function at the beginning of every day.
    float Detection_sigmoid_midpoint  = 0.0;
    float Detection_sigmoid_v = 0.25;
    float Detection_sigmoid_delay = 0.0;
    float Detection_base = 0;

    // Social distancing parameters
    bool SD_intervention = false;
    int n_detected_SD_trigger = 400; // cumulative attack rate triggering intervention

    // Horizon length of each agent is assumed to be sampled from a Normal distribution with mean determining by a sigmoid-like function: 
    // horizon-mean-to-sample = (2 * opinion-horizon-mean) / (1+e^(-opinion-horizon-sigmoid-u * (day - opinion-horizon-sigmoid-delay))
    // To reduce this sigmoid-like function to the constant function (for which we will sample each agent's horizon length from a Normal distribution with fixed mean), we can set opinion_horizon_mean_sigmoid_v = 0
    // This sigmoid-like function will make the transition from 0 to (2 * opinion-horizon-mean) happening around the day: opinion_horizon_mean_sigmoid_delay.
    // The setting for opinion_horizon_mean_sigmoid_v will determine how the transition happen (very sharp and step-like transition, u=2, or gradual increase, u=0.25)
    float opinion_horizon_std                   = 1.0;  // std of Normal distribution for opinion horizon
    float opinion_horizon_mean_sigmoid_midpoint = 7.0;  // mean of Normal distribution for opinion horizon
    float opinion_horizon_mean_sigmoid_v        = 0.25; // parameter to make the transition happen sharply/gently  
    float opinion_horizon_mean_sigmoid_delay    = 90;   // parameter to determine the day around which the transition happen
    float opinion_horizon_base                  = 7;    // An addition to the opinion sigmoid function 

    // After determining the mean of the Normal distibution to sample horizon length from, we can decide to apply averaging or weighted averaging to the horizon of detected incidence for each agent.
    float opinion_horizon_weight    = 1.0;         // Weight to be applied in the formula for weighted average for the horizon: weighted average incidence = sum ( w^(t-t_current + 1) * incidence(t) ) / sum ( w^(t-t_current + 1) ), t = t_current - horison : t_current - 1 

    bool  opinion_based_compliance  = false;       // Enable/Disable SD compliance based on opinion
    float beta_perc_initial         = 0.5;         // perceived infection rate from contact (initial value)
    float beta_perc_fatigue_rate    = 0.0002;        // perceived infection rate from contact
    float opinion_threshold         = 0.5;         // perceived infection threshold from contact

    float social_pressure_weight    = 0.4;         // weight for the impact from agents in shared social contexts to an agent. 
                                                   // weight for social influence is social_pressure_weight, while the weight from the agent itself will be scaled with the weight (1-social_pressure_weight)

    // We can decompose social_pressure into different pressure components from different contexts. Each context will be assigned with a specific weight showing the influence of opinion of other people in this
    // context to the agent of interest.
    float social_pressure_household         = 0.5;
    float social_pressure_householdcluster  = 0.0;
    float social_pressure_SLA               = 0.0;
    float social_pressure_CD                = 0.0;

    float social_pressure_workinggroup      = 0.5;
    float social_pressure_class             = 0.0;
    float social_pressure_grade             = 0.0;
    float social_pressure_school            = 0.0;

    // We assume a fraction of population is always in favour of complying with SD measures (their risk perception always stay high), and a fraction of population never comply SD measures (they are essential 
    // workers or their risk perception is always low).
    float fraction_always_compliant = 0.2;
    float fraction_never_compliant  = 0.25;

    float compliance_rate_SD = 0.5;   // fraction of people who comply
    int duration_SD = 200;            // duration of social distancing measures (in days)
    float FoI_SD_community = 0.25;    // factor by which an infected person's community contribution to FoI is multiplied with SD
    float FoI_SD_work = 0.1;          // factor by which an infected person's work/school contribution to FoI is multiplied with SD
    float FoI_SD_home = 1;            // factor by which an infected person's home contribution to FoI is multiplied with SD
    int SD_assignment_period = 1000;  // duration of compliance period for each agent.
    vector<float> daily_SD_level;     // daily SD level
    vector<float> daily_fatigue_rate; // daily fatigue rate

    // Create a vector to calculate the detected incidence for opinion dynamics.
    std::vector<unsigned> detected_incidence_buffer;

    // Build a list of agents' character (0: never compliant, 1: always compliant, 2: opinion-based)
    std::vector<unsigned> agents_character;

    // for bookkeeping everyday
    double SD_compliance_level = 0;

    // School closure parameters
    bool School_closure = false;
    int n_detected_SC_trigger = 100;
    int duration_SC = 121;
    float FoI_SC_community = 0.5;
    float FoI_SC_work = 0.0;
    float FoI_SC_home = 1;
    float p_parent_stays_home = 0.25;

    // Case isolation parameters
    bool CI_intervention = false;
    float compliance_rate_CI = 0.7;
    float FoI_CI_community = 0.25;
    float FoI_CI_work = 0.25;
    float FoI_CI_home = 1;
    int g_CI_agents = 0;            // the count of agents affected by CI, for output.

    // Home quarantine parameters
    bool HQ_intervention = false;   // Home quarantine, in response to illness in household
    float compliance_rate_HQ = 0.5; // compliance probability given CI compliance
    int duration_HQ = 7;            // length of quarantine period in days
    float FoI_HQ_community = 0.25;
    float FoI_HQ_work = 0.25;
    float FoI_HQ_home = 2.0;
    int g_HQ_agents = 0;            // the count of agents affected by HQ, for output.

    // A random number representing a class instance.
    unsigned long rnd_num;

    // File names and directories
    std::string current_dir = (char *)get_current_dir_name();
    std::string database_output = current_dir + "/AmTraC_outputs/2021_population/";
    std::string database_input = current_dir + "/AmTraC_inputs/2021_population/";
    std::string non_pharm_intervention_input_fname = "non_pharm_intervention_input/non_pharm_intervention_input.dat";
    std::string daily_SD_level_input_fname = "non_pharm_intervention_input/daily_SD_level.dat";
    std::string daily_fatigue_rate_input_fname = "non_pharm_intervention_input/daily_fatigue_rate.dat";
    std::string pharm_intervention_input_fname = "pharm_intervention_input/pharm_intervention_input.dat";
    std::string infection_parameter_input_fname = "infection_input/infection_input.dat";
    std::string non_pharm_intervention_input_file, daily_SD_level_input_file, pharm_intervention_input_file, virus_parameter_input_file, pop_fname, base_dir;

    // Variables for Summarising purposes
    unsigned long g_detected_incidence, g_detected_prevalence, g_detected_cumulative,
        g_infected_incidence, g_infected_prevalence, g_infected_cumulative,
        g_illness_incidence, g_illness_prevalence, g_illness_cumulative,
        g_number_removed;

    // These will count DETECTED inicidence and prevalence, not illnesses
    std::map<unsigned long, unsigned long> g_sla_prevalence_map,
        g_sla_number_removed_map,
        g_sla_incidence_map,
        g_sla_attack_rate_map;

    // These will count DETECTED incidence and prevalence, not illnesses
    std::map<AgeGroup, std::map<unsigned long, unsigned long>> g_ag_sla_prevalence_map,
        g_ag_sla_number_removed_map,
        g_ag_sla_incidence_map;

    // Adding an age-group global incidence and prevalence output for illness, detection, and infection, to avoid any confusion and facilitate analysis / sanity checking
    std::map<AgeGroup, unsigned long> g_ag_infection_incidence_map,
        g_ag_infection_prevalence_map,
        g_ag_illness_incidence_map,
        g_ag_illness_prevalence_map,
        g_ag_detection_incidence_map,
        g_ag_detection_prevalence_map,
        ag_pops_map;

    // Epidemic Seeding Variables
    SeedType seed_type;

    // Main data structures
    agent_map_type agents;
    hh_map_type households;
    hc_map_type household_clusters;
    cd_map_type census_districts;
    wg_map_type working_groups;
    wg_map_type schools, grades, classes;
    dz_map_type destination_zones;
    sla_map_type statistical_local_areas;

    // Bookkeeping - Output stats to files
    std::ofstream outfile_stream;
    std::ofstream outfile_stream_breakdown;
    std::ofstream outfile_stream_incidence;
    std::ofstream outfile_stream_prevalence;
    std::ofstream outfile_stream_CI_HQ;
    std::ofstream outfile_stream_infection_I_ag;
    std::ofstream outfile_stream_illness_I_ag;
    std::ofstream outfile_stream_detection_I_ag;
    std::ofstream outfile_stream_infection_P_ag;
    std::ofstream outfile_stream_illness_P_ag;
    std::ofstream outfile_stream_detection_P_ag;
    std::ofstream outfile_stream_P_vaccination;
    std::ofstream outfile_stream_G_vaccination;

    // New output files for V9 - compartments
    std::ofstream outfile_stream_N_health_states;
    std::ofstream outfile_stream_N_immunity_states;
    std::ofstream outfile_stream_N_recovered_susceptibles;

    // New output for Opinion Dynamics
    std::ofstream outfile_stream_Detected_incidence_SD_compliance;

    std::map<AgeGroup, std::shared_ptr<std::ofstream>> ag_outfile_stream_incidence;
    std::map<AgeGroup, std::shared_ptr<std::ofstream>> ag_outfile_stream_prevalence;

    std::vector<unsigned> seed_thread;
    std::vector<unsigned long> agent_ids;
    std::vector<unsigned long> hh_ids;
    std::vector<unsigned long> hc_ids;
    std::vector<unsigned long> sla_ids;
    std::vector<unsigned long> cd_ids;
    std::vector<unsigned long> dz_ids;
    std::vector<unsigned long> wg_ids;
    std::vector<unsigned long> school_ids;
    std::vector<unsigned long> grade_ids;
    std::vector<unsigned long> class_ids;
};


EpidemicSimulator::EpidemicSimulator(int argc, char *argv[])
{
    // Prepare for parallelisation. Separate seeds for each parallel thread.
    for (int seed_idx = 0; seed_idx < omp_get_num_threads() + 128; seed_idx++)
    {
        seed_thread.emplace_back((unsigned) time(NULL) + 42 * seed_idx);
    }

    g_detected_incidence = 0;
    g_detected_prevalence = 0;
    g_detected_cumulative = 0;
    g_infected_incidence = 0;
    g_infected_prevalence = 0;
    g_infected_cumulative = 0;
    g_illness_incidence = 0;
    g_illness_prevalence = 0;
    g_illness_cumulative = 0;
    g_number_removed = 0;

    // Removes all elements from the map containers, leaving the container with a size of 0.
    g_sla_prevalence_map.clear();
    g_sla_number_removed_map.clear();
    g_sla_incidence_map.clear();
    g_sla_attack_rate_map.clear();
    g_ag_sla_prevalence_map.clear();
    g_ag_sla_number_removed_map.clear();
    g_ag_sla_incidence_map.clear();
    g_ag_infection_incidence_map.clear();
    g_ag_infection_prevalence_map.clear();
    g_ag_illness_incidence_map.clear();
    g_ag_illness_prevalence_map.clear();
    g_ag_detection_incidence_map.clear();
    g_ag_detection_prevalence_map.clear();
    ag_pops_map.clear();

    // Generate a number representing a class instance for many purposes (e.g., avoiding conflicts in creating base_dir when multiple jobs are initiated at the same time)
    std::random_device rd;
    std::mt19937 rng(rd());
    std::uniform_int_distribution<int> uni(0, 100000);
    rnd_num = uni(rng);

    // Assign parameters defined from the command line.
    parse_arguments(argc, argv);

    // Select infection seeding type
    seed_type = SeedType::Airport;

    // Update file paths after processing the command line.
    non_pharm_intervention_input_file = database_input + non_pharm_intervention_input_fname;
    daily_SD_level_input_file = database_input + daily_SD_level_input_fname;
    pharm_intervention_input_file = database_input + pharm_intervention_input_fname;
    virus_parameter_input_file = database_input + infection_parameter_input_fname;
    pop_fname = database_input + "population.dat";

    // Create output folder
    base_dir = basedir_generator();

    // Random seed for random number generation.
    srand(time(NULL) + rnd_num);

    // Load parameters from files if existing
    parse_epidemic_params();
    parse_pharm_intervention_params();
    parse_non_pharm_intervention_params();
    parse_daily_SD_level();
    parse_daily_fatigue_rate();

    // reset and initialise all parameters
    reset();

    // save parameters
    print_params_to_files();
};

// ------------------------------
// Parse argument inputs (code "borrowed" from https://www.gnu.org/software/libc/manual/html_node/Getopt-Long-Option-Example.html#Getopt-Long-Option-Example).
void EpidemicSimulator::parse_arguments(int argc, char *argv[])
{
    static struct option long_options[] = {
        /*
         * Following options don’t set a flag.
         * We distinguish them by their indices.
         */
        {"ptrans", no_argument, 0, 'p'},
        {"trans_scaler_mean", no_argument, 0, 't'},
        {"trans_scaler_std", no_argument, 0, 'y'},
        {"runs", no_argument, 0, 'r'},
        {"days", no_argument, 0, 'd'},
        {"database_output", no_argument, 0, 'b'},
        {"database_input", no_argument, 0, 'f'},
        {"non_pharm_intervention_input_fname", no_argument, 0, 'q'},
        {"daily_SD_level_input_fname", no_argument, 0, 'l'}, 
        {"daily_fatigue_rate_input_fname", no_argument, 0, 'g'},
        {"pharm_intervention_input_fname", no_argument, 0, 'h'},
        {"infection_parameter_input_fname", no_argument, 0, 'v'},
        {0, 0, 0, 0}};

    for (;;)
    {
        /* getopt_long stores the option index here. */
        int option_index = 0;
        char c = getopt_long(argc, argv, "p:t:y:r:d:b:f:q:l:g:h:v:", long_options, &option_index);

        /* Options are all parsed */
        if (c == -1)
        {
            break;
        }

        switch (c)
        {
        case 'p':
            p_trans = std::stof(optarg);
            break;

        case 't':
            trans_scaler_mean = std::stof(optarg);
            break;

        case 'y':
            trans_scaler_std = std::stof(optarg);
            break;

        case 'r':
            n_runs = std::stoi(optarg);
            break;

        case 'd':
            n_days = std::stoi(optarg);
            break;

        case 'b':
            database_output = std::string(optarg);
            break;

        case 'f':
            database_input = std::string(optarg);
            break;

        case 'q':
            non_pharm_intervention_input_fname = std::string(optarg);
            break;

        case 'l':
            daily_SD_level_input_fname = std::string(optarg);
            break;

        case 'g':
            daily_fatigue_rate_input_fname = std::string(optarg);
            break;

        case 'h':
            pharm_intervention_input_fname = std::string(optarg);
            break;

        case 'v':
            infection_parameter_input_fname = std::string(optarg);
            break;

        case '?':
            /* getopt_long already printed an error message. */
            break;

        default:
            abort();
        }
        std::cout << "option -" << c << " with value `" << optarg << "'" << std::endl;
    }

    /* Print any remaining command line arguments (not options). */
    if (optind < argc)
    {
        std::cout << "Non-option ARGV-elements: ";
        while (optind < argc)
        {
            std::cout << argv[optind++] << ", ";
        }
        std::cout << std::endl;
    }
}

// ------------------------------
void EpidemicSimulator::print_summary()
{
    std::cout << "Starting simulations with following settings:"
              << "\n\tOutput directory: " << database_output
              << "\n\tInput directory: " << database_input
              << "\n\tRuns: " << n_runs
              << "\n\tDays per run: " << n_days
              << "\n\tP(transmission): " << p_trans
              << "\n\tTransmission scaler mean: " << trans_scaler_mean
              << std::endl;
};

// ------------------------------
void EpidemicSimulator::print_params_to_files()
{
    std::stringstream config_savefile;
    config_savefile << base_dir << "/config.dat";
    std::ofstream configfile_stream(config_savefile.str());
    configfile_stream
        << "\n"
        << "-----general parameters-----"
        << "\n"
        << "transmission probability: " << p_trans
        << "\n"
        << "transmission scalar mean (scales R0): " << trans_scaler_mean
        << "\n"
        << "transmission scalar std (scales R0): " << trans_scaler_std
        << "\n"
        << "population: " << agents.size()
        << "\n"
        << "number of days: " << n_days
        << "\n"
        << "number of runs: " << n_runs
        << "\n"
        << "\n"
        << "-----disease parameters-----"
        << "\n"
        << "incubation period mu: " << incubation_period_mu
        << "\n"
        << "incubation period sigma: " << incubation_period_sig
        << "\n"
        << "latent period (mean), time between infection and onset of infectiousness (days): " << latent_period_mean
        << "\n"
        << "latent period (std), time between infection and onset of infectiousness (days): " << latent_period_std
        << "\n"
        << "days from end of incubation to recovery (upper bound): " << recovery_period_hi
        << "\n"
        << "days from end of incubation to recovery (lower bound): " << recovery_period_low
        << "\n"
        << "days from recovery to before being able to be infected again (mean): " << after_recovery_period_mean
        << "\n"
        << "days from recovery to before being able to be infected again (std): " << after_recovery_period_std
        << "\n"
        << "adult symptomatic fraction: " << adult_symptomatic_fraction
        << "\n"
        << "child symptomatic fraction: " << child_symptomatic_fraction
        << "\n"
        << "asymptomatic infectivity factor: " << asymptomatic_infectivity
        << "\n"
        
        << "\n" 
        << "illness detection probability (per ill individual, per day) - sigmoid_midpoint: " << Detection_sigmoid_midpoint
        << "\n" 
        << "illness detection probability (per ill individual, per day) - sigmoid_v: " << Detection_sigmoid_v
        << "\n" 
        << "illness detection probability (per ill individual, per day) - sigmoid_delay: " << Detection_sigmoid_delay
        << "\n" 
        << "illness detection probability (per ill individual, per day) - sigmoid_base: " << Detection_base
        << "\n"
        << "pre and asymptomatic detection probability (i.e., from testing): " << Detection_asymp

        << "\n"
        << "co-infection rate: " << co_infection_rate
        << "\n"
        << "\n"
        << "-----vaccine parameters-----"
        << "\n"
        << "pharmaceutical intervention (or pre-vac) on? [0=no, 1=yes]: " << pharm_intervention
        << "\n"
        << "number of people pre-vaccinated (priority vaccine): " << n_vac_pre_epidemic_PV
        << "\n"
        << "fraction of people pre-vaccinated over entire population (priority vaccine): " << fraction_vac_pre_epidemic_PV
        << "\n"
        << "number of people vaccinated per week (priority vaccine): " << n_vac_per_week_PV
        << "\n"
        << "number of people pre-vaccinated (general vaccine): " << n_vac_pre_epidemic_GV
        << "\n"
        << "fraction of people pre-vaccinated over entire population (general vaccine): " << fraction_vac_pre_epidemic_GV
        << "\n"
        << "number of people vaccinated per week (general vaccine): " << n_vac_per_week_GV
        << "\n"
        << "probability of vaccination given Age 65+: " << p_vac_65_and_over
        << "\n"
        << "probability of vaccination given Age 18-64: " << p_vac_18_to_64
        << "\n"
        << "probability of vaccination given Age under 18: " << p_vac_children

        << "\n"
        << "immunity against symptomatic infection (after an infection): " << VEc_NR
        << "\n"
        << "immunity against forward infection (after an infection): " << VEt_NR
        << "\n"
        << "immunity against forward infection (after an infection): " << VEt_NR
        << "\n"
        << "daily waning rate of natural immunity against symptomatic infection: " << daily_waning_VEc_NR
        << "\n"
        << "daily waning rate of natural immunity against forward infection: " << daily_waning_VEt_NR

        << "\n"
        << "vaccine efficacy against symptomatic infection (priority vaccine): " << VEc_PV
        << "\n"
        << "vaccine efficacy for infectiousness (priority vaccine): " << VEt_PV
        << "\n"
        << "daily waning rate of priority vaccine efficacy against symptomatic infection: " << daily_waning_VEc_PV
        << "\n"
        << "daily waning rate of priority vaccine efficacy against forward infection: " << daily_waning_VEt_PV

        << "\n"
        << "vaccine efficacy against symptomatic infection (general vaccine): " << VEc_GV
        << "\n"
        << "vaccine efficacy for infectiousness (general vaccine): " << VEt_GV
        << "\n"
        << "daily waning rate of general vaccine efficacy against symptomatic infection: " << daily_waning_VEc_GV
        << "\n"
        << "daily waning rate of general vaccine efficacy against forward infection: " << daily_waning_VEt_GV
        << "\n"
        

        << "\n"
        << "----Social Distancing for COV19----"
        << "\n"
        << "Social distancing on? [0=no, 1=yes]: " << SD_intervention
        << "\n"
        << "cumulative attack rate threshold: " << n_detected_SD_trigger
        << "\n"
        << "SD compliance rate: " << compliance_rate_SD
        << "\n"
        << "duration of social distancing: " << duration_SD
        << "\n"
        << "Factor of social distancing for FoI in community: " << FoI_SD_community
        << "\n"
        << "Factor of social distancing for FoI at work and school: " << FoI_SD_work
        << "\n"
        << "Factor of social distancing for FoI at home: " << FoI_SD_home
        << "\n"
        << "SD assignment period: " << SD_assignment_period

        << "\n" 
        << "Opinion-based compliance? [0=no, 1=yes]: " << opinion_based_compliance
        << "\n" 
        << "Perceived infection rate from contact (initial value): " << beta_perc_initial
        << "\n" 
        << "Perceived infection fatique rate from contact: " << beta_perc_fatigue_rate
        << "\n" 
        << "Opinion threshold: " << opinion_threshold
        << "\n" 
        << "Opinion horizon std: " << opinion_horizon_std
        << "\n"
        << "Opinion horizon mean sigmoid midpoint: " << opinion_horizon_mean_sigmoid_midpoint
        << "\n"  
        << "opinion horizon mean sigmoid v: " << opinion_horizon_mean_sigmoid_v
        << "\n" 
        << "opinion horizon mean sigmoid delay (days): " << opinion_horizon_mean_sigmoid_delay
        << "\n" 
        << "opinion horizon base: " << opinion_horizon_base
        << "\n" 
        << "Opinion horizon weight: " << opinion_horizon_weight
        << "\n" 
        << "Weight of social pressure on an individual's opinion: " << social_pressure_weight
        << "\n" 
        << "Social pressure - Component weight from members sharing a household: " << social_pressure_household
        << "\n" 
        << "Social pressure - Component weight from members sharing a household cluster: " << social_pressure_householdcluster
        << "\n" 
        << "Social pressure - Component weight from members sharing a neighborhood (SA1): " << social_pressure_CD
        << "\n" 
        << "Social pressure - Component weight from members sharing a community (SA2): " << social_pressure_SLA
        << "\n" 
        << "Social pressure - Component weight from members sharing a working group: " << social_pressure_workinggroup
        << "\n" 
        << "Social pressure - Component weight from members sharing a class: "  << social_pressure_class
        << "\n" 
        << "Social pressure - Component weight from members sharing a grade: "  << social_pressure_grade
        << "\n" 
        << "Social pressure - Component weight from members sharing a school: " << social_pressure_school
        << "\n" 
        << "Fraction of population always staying compliant based on their opinion: " << fraction_always_compliant
        << "\n" 
        << "Fraction of population never  staying compliant based on their opinion: " << fraction_never_compliant

        << "\n"
        << "\n"
        << "---School closure for COV19---"
        << "\n"
        << "School closure on? [0=no, 1=yes]: " << School_closure
        << "\n"
        << "cumulative attack rate threshold: " << n_detected_SC_trigger
        << "\n"
        << "duration of school closure: " << duration_SC
        << "\n"
        << "Factor of school closure for FoI in community: " << FoI_SC_community
        << "\n"
        << "Factor of school closure for FoI at work and school: " << FoI_SC_work
        << "\n"
        << "Factor of school closure for FoI at home: " << FoI_SC_home
        << "\n"
        << "\n"
        << "----Case Isolation for COV19----"
        << "\n"
        << "Case Isolation on? [0=no, 1=yes]: " << CI_intervention
        << "\n"
        << "CI compliance rate: " << compliance_rate_CI
        << "\n"
        << "Factor of case isolation for FoI in community: " << FoI_CI_community
        << "\n"
        << "Factor of case isolation for FoI at work and school: " << FoI_CI_work
        << "\n"
        << "Factor of case isolation for FoI at home: " << FoI_CI_home
        << "\n"
        << "\n"
        << "----Home Quarantine for COV19----"
        << "\n"
        << "Home Quarantine on? [0=no, 1=yes]: " << HQ_intervention
        << "\n"
        << "HQ compliance rate: " << compliance_rate_HQ
        << "\n"
        << "Quarantine duration (days): " << duration_HQ
        << "\n"
        << "Factor of home quarantine for FoI in community: " << FoI_HQ_community
        << "\n"
        << "Factor of home quarantine for FoI at work and school: " << FoI_HQ_work
        << "\n"
        << "Factor of home quarantine for FoI at home: " << FoI_HQ_home
        << std::endl;
    configfile_stream.close();

    std::cout
        << "\n"
        << "-----general parameters-----"
        << "\n"
        << "transmission probability: " << p_trans
        << "\n"
        << "transmission scalar mean (scales R0): " << trans_scaler_mean
        << "\n"
        << "transmission scalar std (scales R0): " << trans_scaler_std
        << "\n"
        << "population: " << agents.size()
        << "\n"
        << "number of days: " << n_days
        << "\n"
        << "number of runs: " << n_runs
        << "\n"
        << "\n"
        << "-----disease parameters-----"
        << "\n"
        << "incubation period mu: " << incubation_period_mu
        << "\n"
        << "incubation period sigma: " << incubation_period_sig
        << "\n"
        << "days from end of incubation to recovery (upper bound): " << recovery_period_hi
        << "\n"
        << "days from end of incubation to recovery (lower bound): " << recovery_period_low
        << "\n"
        << "latent period mean, time between infection and onset of infectiousness (days): " << latent_period_mean
        << "\n"
        << "days from recovery to before being able to be infected again (mean): " << after_recovery_period_mean
        << "\n"
        << "days from recovery to before being able to be infected again (std): " << after_recovery_period_std
        << "\n"
        << "adult symptomatic fraction: " << adult_symptomatic_fraction
        << "\n"
        << "child symptomatic fraction: " << child_symptomatic_fraction
        << "\n"
        << "asymptomatic infectivity factor: " << asymptomatic_infectivity
        << "\n"
        << "co_infection rate: " << co_infection_rate
        << "\n" 
        << "illness detection probability (per ill individual, per day) - sigmoid_midpoint: " << Detection_sigmoid_midpoint
        << "\n" 
        << "illness detection probability (per ill individual, per day) - sigmoid_v: " << Detection_sigmoid_v
        << "\n" 
        << "illness detection probability (per ill individual, per day) - sigmoid_delay: " << Detection_sigmoid_delay
        << "\n" 
        << "illness detection probability (per ill individual, per day) - sigmoid_base: " << Detection_base
        << "\n"
        << "pre and asymptomatic detection probability (i.e., from testing): " << Detection_asymp
        << "\n"
        << "\n"
        << "-----vaccine parameters-----"
        << "\n"
        << "pharmaceutical intervention (or pre-vac) on? [0=no, 1=yes]: " << pharm_intervention
        << "\n"
        << "number of people pre-vaccinated (priority vaccine): " << n_vac_pre_epidemic_PV
        << "\n"
        << "fraction of people pre-vaccinated over entire population (priority vaccine): " << fraction_vac_pre_epidemic_PV
        << "\n"
        << "number of people vaccinated per week (priority vaccine): " << n_vac_per_week_PV
        << "\n"
        << "number of people pre-vaccinated (general vaccine): " << n_vac_pre_epidemic_GV
        << "\n"
        << "fraction of people pre-vaccinated over entire population (general vaccine): " << fraction_vac_pre_epidemic_GV
        << "\n"
        << "number of people vaccinated per week (general vaccine): " << n_vac_per_week_GV
        << "\n"
        << "probability of vaccination given Age 65+: " << p_vac_65_and_over
        << "\n"
        << "probability of vaccination given Age 18-64: " << p_vac_18_to_64
        << "\n"
        << "probability of vaccination given Age under 18: " << p_vac_children

        << "\n"
        << "immunity against symptomatic infection (after an infection): " << VEc_NR
        << "\n"
        << "immunity against infectiousness (after an infection): " << VEt_NR
        << "\n"
        << "daily waning rate of natural immunity against symptomatic infection: " << daily_waning_VEc_NR
        << "\n"
        << "daily waning rate of natural immunity against forward infection: " << daily_waning_VEt_NR


        << "\n"
        << "vaccine efficacy against symptomatic infection (priority vaccine): " << VEc_PV
        << "\n"
        << "vaccine efficacy for infectiousness (priority vaccine): " << VEt_PV
        << "\n"
        << "daily waning rate of priority vaccine efficacy against symptomatic infection: " << daily_waning_VEc_PV
        << "\n"
        << "daily waning rate of priority vaccine efficacy against forward infection: " << daily_waning_VEt_PV

        << "\n"
        << "vaccine efficacy against symptomatic infection (general vaccine): " << VEc_GV
        << "\n"
        << "vaccine efficacy for infectiousness (general vaccine): " << VEt_GV
        << "\n"
        << "daily waning rate of general vaccine efficacy against symptomatic infection: " << daily_waning_VEc_GV
        << "\n"
        << "daily waning rate of general vaccine efficacy against forward infection: " << daily_waning_VEt_GV
        << "\n"

        << "\n"
        << "----Social Distancing for COV19----"
        << "\n"
        << "Social distancing on? [0=no, 1=yes]: " << SD_intervention
        << "\n"
        << "cumulative attack rate threshold: " << n_detected_SD_trigger
        << "\n"
        << "Opinion-based compliance? [0=no, 1=yes]: " << opinion_based_compliance
        << "\n" 
        << "Perceived infection rate from contact (initial): " << beta_perc_initial
        << "\n" 
        << "Perceived infection fatique rate from contact: "   << beta_perc_fatigue_rate
        << "\n" 
        << "SD compliance rate: " << compliance_rate_SD
        << "\n" 
        << "duration of social distancing: " << duration_SD
        << "\n" 
        << "Factor of social distancing for FoI in community: " << FoI_SD_community
        << "\n" 
        << "Factor of social distancing for FoI at work and school: " << FoI_SD_work
        << "\n" 
        << "Factor of social distancing for FoI at home: " << FoI_SD_home
        << "\n" 
        << "Assignment period: " << SD_assignment_period
        << "\n" 
        << "Opinion threshold: " << opinion_threshold
        << "\n" 
        << "Opinion horizon mean sigmoid midpoint: " << opinion_horizon_mean_sigmoid_midpoint
        << "\n" 
        << "opinion horizon mean sigmoid v: " << opinion_horizon_mean_sigmoid_v
        << "\n" 
        << "opinion horizon mean sigmoid delay (days): " << opinion_horizon_mean_sigmoid_delay
        << "\n" 
        << "opinion horizon base: " << opinion_horizon_base
        << "\n" 
        << "Opinion horizon (std): " << opinion_horizon_std
        << "\n" 
        << "Opinion horizon weight: " << opinion_horizon_weight
        << "\n" 
        << "Weight of social pressure on an individual's opinion: " << social_pressure_weight
        << "\n" 
        << "Social pressure - Component weight from members sharing a household: " << social_pressure_household
        << "\n" 
        << "Social pressure - Component weight from members sharing a household cluster: " << social_pressure_householdcluster
        << "\n" 
        << "Social pressure - Component weight from members sharing a neighborhood (SA1): " << social_pressure_CD
        << "\n" 
        << "Social pressure - Component weight from members sharing a community (SA2): " << social_pressure_SLA
        << "\n" 
        << "Social pressure - Component weight from members sharing a working group: " << social_pressure_workinggroup
        << "\n" 
        << "Social pressure - Component weight from members sharing a class: "  << social_pressure_class
        << "\n" 
        << "Social pressure - Component weight from members sharing a grade: "  << social_pressure_grade
        << "\n" 
        << "Social pressure - Component weight from members sharing a school: " << social_pressure_school
        << "\n" 
        << "Fraction of population always staying compliant based on their opinion: " << fraction_always_compliant
        << "\n" 
        << "Fraction of population never  staying compliant based on their opinion: " << fraction_never_compliant
        << "\n"
        << "\n"
        << "\n"
        << "---School closure for COV19---"
        << "\n"
        << "School closure on? [0=no, 1=yes]: " << School_closure
        << "\n"
        << "cumulative attack rate threshold: " << n_detected_SC_trigger
        << "\n"
        << "duration of school closure: " << duration_SC
        << "\n"
        << "Factor of school closure for FoI in community: " << FoI_SC_community
        << "\n"
        << "Factor of school closure for FoI at work and school: " << FoI_SC_work
        << "\n"
        << "Factor of school closure for FoI at home: " << FoI_SC_home
        << "\n"
        << "Chance that a parent stays home (per student): " << p_parent_stays_home
        << "\n"
        << "\n"
        << "----Case Isolation for COV19----"
        << "\n"
        << "Case Isolation on? [0=no, 1=yes]: " << CI_intervention
        << "\n"
        << "CI compliance rate: " << compliance_rate_CI
        << "\n"
        << "Factor of case isolation for FoI in community: " << FoI_CI_community
        << "\n"
        << "Factor of case isolation for FoI at work and school: " << FoI_CI_work
        << "\n"
        << "Factor of case isolation for FoI at home: " << FoI_CI_home
        << "\n"
        << "\n"
        << "----Home Quarantine for COV19----"
        << "\n"
        << "Home Quarantine on? [0=no, 1=yes]: " << HQ_intervention
        << "\n"
        << "HQ compliance rate: " << compliance_rate_HQ
        << "\n"
        << "Quarantine duration (days): " << duration_HQ
        << "\n"
        << "Factor of home quarantine for FoI in community: " << FoI_HQ_community
        << "\n"
        << "Factor of home quarantine for FoI at work and school: " << FoI_HQ_work
        << "\n"
        << "Factor of home quarantine for FoI at home: " << FoI_HQ_home
        << std::endl;
};

// ------------------------------
// ------------------------------
// ------------------------------

void EpidemicSimulator::initiate_a_random_seed(unsigned run_id)
{
    unsigned seed = time(NULL) + rnd_num;
    srand(seed);
    std::cout << "Used random number seed: " << seed << std::endl;
    // seed generator for distributed parameters
    generator.seed(seed + 1);
}

// ------------------------------
// ------------------------------
void EpidemicSimulator::reset()
{
    std::cout << "Reseting parameters ..." << std::endl;
    g_detected_incidence = 0;
    g_detected_prevalence = 0;
    g_detected_cumulative = 0;
    g_infected_incidence = 0;
    g_infected_prevalence = 0;
    g_infected_cumulative = 0;
    g_illness_incidence = 0;
    g_illness_prevalence = 0;
    g_illness_cumulative = 0;
    g_number_removed = 0;

    // After parsing the per-cycle parameters, we need to reset the daily parameters.
    daily_waning_VEc_NR = per_cycle_waning_VEc_NR * 2;
    daily_waning_VEt_NR = per_cycle_waning_VEt_NR * 2;
    daily_waning_VEc_PV = per_cycle_waning_VEc_PV * 2;
    daily_waning_VEt_PV = per_cycle_waning_VEt_PV * 2;
    daily_waning_VEc_GV = per_cycle_waning_VEc_GV * 2;
    daily_waning_VEt_GV = per_cycle_waning_VEt_GV * 2;
    

    /* Scale contact rates by `p_trans' for inputting to agents */
    // trans_scaler is now considered to be dependent on virus, thus it is set to be agent-based, not context-based.
    Epidemic::init_transmission_rates(p_trans);

    std::cout << "Creating a new surrogate population ..." << std::endl;
    cout << "Reading population..." << endl;
    CreatePopulation::readPopulation(agents, households, household_clusters, census_districts, destination_zones, statistical_local_areas,
                                     working_groups, schools, grades, classes, database_input, pop_fname);
    cout << "Done reading and creating a surrogate population." << endl;

    // Prepare to simulation with parallelisation
    for (auto &x : agents)
    {
        agent_ids.emplace_back(x.first);
    }
    for (auto &x : households)
    {
        hh_ids.emplace_back(x.first);
    }
    for (auto &x : household_clusters)
    {
        hc_ids.emplace_back(x.first);
    }
    for (auto &x : census_districts)
    {
        cd_ids.emplace_back(x.first);
    }
    for (auto &x : destination_zones)
    {
        dz_ids.emplace_back(x.first);
    }
    for (auto &x : statistical_local_areas)
    {
        sla_ids.emplace_back(x.first);
    }
    for (auto &x : working_groups)
    {
        wg_ids.emplace_back(x.first);
    }
    for (auto &x : schools)
    {
        school_ids.emplace_back(x.first);
    }
    for (auto &x : grades)
    {
        grade_ids.emplace_back(x.first);
    }
    for (auto &x : classes)
    {
        class_ids.emplace_back(x.first);
    }

    // Pre-load the global maps
    stringstream SLA_pop_savefile;
    SLA_pop_savefile << base_dir << "/SLA_pop.dat";
    ofstream pop_dat_stream(SLA_pop_savefile.str());
    for (auto sla_entry : statistical_local_areas)
    {
        g_sla_prevalence_map.emplace(sla_entry.first, 0);
        g_sla_number_removed_map.emplace(sla_entry.first, 0);
        g_sla_incidence_map.emplace(sla_entry.first, 0);
        g_sla_attack_rate_map.emplace(sla_entry.first, 0);
        pop_dat_stream << sla_entry.second->getID() << " " << float(sla_entry.second->getAgentIdxs().size()) << endl;
    }
    pop_dat_stream.close();

    for (AgeGroup ag = AgeGroups::Age0to4; ag <= AgeGroups::Age65Plus; ag = AgeGroup(ag + 1))
    {
        g_ag_sla_prevalence_map.emplace(ag, g_sla_prevalence_map);
        g_ag_sla_number_removed_map.emplace(ag, g_sla_number_removed_map);
        g_ag_sla_incidence_map.emplace(ag, g_sla_incidence_map);

        g_ag_infection_incidence_map.emplace(ag, 0);
        g_ag_infection_prevalence_map.emplace(ag, 0);
        g_ag_illness_incidence_map.emplace(ag, 0);
        g_ag_illness_prevalence_map.emplace(ag, 0);
        g_ag_detection_incidence_map.emplace(ag, 0);
        g_ag_detection_prevalence_map.emplace(ag, 0);
        ag_pops_map.emplace(ag, 0);
    }

    std::stringstream AgeGroup_pop_savefile;
    AgeGroup_pop_savefile << base_dir << "/AgeGroup_pop.dat";
    std::ofstream AG_pop_dat_stream(AgeGroup_pop_savefile.str());

    for (auto agent_entry : agents)
    {
        auto ag = (agent_entry.second)->getAgeGroup();
        ag_pops_map.at(ag)++;
    }

    for (AgeGroup ag = AgeGroups::Age0to4; ag <= AgeGroups::Age65Plus; ag = AgeGroup(ag + 1))
    {
        AG_pop_dat_stream << ag_pops_map.at(ag) << " ";
    }

    AG_pop_dat_stream << endl;
    AG_pop_dat_stream.close();

    // Seeded SLAs are those with international airports
    if (seed_type == SeedTypes::Airport)
    {
        Epidemic::init_airport_seeded_slas(database_input, statistical_local_areas);
    }
    else
    {
        cout << "ERROR: Unknown seeding method was used. Please revise the option. In this release, this code base only supports SeedTypes::Airport." << endl;
        exit(1);
    }

    // Counting number of agents
    n_agents = agents.size();

    // Update the exact number/fraction of vaccinated people from pre-emptive vaccination.
    if (n_vac_pre_epidemic_PV < 0)
    {
        n_vac_pre_epidemic_PV = (int)round(n_agents * fraction_vac_pre_epidemic_PV);
    }
    else
    {
        fraction_vac_pre_epidemic_PV = (float)((int)100 * fraction_vac_pre_epidemic_PV / n_agents) / 100;
    }

    if (n_vac_pre_epidemic_GV < 0)
    {
        n_vac_pre_epidemic_GV = (int)round(n_agents * fraction_vac_pre_epidemic_GV);
    }
    else
    {
        fraction_vac_pre_epidemic_GV = (float)((int)100 * n_vac_pre_epidemic_GV / n_agents) / 100;
    }
};

// ------------------------------
// A utility function for EpidemicSimulator::basedir_generator
int makeDir(const std::string &dir)
{
    int status = mkdir(dir.c_str(), S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
    if (status != 0)
    {
        if (errno == EEXIST)
        {
            std::cout << "Folder " << dir << " already exists." << std::endl;
        }
        else
        {
            std::cerr << "Could not create folder " << dir << " for results (err:" << errno << ")." << std::endl;
        }
    }
    return (status);
}

// ------------------------------
string EpidemicSimulator::basedir_generator()
{
    gettimeofday(&tv, NULL);
    struct tm *tm_info = localtime(&tv.tv_sec);
    strftime(buffer, 26, "%Y-%m-%d_%H-%M-%S", tm_info);

    std::stringstream base_dir_ss;
    base_dir_ss << database_output
                << "D" << buffer
                << "_I" << rnd_num;

    if (seed_type == SeedTypes::Airport)
    {
        base_dir_ss << "_AP";
    }
    else
    {
        std::cerr << "Undefined seeding method: " << seed_type << ". Exiting." << std::endl;
        exit(1);
    }

    std::string base_dir(base_dir_ss.str());

    std::size_t dot_loc;
    while ((dot_loc = base_dir.find(".")) != std::string::npos)
    {
        base_dir.replace(dot_loc, 1, "-");
    }

    if (makeDir(base_dir) != 0)
    {
        std::cerr << "Making directory: " << base_dir << " failed. Exiting." << std::endl;
        exit(1);
    }

    return base_dir;
};

// ------------------------------
// ------------------------------
// ------------------------------

void EpidemicSimulator::rezero_before_simulating()
{
// Rezero the SLA prevalence
#pragma omp parallel for
    for (unsigned i = 0; i < cd_ids.size(); i++)
    {
        census_districts[cd_ids[i]]->remove_n_infecteds_tot();
    }

// Zero g_sla_attack_rate_map for the run
#pragma omp parallel for
    for (unsigned i = 0; i < sla_ids.size(); i++)
    {
        g_sla_attack_rate_map.at(sla_ids[i]) = 0;
    }

    g_detected_cumulative = 0;
    g_illness_cumulative = 0;
    g_infected_cumulative = 0;
}

// ------------------------------
// ------------------------------

void EpidemicSimulator::clearDayTimeInfectionLists()
{

    #pragma omp parallel for
    for (int i = 0; i < dz_ids.size(); i++)
    {
        destination_zones[dz_ids[i]]->clear_infection_lists();
    }

    #pragma omp parallel for
    for (int i = 0; i < wg_ids.size(); i++)
    {
        working_groups[wg_ids[i]]->clear_infection_lists();
    }

    #pragma omp parallel for
    for (int i = 0; i < school_ids.size(); i++)
    {
        schools[school_ids[i]]->clear_infection_lists();
    }

    #pragma omp parallel for
    for (int i = 0; i < grade_ids.size(); i++)
    {
        grades[grade_ids[i]]->clear_infection_lists();
    }

    #pragma omp parallel for
    for (int i = 0; i < class_ids.size(); i++)
    {
        classes[class_ids[i]]->clear_infection_lists();
    }
}

// ------------------------------

void EpidemicSimulator::clearNightTimeInfectionLists()
{
    #pragma omp parallel for
    for (int i = 0; i < hh_ids.size(); i++)
    {
        households[hh_ids[i]]->clear_infection_lists();
    }

    #pragma omp parallel for
    for (int i = 0; i < hc_ids.size(); i++)
    {
        household_clusters[hc_ids[i]]->clear_infection_lists();
    }

    #pragma omp parallel for
    for (int i = 0; i < cd_ids.size(); i++)
    {
        census_districts[cd_ids[i]]->clear_infection_lists();
    }

    #pragma omp parallel for
    for (int i = 0; i < sla_ids.size(); i++)
    {
        statistical_local_areas[sla_ids[i]]->clear_infection_lists();
    }
}

// ------------------------------

void EpidemicSimulator::rezero_everyday()
{
#pragma omp parallel for
    for (int i = 0; i < hh_ids.size(); i++)
    {
        households[hh_ids[i]]->reZero();
    }

#pragma omp parallel for
    for (int i = 0; i < hc_ids.size(); i++)
    {
        household_clusters[hc_ids[i]]->reZero();
    }

#pragma omp parallel for
    for (int i = 0; i < cd_ids.size(); i++)
    {
        census_districts[cd_ids[i]]->reZero();
    }

#pragma omp parallel for
    for (int i = 0; i < dz_ids.size(); i++)
    {
        destination_zones[dz_ids[i]]->reZero();
    }

#pragma omp parallel for
    for (int i = 0; i < wg_ids.size(); i++)
    {
        working_groups[wg_ids[i]]->reZero();
    }

#pragma omp parallel for
    for (int i = 0; i < school_ids.size(); i++)
    {
        schools[school_ids[i]]->reZero();
    }

#pragma omp parallel for
    for (int i = 0; i < grade_ids.size(); i++)
    {
        grades[grade_ids[i]]->reZero();
    }

#pragma omp parallel for
    for (int i = 0; i < class_ids.size(); i++)
    {
        classes[class_ids[i]]->reZero();
    }

#pragma omp parallel for
    for (int i = 0; i < sla_ids.size(); i++)
    {
        statistical_local_areas[sla_ids[i]]->reZero();

        /* Probably should just have these in each statistical area */
        g_sla_prevalence_map.at(sla_ids[i]) = 0;
        g_sla_incidence_map.at(sla_ids[i]) = 0;
        g_sla_number_removed_map.at(sla_ids[i]) = 0;

        for (AgeGroup ag = AgeGroups::Age0to4; ag <= AgeGroups::Age65Plus; ag = AgeGroup(ag + 1))
        {
            g_ag_sla_prevalence_map.at(ag).at(sla_ids[i]) = 0;
            g_ag_sla_incidence_map.at(ag).at(sla_ids[i]) = 0;
            g_ag_sla_number_removed_map.at(ag).at(sla_ids[i]) = 0;
        }
    }

    /* Zero our global variables too */
    g_detected_incidence = 0;
    g_detected_prevalence = 0;

    g_infected_incidence = 0;  // this will count infections
    g_infected_prevalence = 0; // Rezero the prevalence (asymptomatic and symptomatic)

    g_illness_incidence = 0;
    g_illness_prevalence = 0;

    for (AgeGroup ag = AgeGroups::Age0to4; ag <= AgeGroups::Age65Plus; ag = AgeGroup(ag + 1))
    {
        g_ag_infection_incidence_map.at(ag) = 0;
        g_ag_illness_incidence_map.at(ag) = 0;
        g_ag_detection_incidence_map.at(ag) = 0;
        g_ag_infection_prevalence_map.at(ag) = 0;
        g_ag_illness_prevalence_map.at(ag) = 0;
        g_ag_detection_prevalence_map.at(ag) = 0;
    }

    g_number_removed = 0;

    // Empty infection lists.
    this->clearDayTimeInfectionLists();
    this->clearNightTimeInfectionLists();
}

// ------------------------------

void EpidemicSimulator::simulate_all_runs()
{
    for (unsigned run = 0; run < n_runs; ++run)
    {
        double tStart = omp_get_wtime();
        initiate_bookkeeping_everyday(run);
        simulate_a_single_run(run);
        close_bookkeeping_everyday();
        std::cout << "\nTime taken: " << (double)(omp_get_wtime() - tStart) << " seconds.\n"
                  << std::endl;
    }
}

// ------------------------------
// ------------------------------
// ------------------------------

void EpidemicSimulator::reset_run(unsigned run_id, std::vector<bool> &list_SD_compliant)
{
    initiate_a_random_seed(run_id);
    rezero_before_simulating();

    std::cout << "Initialising all agents..." << std::endl;
    #pragma omp parallel for
    for (unsigned long i = 0; i < agent_ids.size(); i++)
    {
        agents[agent_ids[i]]->init();
    }

    std::cout << "Saving the list of agents who are compliant to SD measures ..." << std::endl;
    list_SD_compliant.reserve(n_agents + 1);
    for (unsigned agent_id = 0; agent_id < n_agents; agent_id++)
    {
        list_SD_compliant.emplace_back(false);
    }
    std::cout << "Done saving the list of SD-compliant agents." << std::endl;

    unsigned long n_never_compliant_count = 0;
    unsigned long n_never_compliant = fraction_never_compliant * n_agents;
    unsigned long n_always_compliant_count = 0;
    unsigned long n_always_compliant = fraction_always_compliant * n_agents;
    unsigned long n_opinion_compliant_count = 0;

    agents_character.clear();
    for (int agent_id=0; agent_id < n_agents; agent_id++) {
        std::vector<float> list_rand_weights = {(float) fraction_never_compliant, (float) fraction_always_compliant, (float) (1.0 - fraction_never_compliant - fraction_always_compliant)};
        std::mt19937 gen(rand_r(&seed_thread[omp_get_thread_num()]) + time(NULL));
        std::discrete_distribution<> dist(list_rand_weights.begin(), list_rand_weights.end());

        unsigned int cat = dist(gen);
        agents_character.emplace_back(cat);
        if (cat == 0) {
            n_never_compliant_count++;
            agents[agent_id]->set_rational_opinion_on_compliance(0.0);
        } else if (cat == 1) {
            n_always_compliant_count++;
            agents[agent_id]->set_rational_opinion_on_compliance(1.0);
        } else {
            n_opinion_compliant_count++;
            agents[agent_id]->set_rational_opinion_on_compliance(0.0);
        }
    }

    if (opinion_based_compliance) {
        std::cout << "******* OPINION CHARACTER ASSIGNMENT: " << n_never_compliant_count << " agents never complying to SD, " << n_always_compliant_count << " agents always complying to SD, " << n_opinion_compliant_count << " agents whose SD compliance is based on their opinion." << std::endl;
    }

    if (SD_intervention)
    {
        assign_SD_compliance();
        std::cout << "Assigning social distancing compliance to agents ..." << std::endl;
        
        // Save list of compliant/noncompliant agents
        #pragma omp parallel for
        for (unsigned long i = 0; i < agent_ids.size(); i++)
        {
            if (agents[agent_ids[i]]->getSDState() == SD_States::SD_compliant)
            {
                list_SD_compliant[agent_ids[i]] = true;
            }
            else
            {
                list_SD_compliant[agent_ids[i]] = false;
            }
        }
    }

    if (CI_intervention)
    {
        assign_CI_compliance();
        std::cout << "assigning case isolation compliance to agents at rate: " << compliance_rate_CI << std::endl;
        if (HQ_intervention)
        {
            assign_HQ_compliance();
            std::cout << "done assigning CI compliance, now assigning HQ compliance at rate: " << compliance_rate_HQ << std::endl;
        }
    }

    if (School_closure)
    {
        assign_SC_compliance_students_teachers();
        std::cout << "Assigning school closure compliance to students and teachers (parents comply dynamically)" << std::endl;
    }

    if (HQ_intervention && (!CI_intervention))
    {
        std::cout << "WARNING: Home quarantine has not been implemented, it is conditional on Case Isolation" << std::endl;
    }

    set_Compliance_states_all();

    num_vac = 0;
    num_vac_65_plus = 0;
    num_vac_18_to_64 = 0;
    num_vac_children = 0;

    num_vac_PV = 0;
    num_vac_65_plus_PV = 0;
    num_vac_18_to_64_PV = 0;
    num_vac_children_PV = 0;

    num_vac_GV = 0;
    num_vac_65_plus_GV = 0;
    num_vac_18_to_64_GV = 0;
    num_vac_children_GV = 0;

    n_airport_infections = 0;

    n_recovered = 0;
}

// ------------------------------

void EpidemicSimulator::simulate_a_single_run(unsigned run_id)
{
    cout << "\n\n#-----------------------#\n\tRun #" << run_id << "\n#-----------------------#\n"
         << endl;

    bool SD_trigger = false;
    int SD_timer = 0;
    int days_since_last_SD_assignment = 0;

    bool SC_trigger = false;
    int SC_timer = 0;

    std::vector<bool> list_SD_compliant;
    reset_run(run_id, list_SD_compliant);

    // make lists of unvaccinated agents in each age bracket
    std::vector<unsigned long> unvaccinated_child_IDs;
    std::vector<unsigned long> unvaccinated_adult_IDs;
    std::vector<unsigned long> unvaccinated_senior_IDs;
    std::stack<unsigned long> unvaccinated_child_IDs_s;
    std::stack<unsigned long> unvaccinated_adult_IDs_s;
    std::stack<unsigned long> unvaccinated_senior_IDs_s;

    if (pharm_intervention)
    {
        for (auto agent_entry : agents)
        {
            auto c_agent = agent_entry.second;
            if (c_agent->getAgeGroup() == AgeGroups::Age65Plus)
            {
                unvaccinated_senior_IDs.push_back(c_agent->getID());
            }
            else if (c_agent->getAgeGroup() == AgeGroups::Age0to4 || c_agent->getAgeGroup() == AgeGroups::Age5to18)
            {
                unvaccinated_child_IDs.push_back(c_agent->getID());
            }
            else
            {
                unvaccinated_adult_IDs.push_back(c_agent->getID());
            }
        }

        unsigned shuf_seed = 0;
        while (shuf_seed == 0)
        {
            shuf_seed = rand();
        }

        std::shuffle(unvaccinated_senior_IDs.begin(), unvaccinated_senior_IDs.end(), std::default_random_engine(shuf_seed));
        std::shuffle(unvaccinated_adult_IDs.begin(), unvaccinated_adult_IDs.end(), std::default_random_engine(shuf_seed + (unsigned)rand()));
        std::shuffle(unvaccinated_child_IDs.begin(), unvaccinated_child_IDs.end(), std::default_random_engine(shuf_seed + (unsigned)rand()));

        for (int i = 0; i < unvaccinated_senior_IDs.size(); i++)
        {
            unvaccinated_senior_IDs_s.push(unvaccinated_senior_IDs.at(i));
        }
        for (int i = 0; i < unvaccinated_adult_IDs.size(); i++)
        {
            unvaccinated_adult_IDs_s.push(unvaccinated_adult_IDs.at(i));
        }
        for (int i = 0; i < unvaccinated_child_IDs.size(); i++)
        {
            unvaccinated_child_IDs_s.push(unvaccinated_child_IDs.at(i));
        }

        std::cout << "\nBefore pre-emptive vaccination:" << std::endl;
        std::cout << "\t- There are " << unvaccinated_senior_IDs_s.size() << " unvaccinated agents 65+" << std::endl;
        std::cout << "\t- There are " << unvaccinated_adult_IDs_s.size() << " unvaccinated agents 18-64" << std::endl;
        std::cout << "\t- There are " << unvaccinated_child_IDs_s.size() << " unvaccinated agents under 18" << std::endl;

        std::cout << "\nAfter pre-emptive vaccination:" << std::endl;
        pre_epidemic_vaccination(unvaccinated_senior_IDs_s,
                                 unvaccinated_adult_IDs_s,
                                 unvaccinated_child_IDs_s,
                                 0);
    }

    g_infected_prevalence = getNumAgentsInfected();
    g_illness_prevalence = getNumAgentsIll();

    std::cout << "\nStarting simulation (max days=" << n_days
              << " and " << g_infected_prevalence << "/" << agents.size() << " agents already infected, "
              << g_illness_prevalence << "/" << agents.size() << " agents already symptomatic.) ..." << std::endl;

    // initialize the weekday counter
    unsigned day_of_week = rand() % 7;
    bool weekend = false;

    detected_incidence_buffer.clear();
    SD_compliance_level = 0;
    for (unsigned day = 0; day < n_days && (day < 10 || (g_infected_prevalence > 0 && g_number_removed < agents.size())); ++day)
    {
        std::cout << "\n\t=== Start of day " << day << " ===" << std::endl;
        
        // We will first calculate the detection rate for symptomatic case for today.
        Detection_ = 2.0 * Detection_sigmoid_midpoint / (1 + pow(M_E, (-1.0 * Detection_sigmoid_v * (day - Detection_sigmoid_delay)))) + Detection_base;
        std::cout << "\t Detection Rate = " << Detection_ << ", which is calculated from the sigmoid function with sigmoid_midpoint = " << Detection_sigmoid_midpoint << ", sigmoid_v = " << Detection_sigmoid_v << ", sigmoid_delay = " << Detection_sigmoid_delay << ", and sigmoid_base = " << Detection_base << "\n" << std::endl;

        // Re-assign SD commpliance for all agents
        if (SD_trigger) {
            if (opinion_based_compliance) {
                std::cout << "\t ***** SD level on day " << day << " is assigned based on agents' opinion." << std::endl;
                std::cout << "\t ***** Fatigue rate on day " << day << " is " << daily_fatigue_rate[day] << std::endl;
                SD_compliance_level = assign_SD_compliance_opinion_Qmod(days_since_last_SD_assignment, detected_incidence_buffer, list_SD_compliant, daily_fatigue_rate[day], day);
            } else {
                std::cout << "\t***** SD level on day " << day << ": " << daily_SD_level[day] << std::endl;
                if (daily_SD_level[day] != daily_SD_level[day - 1]) {
                    days_since_last_SD_assignment = 0;
                }
                assign_SD_compliance_Qmod(daily_SD_level[day], days_since_last_SD_assignment, list_SD_compliant);
            }
            
            set_Compliance_states_all_Qmod();
            days_since_last_SD_assignment = (days_since_last_SD_assignment + 1) % SD_assignment_period;
        }

        // Follow main_R0 style in assigning weekdays/weekends
        if (day_of_week == 5 || day_of_week == 6)
        {
            weekend = true;
            std::cout << "\tIt's the weekend" << std::endl;
        }
        else
        {
            weekend = false;
            std::cout << "\tIt's a weekday" << std::endl;
        }

        // Zero the buckets: households, household_clusters, census_districts, destination_zones, working_groups, schools, grades, classes, statistical_local_areas
        rezero_everyday();

        if ((seed_type == SeedTypes::Airport) && ((g_detected_cumulative < 20) || !SD_intervention))
        {
            // Seed at SLAs around international airports
            std::vector<unsigned long> infected_agent_ids;
            std::cout << "\n\tSeeding via international airports ..." << std::endl;

            // Make the lists of variants and weights to sample from.
            std::vector<Virus> variant_list;
            std::vector<float> variant_weight_list;

            // For this simple setup, assume that only one variant is spreading. Its characteristic is defined by the old setup with global variables.
            // An extension is easily implemented with the extension of the arrays of variant info and weights.
            Virus ancestral(
                trans_scaler_mean, trans_scaler_std,                    // trans scaler normal distribution
                latent_period_mean, latent_period_mean,                 // latent_period normal distribution
                incubation_period_mu, incubation_period_sig,            // incubation period lognormal distribution
                recovery_period_hi, recovery_period_low,                // recovery period uniform distribution
                after_recovery_period_mean, after_recovery_period_std,  // the period after recovery within that the agent cannot be infected again
                adult_symptomatic_fraction,                             // adult_symptomatic_fraction
                child_symptomatic_fraction,                             // child_symptomatic_fraction
                asymptomatic_infectivity,                               // asymptomatic_infectivity,
                0);
            float weight_of_ancestral = 1;

            variant_list.emplace_back(ancestral);
            variant_weight_list.emplace_back(weight_of_ancestral);

            // Seeding with variant information and sampling weights.
            Epidemic::seed_epi_airports(agents, infected_agent_ids, variant_list, variant_weight_list, day, VEc_NR, VEt_NR, daily_waning_VEc_NR, daily_waning_VEt_NR);

            for (auto &agent_id : infected_agent_ids)
            {
                auto c_agent = agents.at(agent_id);
                n_airport_infections++;
                g_infected_incidence++;
                g_ag_infection_incidence_map.at(c_agent->getAgeGroup())++;
            }
            std::cout << "\tInfected " << infected_agent_ids.size() << " agents in SLAs near airports/urban centres." << std::endl;
        }

        if (SD_timer == 1)
        {
            std::cout << "\tSD trigger reached." << std::endl;
        }
        if (SC_timer == 1)
        {
            std::cout << "\tSC trigger reached. Children, Teachers, and some parents will stay home." << std::endl;
        }
        if (SD_trigger)
        {
            SD_timer += 1;
            std::cout << "\tSD applied, timer: " << SD_timer << std::endl;
        }
        if (SC_trigger)
        {
            assign_SC_compliance_parents();
            SC_timer += 1;
            std::cout << "\tSC applied, timer: " << SC_timer << std::endl;
        }

        if (!weekend)
        {
            std::cout << "\tIt's not a weekend so we're simulating daytime then nighttime" << std::endl;
            // Update probabilities and infect agents during the first part of the day (at work/school)
            if (SD_intervention || CI_intervention || School_closure)
            {
                std::cout << "\n\tUpdating infection and detection status ..." << std::endl;
                step_and_detect(day);
                std::cout << "\tApplying/updating fraction of interaction for daytime ..." << std::endl;
                update_Isolation_states(SD_trigger, CI_intervention, HQ_intervention, SC_trigger);
                set_Isolation_Factors_work(FoI_SD_work, FoI_CI_work, FoI_HQ_work, FoI_SC_work);
                std::cout << "\tSimulating agents during the day period by calling simulateDaytime_SD ..." << std::endl;
                simulateDaytime_SD(day);

                std::cout << "\n\tUpdating infection status ..." << std::endl;
                step_without_detection(day);
                std::cout << "\tApplying/updating fraction of interaction for nighttime ..." << std::endl;
                update_Isolation_states(SD_trigger, CI_intervention, HQ_intervention, SC_trigger);
                set_Isolation_Factors_home(FoI_SD_home, FoI_CI_home, FoI_HQ_home, FoI_SC_home);
                set_Isolation_Factors_community(FoI_SD_community, FoI_CI_community, FoI_HQ_community, FoI_SC_community);
                std::cout << "\tSimulating agents during the night period by calling simulateNighttime_SD ..." << std::endl;
                simulateNighttime_SD(day);
            }
            else
            {
                std::cout << "... simulating agents during the day period by calling ..." << std::endl;
                simulateDaytime(day);
                std::cout << "... simulating agents during the night period by calling simulateNighttime ..." << std::endl;
                simulateNighttime(day);
            }
        }
        else
        {
            std::cout << "\n\tIt is a weekend so we're essentially simulating nighttime twice" << std::endl;

            if (SD_intervention || CI_intervention || School_closure)
            {
                std::cout << "\n\tUpdating infection and detection status ..." << std::endl;
                step_and_detect(day);
                std::cout << "\tApplying/updating fraction of interaction for daytime (weekend) ..." << std::endl;
                update_Isolation_states(SD_trigger, CI_intervention, HQ_intervention, SC_trigger);
                set_Isolation_Factors_home(FoI_SD_home, FoI_CI_home, FoI_HQ_home, FoI_SC_home);
                set_Isolation_Factors_community(FoI_SD_community, FoI_CI_community, FoI_HQ_community, FoI_SC_community);
                std::cout << "\tSimulating agents during the daytime period (weekend) by calling simulateDaytimeWeekend_SD ..." << std::endl;
                simulateDaytimeWeekend_SD(day);

                std::cout << "\n\tUpdating infection status ..." << std::endl;
                step_without_detection(day);
                std::cout << "\tApplying/updating fraction of interaction for nighttime (weekend) ..." << std::endl;
                update_Isolation_states(SD_trigger, CI_intervention, HQ_intervention, SC_trigger);
                set_Isolation_Factors_home(FoI_SD_home, FoI_CI_home, FoI_HQ_home, FoI_SC_home);
                set_Isolation_Factors_community(FoI_SD_community, FoI_CI_community, FoI_HQ_community, FoI_SC_community);
                std::cout << "\tSimulating agents during the night period (weekend) by calling simulateNighttime_SD ..." << std::endl;
                simulateNighttime_SD(day);
            }
            else
            {
                std::cout << "... simulating agents during the day (weekend) period by calling simulateDaytimeWeekend ..." << std::endl;
                simulateDaytimeWeekend(day);
                std::cout << "... simulating agents during the night period by calling simulateNighttime ..." << std::endl;
                simulateNighttime(day);
            }
        }

        if (CI_intervention && HQ_intervention)
        {
            std::cout << "\tUpdating home quarantine clocks" << std::endl;
            update_HQ(); // advances clock for quarantined agents. turns quarantine off if duration is exceeded
        }

        // update detected_incidence_buffer for opinion dynamics
        detected_incidence_buffer.emplace_back(g_detected_incidence);

        g_detected_cumulative += g_detected_incidence;
        g_illness_cumulative += g_illness_incidence;
        g_infected_cumulative += g_infected_incidence;

        // Check whether to enable NPIs
        if ((g_detected_cumulative >= n_detected_SD_trigger) && !SD_trigger && SD_intervention && (SD_timer < duration_SD + 1))
        {
            SD_trigger = true;
            SD_timer = 1;
        }

        if ((g_detected_cumulative >= n_detected_SC_trigger) && !SC_trigger && School_closure && (SC_timer < duration_SC + 1))
        {
            SC_trigger = true;
            SC_timer = 1;
        }

        if (SD_timer == duration_SD + 1)
        {
            SD_trigger = false;
            std::cout << "SD duration exceeded, stopping social distancing" << std::endl;
        }

        if (SC_timer == duration_SC + 1)
        {
            SC_trigger = false;
            std::cout << "SC duration exceeded, re-opening schools" << std::endl;
        }

        // Bookkeeping
        bookkeeping_everyday(run_id, day);
        
        // vaccination
        if (pharm_intervention)
        {
            progressive_vaccination_priority(unvaccinated_senior_IDs_s, unvaccinated_adult_IDs_s, unvaccinated_child_IDs_s, day);
            progressive_vaccination_general(unvaccinated_senior_IDs_s, unvaccinated_adult_IDs_s, unvaccinated_child_IDs_s, day);
        }

        // Follow main_R0 in assigning weekdays/weekends
        day_of_week += 1;
        if (day_of_week >= 7)
        {
            day_of_week = 0;
        }
    }

    g_infected_prevalence = getNumAgentsInfected();
    g_illness_prevalence = getNumAgentsIll();

    std::cout << "\nEnd of Run " << run_id
              << " and " << g_infected_prevalence << "/" << agents.size() << " agents already infected, "
              << g_illness_prevalence << "/" << agents.size() << " agents already symptomatic." << std::endl;
};


// parse input parameter file
void EpidemicSimulator::parse_epidemic_params()
{
    std::ifstream param_stream(virus_parameter_input_file);
    if (!param_stream.good())
    {
        std::cout << "Parameter input file named: " << virus_parameter_input_file << " does not exist, using default parameters" << std::endl;
    }
    else
    {
        std::string cline;
        while (std::getline(param_stream, cline))
        {
            std::stringstream i_ss;
            i_ss << cline;
            std::string first_element;
            i_ss >> first_element;
            if (first_element == "trans_scaler_mean:")
            {
                i_ss >> trans_scaler_mean;
            }
            if (first_element == "trans_scaler_std:")
            {
                i_ss >> trans_scaler_std;
            }
            else if (first_element == "incubation_period_mu:")
            {
                i_ss >> incubation_period_mu;
            }
            else if (first_element == "incubation_period_sig:")
            {
                i_ss >> incubation_period_sig;
            }
            else if (first_element == "recovery_period_hi:")
            {
                i_ss >> recovery_period_hi;
            }
            else if (first_element == "recovery_period_low:")
            {
                i_ss >> recovery_period_low;
            }
            else if (first_element == "latent_period_mean:")
            {
                i_ss >> latent_period_mean;
            }
            else if (first_element == "latent_period_std:")
            {
                i_ss >> latent_period_std;
            }
            else if (first_element == "after_recovery_period_mean:")
            {
                i_ss >> after_recovery_period_mean;
            }
            else if (first_element == "after_recovery_period_std:")
            {
                i_ss >> after_recovery_period_std;
            }
            else if (first_element == "adult_symptomatic_fraction:")
            {
                i_ss >> adult_symptomatic_fraction;
            }
            else if (first_element == "child_symptomatic_fraction:")
            {
                i_ss >> child_symptomatic_fraction;
            }
            else if (first_element == "asymptomatic_infectivity:")
            {
                i_ss >> asymptomatic_infectivity;
            }
            
            if (first_element == "Detection_sigmoid_midpoint:")
            {
                i_ss >> Detection_sigmoid_midpoint;
            }
            if (first_element == "Detection_sigmoid_v:")
            {
                i_ss >> Detection_sigmoid_v;
            }
            if (first_element == "Detection_sigmoid_delay:")
            {
                i_ss >> Detection_sigmoid_delay;
            }
            if (first_element == "Detection_base:")
            {
                i_ss >> Detection_base;
            }

            else if (first_element == "Detection_asymp:")
            {
                i_ss >> Detection_asymp;
            }
            else if (first_element == "co_infection_rate:")
            {
                i_ss >> co_infection_rate;
            }
        }
        param_stream.close();
    }
};

void EpidemicSimulator::parse_pharm_intervention_params()
{
    ifstream pharm_intervention_stream(pharm_intervention_input_file);
    if (!pharm_intervention_stream.good())
    {
        cout << "File " << pharm_intervention_input_file << " does not exist." << endl;
    }
    else
    {
        string cline;
        unsigned long c_sla;

        while (getline(pharm_intervention_stream, cline))
        {
            stringstream i_ss;
            i_ss << cline;
            string first_element;
            i_ss >> first_element;
            if (first_element == "pharm_intervention:")
            {
                i_ss >> pharm_intervention;
            }
            else if (first_element == "n_vac_per_week_PV:")
            {
                i_ss >> n_vac_per_week_PV;
            }
            else if (first_element == "n_vac_pre_epidemic_PV:")
            {
                i_ss >> n_vac_pre_epidemic_PV;
            }
            else if (first_element == "fraction_vac_pre_epidemic_PV:")
            {
                i_ss >> fraction_vac_pre_epidemic_PV;
            }
            else if (first_element == "n_vac_per_week_GV:")
            {
                i_ss >> n_vac_per_week_GV;
            }
            else if (first_element == "n_vac_pre_epidemic_GV:")
            {
                i_ss >> n_vac_pre_epidemic_GV;
            }
            else if (first_element == "fraction_vac_pre_epidemic_GV:")
            {
                i_ss >> fraction_vac_pre_epidemic_GV;
            }
            else if (first_element == "p_vac_65_and_over:")
            {
                i_ss >> p_vac_65_and_over;
            }
            else if (first_element == "p_vac_18_to_64:")
            {
                i_ss >> p_vac_18_to_64;
            }
            else if (first_element == "p_vac_children:")
            {
                i_ss >> p_vac_children;
            }
            else if (first_element == "VEc_NR:")
            {
                i_ss >> VEc_NR;
            }
            else if (first_element == "VEt_NR:")
            {
                i_ss >> VEt_NR;
            }
            else if (first_element == "per_cycle_waning_VEc_NR:")
            {
                i_ss >> per_cycle_waning_VEc_NR;
            }
            else if (first_element == "per_cycle_waning_VEt_NR:")
            {
                i_ss >> per_cycle_waning_VEt_NR;
            }
            else if (first_element == "VEc_PV:")
            {
                i_ss >> VEc_PV;
            }
            else if (first_element == "VEt_PV:")
            {
                i_ss >> VEt_PV;
            }
            else if (first_element == "per_cycle_waning_VEc_PV:")
            {
                i_ss >> per_cycle_waning_VEc_PV;
            }
            else if (first_element == "per_cycle_waning_VEt_PV:")
            {
                i_ss >> per_cycle_waning_VEt_PV;
            }
            else if (first_element == "VEc_GV:")
            {
                i_ss >> VEc_GV;
            }
            else if (first_element == "VEt_GV:")
            {
                i_ss >> VEt_GV;
            }
            else if (first_element == "per_cycle_waning_VEc_GV:")
            {
                i_ss >> per_cycle_waning_VEc_GV;
            }
            else if (first_element == "per_cycle_waning_VEt_GV:")
            {
                i_ss >> per_cycle_waning_VEt_GV;
            }
        }

        pharm_intervention_stream.close();
    }
};

// ------------------------------
// ------------------------------

void EpidemicSimulator::parse_non_pharm_intervention_params()
{
    std::string non_pharm_intervention_input_file = database_input + non_pharm_intervention_input_fname;
    std::ifstream non_pharm_intervention_file(non_pharm_intervention_input_file);
    if (!non_pharm_intervention_file.good())
    {
        std::cout << "File non_pharm_intervention_input.dat named: " << non_pharm_intervention_input_file << " does not exist, using default parameters" << std::endl;
    }
    else
    {
        std::string param_file(non_pharm_intervention_input_file);
        std::ifstream param_stream(param_file);
        std::string cline;

        while (std::getline(param_stream, cline))
        {
            std::stringstream i_ss;
            i_ss << cline;
            std::string first_element;
            i_ss >> first_element;

            if (first_element == "SD_intervention:")
            {
                i_ss >> SD_intervention;
            }
            else if (first_element == "n_detected_SD_trigger:")
            {
                i_ss >> n_detected_SD_trigger;
            }
            else if (first_element == "opinion_based_compliance:")
            {
                i_ss >> opinion_based_compliance;
            }
            else if (first_element == "beta_perc_initial:")
            {
                i_ss >> beta_perc_initial;
            }
            else if (first_element == "beta_perc_fatigue_rate:")
            {
                i_ss >> beta_perc_fatigue_rate;
            } 
            else if (first_element == "opinion_threshold:")
            {
                i_ss >> opinion_threshold;
            }
            else if (first_element == "opinion_horizon_mean_sigmoid_midpoint:")
            {
                i_ss >> opinion_horizon_mean_sigmoid_midpoint;
            }
            else if (first_element == "opinion_horizon_mean_sigmoid_v:")
            {
                i_ss >> opinion_horizon_mean_sigmoid_v;
            }
            else if (first_element == "opinion_horizon_mean_sigmoid_delay:")
            {
                i_ss >> opinion_horizon_mean_sigmoid_delay;
            }
            else if (first_element == "opinion_horizon_base:")
            {
                i_ss >> opinion_horizon_base;
            }
            else if (first_element == "opinion_horizon_std:")
            {
                i_ss >> opinion_horizon_std;
            }
            else if (first_element == "opinion_horizon_weight:")
            {
                i_ss >> opinion_horizon_weight;
            }
            else if (first_element == "social_pressure_weight:")
            {
                i_ss >> social_pressure_weight;
            }
            else if (first_element == "social_pressure_household:")
            {
                i_ss >> social_pressure_household;
            }
            else if (first_element == "social_pressure_householdcluster:")
            {
                i_ss >> social_pressure_householdcluster;
            }
            else if (first_element == "social_pressure_SLA:")
            {
                i_ss >> social_pressure_SLA;
            }
            else if (first_element == "social_pressure_CD:")
            {
                i_ss >> social_pressure_CD;
            }
            else if (first_element == "social_pressure_workinggroup:")
            {
                i_ss >> social_pressure_workinggroup;
            }
            else if (first_element == "social_pressure_class:")
            {
                i_ss >> social_pressure_class;
            }
            else if (first_element == "social_pressure_grade:")
            {
                i_ss >> social_pressure_grade;
            }
            else if (first_element == "social_pressure_school:")
            {
                i_ss >> social_pressure_school;
            }
            else if (first_element == "fraction_always_compliant:")
            {
                i_ss >> fraction_always_compliant;
            }
            else if (first_element == "fraction_never_compliant:")
            {
                i_ss >> fraction_never_compliant;
            }
            else if (first_element == "compliance_rate_SD:")
            {
                i_ss >> compliance_rate_SD;
            }
            else if (first_element == "duration_SD:")
            {
                i_ss >> duration_SD;
            }
            else if (first_element == "FoI_SD_community:")
            {
                i_ss >> FoI_SD_community;
            }
            else if (first_element == "FoI_SD_work:")
            {
                i_ss >> FoI_SD_work;
            }
            else if (first_element == "FoI_SD_home:")
            {
                i_ss >> FoI_SD_home;
            }
            else if (first_element == "SD_assignment_period:")
            {
                i_ss >> SD_assignment_period;
            }
            else if (first_element == "School_closure:")
            {
                i_ss >> School_closure;
            }
            else if (first_element == "n_detected_SC_trigger:")
            {
                i_ss >> n_detected_SC_trigger;
            }
            else if (first_element == "duration_SC:")
            {
                i_ss >> duration_SC;
            }
            else if (first_element == "FoI_SC_community:")
            {
                i_ss >> FoI_SC_community;
            }
            else if (first_element == "FoI_SC_work:")
            {
                i_ss >> FoI_SC_work;
            }
            else if (first_element == "FoI_SC_home:")
            {
                i_ss >> FoI_SC_home;
            }
            else if (first_element == "p_parent_stays_home:")
            {
                i_ss >> p_parent_stays_home;
            }
            else if (first_element == "CI_intervention:")
            {
                i_ss >> CI_intervention;
            }
            else if (first_element == "compliance_rate_CI:")
            {
                i_ss >> compliance_rate_CI;
            }
            else if (first_element == "FoI_CI_community:")
            {
                i_ss >> FoI_CI_community;
            }
            else if (first_element == "FoI_CI_work:")
            {
                i_ss >> FoI_CI_work;
            }
            else if (first_element == "FoI_CI_home:")
            {
                i_ss >> FoI_CI_home;
            }
            else if (first_element == "HQ_intervention:")
            {
                i_ss >> HQ_intervention;
            }
            else if (first_element == "compliance_rate_HQ:")
            {
                i_ss >> compliance_rate_HQ;
            }
            else if (first_element == "duration_HQ:")
            {
                i_ss >> duration_HQ;
            }
            else if (first_element == "FoI_HQ_community:")
            {
                i_ss >> FoI_HQ_community;
            }
            else if (first_element == "FoI_HQ_work:")
            {
                i_ss >> FoI_HQ_work;
            }
            else if (first_element == "FoI_HQ_home:")
            {
                i_ss >> FoI_HQ_home;
            }
        }
        param_stream.close();
    }
}

// ------------------------------
// ------------------------------

void EpidemicSimulator::parse_daily_SD_level()
{
    // Initially SD level is initialised at compliance_rate_SD for all days
    for (unsigned day = 0; day <= n_days; day++)
    {
        daily_SD_level.emplace_back(compliance_rate_SD);
    }

    std::string daily_SD_level_input_file = database_input + daily_SD_level_input_fname;
    std::ifstream daily_SD_level_file(daily_SD_level_input_file);
    if (!daily_SD_level_file.good())
    {
        std::cout << "File daily_SD_level.dat named: " << daily_SD_level_input_file << " does not exist, using default parameters" << std::endl;
    }
    else
    {
        std::cout << "-- Loading daily SD levels from : " << daily_SD_level_input_file << std::endl;

        std::string param_file(daily_SD_level_input_file);
        std::ifstream param_stream(param_file);
        std::string cline;
        while (std::getline(param_stream, cline))
        {
            std::stringstream i_ss;
            i_ss << cline;

            int day;
            float SD_level;
            i_ss >> day >> SD_level;

            for (int day_i = day; day_i <= n_days; day_i++)
            {
                daily_SD_level[day_i] = SD_level;
            }
        }

        // Print out for checking later
        std::cout << "-- SD levels on daily basis : " << std::endl;
        for (unsigned day = 0; day <= n_days; day++)
        {
            if (day == 0) {
                std::cout << "Day " << day << ": " << daily_SD_level[day] << std::endl;
            } else if (daily_SD_level[day] != daily_SD_level[day-1]) {
                std::cout << "Day " << day << ": " << daily_SD_level[day] << std::endl;
            }
        }
        std::cout << std::endl;
    }
}

// ------------------------------

void EpidemicSimulator::parse_daily_fatigue_rate()
{
    for (unsigned day = 0; day <= n_days; day++)
    {
        daily_fatigue_rate.emplace_back(beta_perc_fatigue_rate);
    }

    std::string daily_fatigue_rate_input_file = database_input + daily_fatigue_rate_input_fname;
    std::ifstream daily_fatigue_rate_file(daily_fatigue_rate_input_file);
    if (!daily_fatigue_rate_file.good()) {
        std::cout << "File daily_fatigue_rate.dat named: " << daily_fatigue_rate_input_file << " does not exist, using default parameters" << std::endl;
    } else {
        std::cout << "\n-- Loading daily fatigue rates from : " << daily_fatigue_rate_input_file << std::endl;

        std::string param_file(daily_fatigue_rate_input_file);
        std::ifstream param_stream(param_file);
        std::string cline;
        while (std::getline(param_stream, cline))
        {
            std::stringstream i_ss;
            i_ss << cline;

            int day;
            float fatigue_rate;
            i_ss >> day >> fatigue_rate;

            for (int day_i = day; day_i <= n_days; day_i++)
            {
                daily_fatigue_rate[day_i] = fatigue_rate;
            }
        }

        // Print out for checking later
        std::cout << "-- Fatigue rate on daily basis : " << std::endl;
        for (unsigned day = 0; day <= n_days; day++)
        {
            if (day == 0) {
                std::cout << "Day " << day << ": " << daily_fatigue_rate[day] << std::endl;
            } else if (daily_fatigue_rate[day] != daily_fatigue_rate[day-1]) {
                std::cout << "Day " << day << ": " << daily_fatigue_rate[day] << std::endl;
            }
            
        }
        std::cout << std::endl;
    }
}

// ------------------------------

void EpidemicSimulator::assign_SC_compliance_students_teachers()
{
    int n_students = 0;
    int n_teachers = 0;
    int n_students_thread = 0;
    int n_teachers_thread = 0;

    #pragma omp parallel firstprivate(n_students_thread, n_teachers_thread)
    {
        #pragma omp for
        for (unsigned long i = 0; i < agent_ids.size(); i++)
        {
            if (agents[agent_ids[i]]->isStudent())
            {
                agents[agent_ids[i]]->setSCState(SC_States::SC_child);
                n_students_thread += 1;
            }
            else if (agents[agent_ids[i]]->isTeacher())
            {
                agents[agent_ids[i]]->setSCState(SC_States::SC_teacher);
                n_teachers_thread += 1;
            }
        }

        #pragma omp critical(assign_SC_compliance_students_teachers_)
        {
            n_students += n_students_thread;
            n_teachers += n_teachers_thread;
        }
    }

    std::cout << "...SC applied to " << n_students << " students" << std::endl;
    std::cout << "...SC applied to " << n_teachers << " teachers" << std::endl;
}


void EpidemicSimulator::assign_SD_compliance()
{
    int n_SD_compliant = 0;
    int n_SD_compliant_thread = 0;
    #pragma omp parallel firstprivate(n_SD_compliant_thread)
    {
        #pragma omp for
        for (unsigned long i = 0; i < agent_ids.size(); i++)
        {
            float agent_compliance = rand_r(&seed_thread[omp_get_thread_num()]) / (float)RAND_MAX;
            if (agent_compliance < compliance_rate_SD)
            {
                agents[agent_ids[i]]->setSDState(SD_States::SD_compliant);
                n_SD_compliant_thread += 1;
            }
        }

        #pragma omp critical(assign_SD_compliance_)
        {
            n_SD_compliant += n_SD_compliant_thread;
        }
    }
    std::cout << " ***** SD compliance assigned to " << n_SD_compliant << " agents by assign_SD_compliance()" << std::endl;
}

// Assign both compliance and noncompliance state to agents
void EpidemicSimulator::assign_SD_compliance_Qmod(float SD_compliance_level, int days_since_last_SD_assignment, std::vector<bool> &list_SD_compliant)
{
    int n_SD_compliant = 0;
    int n_SD_compliant_thread = 0;

    #pragma omp parallel firstprivate(n_SD_compliant_thread)
    {
        #pragma omp for
        for (unsigned long i = 0; i < agent_ids.size(); i++)
        {
            if (days_since_last_SD_assignment == 0)
            {
                float agent_compliance = rand_r(&seed_thread[omp_get_thread_num()]) / (float)RAND_MAX;
                if (agent_compliance < SD_compliance_level)
                {
                    agents[agent_ids[i]]->setSDState(SD_States::SD_compliant);
                    list_SD_compliant[agent_ids[i]] = true;
                    n_SD_compliant_thread += 1;
                }
                else
                {
                    agents[agent_ids[i]]->setSDState(SD_States::SD_noncompliant);
                    list_SD_compliant[agent_ids[i]] = false;
                }
                if (agent_ids[i] == 0)
                {
                    std::cout << "\t***** Assigned new list of SD compliant agents." << std::endl;
                }
            }
            else
            {
                if (list_SD_compliant[agent_ids[i]])
                {
                    agents[agent_ids[i]]->setSDState(SD_States::SD_compliant);
                    n_SD_compliant_thread += 1;
                }
                else
                {
                    agents[agent_ids[i]]->setSDState(SD_States::SD_noncompliant);
                }
            }
        }

        #pragma omp critical(assign_SD_compliance_Qmod_)
        {
            n_SD_compliant += n_SD_compliant_thread;
        }
    }
    std::cout << "\t***** " << days_since_last_SD_assignment << " days since last SD assignment"
              << ", current SD compliance level is " << SD_compliance_level << ", SD compliance assigned to " << n_SD_compliant << " agents by assign_SD_compliance_Qmod()" << std::endl;
}

// ------------------------------

float EpidemicSimulator::assign_SD_compliance_opinion_Qmod(int days_since_last_SD_assignment, std::vector<unsigned>& detected_incidence_buffer, std::vector<bool>& list_SD_compliant, float fatigue_rate, unsigned day) {
    // The sigmoid function to determine the mean of a Normal distribution to sample each agent's horizon length
    float opinion_horizon_mean_to_sample = 2.0 * opinion_horizon_mean_sigmoid_midpoint / (1 + pow(M_E, (-1.0 * opinion_horizon_mean_sigmoid_v * (day - opinion_horizon_mean_sigmoid_delay)))) + opinion_horizon_base;

    // Define a distribution to sample the horizon for each agent
    std::normal_distribution<float> horizon_period_dist(opinion_horizon_mean_to_sample, opinion_horizon_std);
    std::random_device rd;
    std::default_random_engine generator(rd());
    
    if (days_since_last_SD_assignment == 0) {
        // 1st step: if it is the time to update agents' compliance, then we will first update all agents' rational opinions based on mean incidence of the past several days (dependent on each agent's own horizon)
        for (auto agent_entry : agents) {
            int   agent_id = (int) agent_entry.first;
            auto  c_agent  = agent_entry.second;

            if (agents_character[agent_id] == 0) {
                c_agent->set_rational_opinion_on_compliance(0.0);
            } else if (agents_character[agent_id] == 1) {
                c_agent->set_rational_opinion_on_compliance(1.0);
            } else if (agents_character[agent_id] == 2) {
                unsigned last_opinion_calculating_day = (unsigned) std::max((int) detected_incidence_buffer.size() - 1, 0);
                unsigned sampled_horizon_of_this_agent = (unsigned) std::max((float) 1.0, horizon_period_dist(generator));
                double total_weighted_detected_incidence = 0;
                double total_weight = 0;
                if (sampled_horizon_of_this_agent <= (last_opinion_calculating_day+1)) {
                    for (unsigned idx = last_opinion_calculating_day - sampled_horizon_of_this_agent + 1; idx <= last_opinion_calculating_day; idx++) {
                        double weight = pow(opinion_horizon_weight, last_opinion_calculating_day - idx);
                        total_weighted_detected_incidence += 1.0 * weight * detected_incidence_buffer[idx];
                        total_weight += weight;
                    }

                }

                double mean_detected_incidence;
                if (total_weight > 0) {
                    mean_detected_incidence = 1.0 * total_weighted_detected_incidence / total_weight;
                } else {
                    mean_detected_incidence = 0.0;
                }

                // calculate individual opinion based on the agent's observation on mean_detected_incidence
                double beta_perc = beta_perc_initial * pow((1 - fatigue_rate), day);
                double agent_rational_opinion_value  = 1 - pow((1 - beta_perc), 1000.0 * mean_detected_incidence / agents.size());

                c_agent->set_rational_opinion_on_compliance(agent_rational_opinion_value);
            }
        }
    }

    // 2nd step: update agent's compliance based on their rational opinion and their household members' influences
    int n_SD_compliant = 0;
    for (auto agent_entry : agents) {
        int   agent_id = (int) agent_entry.first;
        auto  c_agent  = agent_entry.second;

        if (days_since_last_SD_assignment == 0) {
            if (agent_id == 0) {
                std::cout << "\t ***** Assigned new list of SD compliant agents according to their opinion towards their horizon of detected incidence." << std::endl;
            }
            // Here we will prepare the new list of compliant/non-compliant agents according to their rational opinion and their household members' pressure
            // Their compliances is assigned according to their character: agent's character is 0 (never compliant) or 1 (alway compliant) or 2 (compliant based on opinion).
            if (agents_character[agent_id] == 0) {
                c_agent->setSDState(SD_States::SD_noncompliant);
                list_SD_compliant[agent_id] = false;
            } else if (agents_character[agent_id] == 1) {
                c_agent->setSDState(SD_States::SD_compliant);
                list_SD_compliant[agent_id] = true;
                n_SD_compliant += 1;
            } else {
                // For agents whose compliance changes according to their opinion
                // 1st: Agent's opinion will partially depend on its rational opinion (horizon of detected incidence).
                double agent_rational_opinion  = c_agent->get_rational_opinion_on_compliance();

                // 2nd: calculate social pressure
                // 2.1: for household
                double agent_social_pressure_household_component = 0;
                unsigned other_household_members_count = 0;
                if (social_pressure_household > 0) {
                    other_household_members_count = c_agent->getHH()->getAgentIdxs().size() - 1;
                    for (auto household_agent_id:c_agent->getHH()->getAgentIdxs()) {
                        if (household_agent_id != agent_id){
                            agent_social_pressure_household_component += (1.0 / other_household_members_count)  * agents[household_agent_id]->get_rational_opinion_on_compliance();
                        }
                    }
                }

                // 2.2: for household cluster
                double agent_social_pressure_householdcluster_component = 0;
                unsigned other_householdcluster_members_count = 0;
                if (social_pressure_householdcluster > 0) {
                    other_householdcluster_members_count = c_agent->getHC()->getAgentIdxs().size() - 1;
                    for (auto householdcluster_agent_id:c_agent->getHC()->getAgentIdxs()) {
                        if (householdcluster_agent_id != agent_id){
                            agent_social_pressure_householdcluster_component += (1.0 / other_householdcluster_members_count)  * agents[householdcluster_agent_id]->get_rational_opinion_on_compliance();
                        }
                    }
                }

                // 2.3: for neighborhood (CD)
                double agent_social_pressure_neighborhood_component = 0;
                unsigned other_CD_members_count = 0;
                if (social_pressure_CD > 0) {
                    other_CD_members_count = c_agent->getCD()->getAgentIdxs().size() - 1;
                    for (auto cd_agent_id:c_agent->getCD()->getAgentIdxs()) {
                        if (cd_agent_id != agent_id){
                            agent_social_pressure_neighborhood_component += (1.0 / other_CD_members_count)  * agents[cd_agent_id]->get_rational_opinion_on_compliance();
                        }
                    }
                }

                // 2.4: for community (SLA)
                double agent_social_pressure_community_component = 0;
                unsigned other_SLA_members_count = 0;
                if (social_pressure_SLA > 0) {
                    other_SLA_members_count = c_agent->getSLA()->getAgentIdxs().size() - 1;
                    for (auto sla_agent_id:c_agent->getSLA()->getAgentIdxs()) {
                        if (sla_agent_id != agent_id){
                            agent_social_pressure_community_component += (1.0 / other_SLA_members_count)  * agents[sla_agent_id]->get_rational_opinion_on_compliance();
                        }
                    }
                }

                // 2.5: for working group
                double agent_social_pressure_workinggroup_component = 0;
                unsigned other_workinggroup_members_count = 0;
                if ((social_pressure_workinggroup > 0) && (c_agent->getWGs().size() == 1)) {
                    other_workinggroup_members_count = c_agent->getWGs().front()->getAgentIdxs().size() - 1;
                    for (auto workinggroup_agent_id:c_agent->getWGs().front()->getAgentIdxs()) {
                        if (workinggroup_agent_id != agent_id){
                            agent_social_pressure_workinggroup_component += (1.0 / other_workinggroup_members_count)  * agents[workinggroup_agent_id]->get_rational_opinion_on_compliance();
                        }
                    }
                }

                // 2.6: for class (type=2)
                double agent_social_pressure_class_component = 0;
                unsigned other_class_members_count = 0;
                if ((social_pressure_class > 0) && (c_agent->getWGs().size() == 3)) {
                    other_class_members_count = c_agent->getWGs().at(2)->getAgentIdxs().size() - 1;
                    for (auto class_agent_id:c_agent->getWGs().at(2)->getAgentIdxs()) {
                        if (class_agent_id != agent_id){
                            agent_social_pressure_class_component += (1.0 / other_class_members_count)  * agents[class_agent_id]->get_rational_opinion_on_compliance();
                        }
                    }
                }

                // 2.7: for grade (type=1)
                double agent_social_pressure_grade_component = 0;
                unsigned other_grade_members_count = 0;
                if ((social_pressure_grade > 0) && (c_agent->getWGs().size() == 3)) {
                    other_grade_members_count = c_agent->getWGs().at(1)->getAgentIdxs().size() - 1;
                    for (auto grade_agent_id:c_agent->getWGs().at(1)->getAgentIdxs()) {
                        if (grade_agent_id != agent_id){
                            agent_social_pressure_grade_component += (1.0 / other_grade_members_count)  * agents[grade_agent_id]->get_rational_opinion_on_compliance();
                        }
                    }
                }

                // 2.8: for school (type=0)
                double agent_social_pressure_school_component = 0;
                unsigned other_school_members_count = 0;
                if ((social_pressure_school > 0) && (c_agent->getWGs().size() == 3)) {
                    other_school_members_count = c_agent->getWGs().at(0)->getAgentIdxs().size() - 1;
                    for (auto school_agent_id:c_agent->getWGs().at(0)->getAgentIdxs()) {
                        if (school_agent_id != agent_id){
                            agent_social_pressure_school_component += (1.0 / other_school_members_count)  * agents[school_agent_id]->get_rational_opinion_on_compliance();
                        }
                    }
                }

                // 2.9: calculate overall social pressure
                double agent_social_pressure =   social_pressure_household * agent_social_pressure_household_component + 
                                                 social_pressure_householdcluster * agent_social_pressure_householdcluster_component + 
                                                 social_pressure_CD * agent_social_pressure_neighborhood_component +
                                                 social_pressure_SLA * agent_social_pressure_community_component +
                                                 social_pressure_workinggroup * agent_social_pressure_workinggroup_component +
                                                 social_pressure_school * agent_social_pressure_school_component +
                                                 social_pressure_grade * agent_social_pressure_grade_component +
                                                 social_pressure_class * agent_social_pressure_class_component;


                // 3rd: combine both agent's rational opinion and social pressure to calculate agent's opinion
                double agent_overall_opinion;
                if ((other_household_members_count > 0) || (other_householdcluster_members_count > 0) || (other_CD_members_count > 0) || (other_SLA_members_count > 0) || (other_workinggroup_members_count > 0) || 
                (other_school_members_count > 0) || (other_grade_members_count > 0) || (other_class_members_count > 0)) {
                    agent_overall_opinion = social_pressure_weight * agent_social_pressure + (1 - social_pressure_weight) * agent_rational_opinion; 
                } else {
                    agent_overall_opinion = agent_rational_opinion;
                }

                if (agent_overall_opinion > opinion_threshold)
                {
                    c_agent->setSDState(SD_States::SD_compliant);
                    list_SD_compliant[agent_id] = true;
                    n_SD_compliant += 1;
                } else {
                    c_agent->setSDState(SD_States::SD_noncompliant);
                    list_SD_compliant[agent_id] = false;
                }

            }
        } else {
            if (list_SD_compliant[agent_id])
            {
                c_agent->setSDState(SD_States::SD_compliant);
                n_SD_compliant += 1;
            } else {
                c_agent->setSDState(SD_States::SD_noncompliant);
            }
        }
    }
    std::cout << "\t ***** " << days_since_last_SD_assignment << " days since last SD assignment" << ", current SD compliance level is " << (double) n_SD_compliant/agents.size() << ", SD compliance assigned to " << n_SD_compliant << " agents by assign_SD_compliance_opinion_Qmod()" << std::endl;
    return  (double) n_SD_compliant/agents.size();
}


void EpidemicSimulator::assign_CI_compliance()
{
    int n_CI_compliant = 0;
    int n_CI_compliant_thread = 0;
    #pragma omp parallel firstprivate(n_CI_compliant_thread)
    {
        #pragma omp for
        for (unsigned long i = 0; i < agent_ids.size(); i++)
        {
            float agent_compliance = rand_r(&seed_thread[omp_get_thread_num()]) / (float)RAND_MAX;
            if (agent_compliance < compliance_rate_CI)
            {
                agents[agent_ids[i]]->setCIState(CI_States::CI_compliant);
                n_CI_compliant_thread += 1;
            }
        }

        #pragma omp critical(assign_CI_compliance_)
        {
            n_CI_compliant += n_CI_compliant_thread;
        }
    }

    std::cout << " ***** CI compliance assigned to " << n_CI_compliant << " agents by assign_CI_compliance()" << std::endl;
}


void EpidemicSimulator::assign_HQ_compliance()
{
    int n_HQ_compliant = 0;
    int n_HQ_compliant_thread = 0;
    #pragma omp parallel firstprivate(n_HQ_compliant_thread)
    {
        #pragma omp for
        for (unsigned long i = 0; i < agent_ids.size(); i++)
        {
            float agent_compliance = rand_r(&seed_thread[omp_get_thread_num()]) / (float)RAND_MAX;
            if ((agents[agent_ids[i]]->getCIState() == CI_States::CI_compliant) && (agent_compliance < compliance_rate_HQ))
            {
                agents[agent_ids[i]]->setHQState(HQ_States::HQ_compliant);
                n_HQ_compliant_thread += 1;
            }
        }

        #pragma omp critical(assign_HQ_compliance_)
        {
            n_HQ_compliant += n_HQ_compliant_thread;
        }
    }

    std::cout << "***** HQ compliance assigned to " << n_HQ_compliant << " agents by assign_HQ_compliance()" << std::endl;
}


void EpidemicSimulator::set_Compliance_states_all()
{
    #pragma omp parallel for
    for (unsigned long i = 0; i < agent_ids.size(); i++)
    {
        agents[agent_ids[i]]->eval_Compliance_state();
    }
}


void EpidemicSimulator::set_Compliance_states_all_Qmod()
{
    #pragma omp parallel for
    for (unsigned long i = 0; i < agent_ids.size(); i++)
    {
        agents[agent_ids[i]]->eval_Compliance_state_Qmod();
    }
}


void EpidemicSimulator::assign_SC_compliance_parents()
{
    // reset SC compliance for parents
    #pragma omp parallel for
    for (unsigned long i = 0; i < agent_ids.size(); i++) {
        if (agents[agent_ids[i]]->isAdult() && !agents[agent_ids[i]]->isTeacher()) {
            agents[agent_ids[i]]->setSCState(SC_States::SC_nil);
        }
    }

    // decide which parents are staying home
    int n_parents_SC = 0;

    #pragma omp parallel
    {
        #pragma omp for
        for (unsigned long i = 0; i < agent_ids.size(); i++)
        {
            if (agents[agent_ids[i]]->getSCState() == SC_States::SC_child)
            {
                #pragma omp critical(assign_SC_compliance_parents_1)
                {
                    int parent_index = agents[agent_ids[i]]->findParent_ID(agents);
                    if (parent_index < 0) {
                        std::cout << "WARNING: Something went wrong in findParent_ID(), no parent found" << std::endl;
                    } else {
                        unsigned long parent_ID = (unsigned long)parent_index;
                        if (agents.at(parent_ID)->getSCState() == SC_States::SC_nil)
                        {
                            if ((rand_r(&seed_thread[omp_get_thread_num()]) / (float)RAND_MAX) < p_parent_stays_home)
                            {
                                agents.at(parent_ID)->setSCState(SC_States::SC_parent);
                                n_parents_SC += 1;
                            }
                        }
                    }
                }
            }
        }
    }

    std::cout << "\tParents taking care of children at home: " << n_parents_SC << std::endl;
}

// ------------------------------

void EpidemicSimulator::step_and_detect(unsigned day)
{
#pragma omp parallel for
    for (unsigned long i = 0; i < agent_ids.size(); i++)
    {
        agents[agent_ids[i]]->step(day);
        if (agents[agent_ids[i]]->isInfected())
        {
            bool detected_agent = false;
            // If the agent is symptomatic and not detected check for detection
            if ((agents[agent_ids[i]]->getState() == DiseaseStates::StateSymptomatic) && (agents[agent_ids[i]]->isIllnessDetected() == false))
            {
                detected_agent = agents[agent_ids[i]]->Illness_detection(Detection_);
            }
            else
            {
                if ((agents[agent_ids[i]]->getState() == DiseaseStates::StateAsymptomatic) && (agents[agent_ids[i]]->isIllnessDetected() == false)) // if agent is asymptomatic
                // remember that the state 'asymptomatic' applies to both pre-symptomatic and asymptomatic individuals who are otherwise
                // differentiated by the sign of the incubating_period_ (this value is set to -1 for asymptomatic agents)
                {
                    detected_agent = agents[agent_ids[i]]->Illness_detection(Detection_asymp);
                }
            }
        }
    }
}

// ------------------------------

void EpidemicSimulator::step_without_detection(unsigned day)
{
    #pragma omp parallel for
    for (unsigned long i = 0; i < agent_ids.size(); i++) {
        agents[agent_ids[i]]->step(day);
    }
}

// ------------------------------

// this function sets up the population for correct factor computation
// an agents isolation state is a function of the implemented policies and their compliance state.
void EpidemicSimulator::update_Isolation_states(bool SD_flag, bool CI_flag, bool HQ_flag, bool SC_flag)
{
    int CI_counter = 0;
    int SD_counter = 0;
    int HQ_counter = 0;
    int SC_counter = 0;
    int CI_counter_thread = 0;
    int SD_counter_thread = 0;
    int HQ_counter_thread = 0;
    int SC_counter_thread = 0;
    bool assigned;

    // This script implicily assumes a ranking of isolation states  CI > HQ > SD > SC
    // This should be reflected in the FoI terms.
    
    // have to add a separate loop for CI because HQ is conditioned on it.
    // CI will apply to all compliant agents who are ill, it trumps all other measures they may comply with
    // This means there also needs to be a separate loop for clearing all residual states may as well combine these.

    #pragma omp parallel firstprivate(CI_counter_thread)
    {
        #pragma omp for
        for (unsigned long i = 0; i < agent_ids.size(); i++)
        {
            // first, set the isolation state to nil, to make sure only currently active measures are applied
            // the exception is for quarantined agents, who get reset in the function update_HQ (As this measure has a clock attached to it)
            if (agents[agent_ids[i]]->getIsolationState() != Isolation_States::HQ)
            {
                agents[agent_ids[i]]->setIsolationState(Isolation_States::nil);
            }

            if (CI_flag)
            {
                if ((agents[agent_ids[i]]->getComplianceState() > 1) &&                                                                   // agent is CI compliant
                    ((agents[agent_ids[i]]->getState() == DiseaseStates::StateSymptomatic) || agents[agent_ids[i]]->isIllnessDetected())) // added case isolation for detected asymptomatic agents as well.
                {
                    agents[agent_ids[i]]->setIsolationState(Isolation_States::CI);
                    CI_counter_thread += 1;
                    #pragma omp critical(update_Isolation_states_0)
                    if (!agents[agent_ids[i]]->getHH()->getCIinHousehold()) {
                        agents[agent_ids[i]]->getHH()->setCIinHousehold(true);
                    }
                }
            }
        }
        #pragma omp critical(update_Isolation_states_1)
        {
            CI_counter += CI_counter_thread;
        }
    }

    // now go through and apply all other possible interventions in order of 'strength' of the intervention
    #pragma omp parallel firstprivate(HQ_counter_thread, SD_counter_thread, SC_counter_thread)
    {
        #pragma omp for
        for (unsigned long i = 0; i < agent_ids.size(); i++)
        {
            if (agents[agent_ids[i]]->getIsolationState() != Isolation_States::CI)
            {
                // HQ will apply to all compliant agents who have a symptomatic (detected) illness in their home, unless they
                // are in case isolation themselves or are already quarantined.
                if (HQ_flag)
                {
                    if (agents[agent_ids[i]]->getIsolationState() == Isolation_States::HQ)
                    {
                        HQ_counter_thread += 1;
                        continue;
                    }
                    else if ((agents[agent_ids[i]]->getHH()->getCIinHousehold()) && ((agents[agent_ids[i]]->getComplianceState() > 2) && (agents[agent_ids[i]]->getComplianceState() < 5)))
                    {
                        agents[agent_ids[i]]->setIsolationState(Isolation_States::HQ);
                        HQ_counter_thread += 1;
                        continue;
                    }
                }

                // SD will apply to any compliant agents who are not actively in HQ or CI
                if (SD_flag)
                {
                    if (((agents[agent_ids[i]]->getComplianceState() > 0) && (agents[agent_ids[i]]->getComplianceState() < 4)))
                    {
                        agents[agent_ids[i]]->setIsolationState(Isolation_States::SD);
                        SD_counter_thread += 1;
                        continue;
                    }
                }

                // school closure only applies to teachers, students, and parents who are not
                // already isolated by one of the other measures.
                if (SC_flag)
                {
                    if ((agents[agent_ids[i]]->getSCState() > 0))
                    {
                        agents[agent_ids[i]]->setIsolationState(Isolation_States::SC);
                        SC_counter_thread += 1;
                        continue;
                    }
                }
            }
        }

        #pragma omp critical(update_Isolation_states_2)
        {
            HQ_counter += HQ_counter_thread;
            SD_counter += SD_counter_thread;
            SC_counter += SC_counter_thread;
        }
    }

    g_HQ_agents = HQ_counter;
    g_CI_agents = CI_counter;

    std::cout << "\t"
              << "   CI count: " << CI_counter << ","
              << "\t"
              << "   HQ count: " << HQ_counter << ","
              << "\t"
              << "   SD count: " << SD_counter << ","
              << "\t"
              << "   SC count: " << SC_counter << std::endl;
}

// ------------------------------

void EpidemicSimulator::pre_epidemic_vaccination(std::stack<unsigned long> &unvaccinated_senior_IDs_s,
                                                 std::stack<unsigned long> &unvaccinated_adult_IDs_s,
                                                 std::stack<unsigned long> &unvaccinated_child_IDs_s,
                                                 unsigned day)
{
    unsigned long n_unvaccinated = unvaccinated_senior_IDs_s.size() + unvaccinated_adult_IDs_s.size();

    while ((num_vac < n_vac_pre_epidemic_PV + n_vac_pre_epidemic_GV) && (n_unvaccinated > 0))
    {
        if (unvaccinated_senior_IDs_s.size() > 0)
        {
            for (int i = 1; i < p_vac_65_and_over + 1; i++)
            {
                if ((unvaccinated_senior_IDs_s.size() > 0) && (num_vac < n_vac_pre_epidemic_PV + n_vac_pre_epidemic_GV))
                {
                    auto c_agent = agents.at(unvaccinated_senior_IDs_s.top());
                    if (c_agent->wasIllnessDetected())
                    {
                        unvaccinated_senior_IDs_s.pop();
                    }
                    else
                    {
                        if (num_vac < n_vac_pre_epidemic_PV)
                        {
                            // Add an immunisation to this agent's immunisation history after this agent is vaccinated
                            c_agent->add_an_immunisation_to_history(day, VEc_PV, VEt_PV, daily_waning_VEc_PV, daily_waning_VEt_PV, Immunisation_Types::vaccinated_PV);
                            num_vac_PV++;
                            num_vac_65_plus_PV++;
                        }
                        else
                        {
                            // Add an immunisation to this agent's immunisation history after this agent is vaccinated
                            c_agent->add_an_immunisation_to_history(day, VEc_GV, VEt_GV, daily_waning_VEc_GV, daily_waning_VEt_GV, Immunisation_Types::vaccinated_GV);
                            num_vac_GV++;
                            num_vac_65_plus_GV++;
                        }
                        unvaccinated_senior_IDs_s.pop();
                        num_vac++;
                        num_vac_65_plus++;
                    }
                }
            }
        }

        if (unvaccinated_adult_IDs_s.size() > 0)
        {
            for (int i = 1; i < p_vac_18_to_64 + 1; i++)
            {
                if ((unvaccinated_adult_IDs_s.size() > 0) && (num_vac < n_vac_pre_epidemic_PV + n_vac_pre_epidemic_GV))
                {
                    auto c_agent = agents[unvaccinated_adult_IDs_s.top()];
                    if (c_agent->wasIllnessDetected())
                    {
                        unvaccinated_adult_IDs_s.pop();
                    }
                    else
                    {
                        if (num_vac < n_vac_pre_epidemic_PV)
                        {
                            // Add an immunisation to this agent's immunisation history after this agent is vaccinated
                            c_agent->add_an_immunisation_to_history(day, VEc_PV, VEt_PV, daily_waning_VEc_PV, daily_waning_VEt_PV, Immunisation_Types::vaccinated_PV);
                            num_vac_PV++;
                            num_vac_18_to_64_PV++;
                        }
                        else
                        {
                            // Add an immunisation to this agent's immunisation history after this agent is vaccinated
                            c_agent->add_an_immunisation_to_history(day, VEc_GV, VEt_GV, daily_waning_VEc_GV, daily_waning_VEt_GV, Immunisation_Types::vaccinated_GV);
                            num_vac_GV++;
                            num_vac_18_to_64_GV++;
                        }
                        unvaccinated_adult_IDs_s.pop();
                        num_vac++;
                        num_vac_18_to_64++;
                    }
                }
            }
        }

        if (unvaccinated_child_IDs_s.size() > 0)
        {
            for (int i = 0; i < p_vac_children; i++)
            {
                if (unvaccinated_child_IDs_s.size() > 0 && (num_vac < n_vac_pre_epidemic_PV + n_vac_pre_epidemic_PV))
                {
                    auto c_agent = agents.at(unvaccinated_child_IDs_s.top());
                    if (c_agent->wasIllnessDetected())
                    {
                        unvaccinated_child_IDs_s.pop();
                    }
                    else
                    {
                        if (num_vac < n_vac_pre_epidemic_PV)
                        {
                            // Add an immunisation to this agent's immunisation history after this agent is vaccinated
                            c_agent->add_an_immunisation_to_history(day, VEc_PV, VEt_PV, daily_waning_VEc_PV, daily_waning_VEt_PV, Immunisation_Types::vaccinated_PV);
                            num_vac_PV++;
                            num_vac_children_PV++;
                        }
                        else
                        {
                            // Add an immunisation to this agent's immunisation history after this agent is vaccinated
                            c_agent->add_an_immunisation_to_history(day, VEc_GV, VEt_GV, daily_waning_VEc_GV, daily_waning_VEt_GV, Immunisation_Types::vaccinated_GV);
                            num_vac_GV++;
                            num_vac_children_GV++;
                        }
                        unvaccinated_child_IDs_s.pop();
                        num_vac++;
                        num_vac_children++;
                    }
                }
            }
        }

        n_unvaccinated = unvaccinated_senior_IDs_s.size() +
                         unvaccinated_adult_IDs_s.size();

    }

    std::cout << "\t- Number of pre-vaccinated agents: " << num_vac << std::endl;
    std::cout << "\t- Number of pre-vaccinated children: " << num_vac_children << std::endl;
    std::cout << "\t- Number of pre-vaccinated aged 18-64: " << num_vac_18_to_64 << std::endl;
    std::cout << "\t- Number of pre-vaccinated aged 65+: " << num_vac_65_plus << std::endl;

    std::cout << "\t- Number of pre-vaccinated agents (Priority): " << num_vac_PV << std::endl;
    std::cout << "\t- Number of pre-vaccinated children (Priority): " << num_vac_children_PV << std::endl;
    std::cout << "\t- Number of pre-vaccinated aged 18-64 (Priority): " << num_vac_18_to_64_PV << std::endl;
    std::cout << "\t- Number of pre-vaccinated aged 65+ (Priority): " << num_vac_65_plus_PV << std::endl;

    std::cout << "\t- Number of pre-vaccinated agents (General): " << num_vac_GV << std::endl;
    std::cout << "\t- Number of pre-vaccinated children (General): " << num_vac_children_GV << std::endl;
    std::cout << "\t- Number of pre-vaccinated aged 18-64 (General): " << num_vac_18_to_64_GV << std::endl;
    std::cout << "\t- Number of pre-vaccinated aged 65+ (General): " << num_vac_65_plus_GV << std::endl;
}

// ------------------------------

unsigned long EpidemicSimulator::getNumAgentsRemoved()
{
    unsigned long num = 0;
    unsigned long num_thread = 0;
    #pragma omp parallel firstprivate(num_thread)
    {
        #pragma omp for
        for (unsigned long i = 0; i < agent_ids.size(); i++) {
            if (agents[agent_ids[i]]->isRemoved()) {
                num_thread++;
            }
        }

        #pragma omp critical(getNumAgentsIll_)
        {
            num += num_thread;
        }
    }
    return (num);
}

// ------------------------------

unsigned long EpidemicSimulator::getNumAgentsIll()
{
    unsigned long num = 0;
    unsigned long num_thread = 0;
    #pragma omp parallel firstprivate(num_thread)
    {
        #pragma omp for
        for (unsigned long i = 0; i < agent_ids.size(); i++)
        {
            if (agents[agent_ids[i]]->isIll())
            {
                num_thread++;
            }
        }
        #pragma omp critical(getNumAgentsIll_)
        {
            num += num_thread;
        }
    }
    return (num);
}

// ------------------------------

unsigned long EpidemicSimulator::getNumAgentsInfected()
{
    unsigned long num = 0;
    unsigned long num_thread = 0;
    #pragma omp parallel firstprivate(num_thread)
    {
        #pragma omp for
        for (unsigned long i = 0; i < agent_ids.size(); i++)
        {
            if (agents[agent_ids[i]]->isInfected())
            {
                num_thread++;
            }
        }
        #pragma omp critical(getNumAgentsIll_)
        {
            num += num_thread;
        }
    }
    return (num);
}

// ------------------------------
// ------------------------------

// set isolation factors for FoI based on Agent's isolation states.
void EpidemicSimulator::set_Isolation_Factors_home(float FoI_SD, float FoI_CI, float FoI_HQ, float FoI_SC)
{
    #pragma omp parallel for
    for (unsigned long i = 0; i < agent_ids.size(); i++)
    {
        agents[agent_ids[i]]->setIsolationFactor_home(FoI_SD, FoI_CI, FoI_HQ, FoI_SC);
    }
}

// ------------------------------

void EpidemicSimulator::set_Isolation_Factors_work(float FoI_SD, float FoI_CI, float FoI_HQ, float FoI_SC)
{
    #pragma omp parallel for
    for (unsigned long i = 0; i < agent_ids.size(); i++) {
        agents[agent_ids[i]]->setIsolationFactor_work(FoI_SD, FoI_CI, FoI_HQ, FoI_SC);
    }
}

// ------------------------------

void EpidemicSimulator::set_Isolation_Factors_community(float FoI_SD, float FoI_CI, float FoI_HQ, float FoI_SC)
{
    #pragma omp parallel for
    for (unsigned long i = 0; i < agent_ids.size(); i++) {
        agents[agent_ids[i]]->setIsolationFactor_community(FoI_SD, FoI_CI, FoI_HQ, FoI_SC);
    }
}

// ------------------------------
// ------------------------------

void EpidemicSimulator::sampleSourcesForAnInfection(vector<unsigned> &potential_source_IDs, vector<float> &potential_source_weights, vector<unsigned> &source_IDs)
{
    std::random_device rd;
    std::mt19937 gen(rd());
    std::discrete_distribution<> dist(potential_source_weights.begin(), potential_source_weights.end());

    // An infection happens when the bag of potential sources has at least 1 infectious agent.
    if (potential_source_IDs.size() > 0)
    {
        unsigned int sampled_idx = (unsigned int)dist(gen);
        source_IDs.emplace_back(potential_source_IDs[sampled_idx]);

        // An co-infection happens with a rate and when the bag of potential sources has at least 2 infectious agents.
        if (((rand_r(&seed_thread[omp_get_thread_num()]) / (float)RAND_MAX) < co_infection_rate) && (potential_source_IDs.size() > 1))
        {
            potential_source_IDs.erase(potential_source_IDs.begin() + sampled_idx);
            potential_source_weights.erase(potential_source_weights.begin() + sampled_idx);
            dist = std::discrete_distribution<>(potential_source_weights.begin(), potential_source_weights.end());
            sampled_idx = (unsigned int)dist(gen);

            if (potential_source_IDs[sampled_idx] != source_IDs[0])
            {
                source_IDs.emplace_back(potential_source_IDs[sampled_idx]);
            }
        }
    }
    else
    {
        std::cout << "ERROR: Infection happens when the bag of potential sources is empty." << std::endl;
        exit(1);
    }
}

// ------------------------------
// ------------------------------

void EpidemicSimulator::simulateDaytime_SD(unsigned day)
{
    /* Update transmissibility at work */
    #pragma omp parallel for
    for (unsigned long i = 0; i < agent_ids.size(); i++) {
        if (agents[agent_ids[i]]->isInfected()) {
            agents[agent_ids[i]]->addToDaytimeProbabilityOfInfection_SD(day);
        }
    }

    /* Update susceptibility at work */
    // decide who is now infected
    #pragma omp parallel for
    for (unsigned long i = 0; i < agent_ids.size(); i++)
    {
        if (agents[agent_ids[i]]->isStudent())
        {
            float P_g = 1.0 - (1.0 - agents[agent_ids[i]]->getIsolationFactor_work() * agents[agent_ids[i]]->getDaytimeProbabilityOfInfection(2)) 
                            * (1.0 - agents[agent_ids[i]]->getIsolationFactor_work() * agents[agent_ids[i]]->getDaytimeProbabilityOfInfection(1)) 
                            * (1.0 - agents[agent_ids[i]]->getIsolationFactor_work() * agents[agent_ids[i]]->getDaytimeProbabilityOfInfection(0));
            updateAgentStateAllContexts(agents[agent_ids[i]], P_g, true, day);
        }
        else
        {
            float P_g = 1.0 - (1.0 - agents[agent_ids[i]]->getIsolationFactor_work() * agents[agent_ids[i]]->getDaytimeProbabilityOfInfection());
            updateAgentStateAllContexts(agents[agent_ids[i]], P_g, true, day);
        }
    }

}


void EpidemicSimulator::simulateDaytimeWeekend_SD(unsigned day)
{
    /* Update distributions for home */
    #pragma omp parallel for
    for (unsigned long i = 0; i < agent_ids.size(); i++)
    {
        if (agents[agent_ids[i]]->isInfected())
        {
            agents[agent_ids[i]]->addToNighttimeProbabilityOfInfection_SD(day);
        }
    }
}


void EpidemicSimulator::simulateNighttime_SD(unsigned day)
{
    /* Update distributions for home */
    #pragma omp parallel for
    for (unsigned long i = 0; i < agent_ids.size(); i++) {
        if (agents[agent_ids[i]]->isInfected()) {
            agents[agent_ids[i]]->addToNighttimeProbabilityOfInfection_SD(day);
        }
    }

    /* Decide who is now infected */
    #pragma omp parallel for
    for (unsigned long i = 0; i < agent_ids.size(); i++)
    {
        float P_g = 1.0 - (1.0 - agents[agent_ids[i]]->getHouseholdProbabilityOfInfection() * agents[agent_ids[i]]->getIsolationFactor_home()) 
                        * (1.0 - agents[agent_ids[i]]->getHouseholdClusterProbabilityOfInfection() * agents[agent_ids[i]]->getIsolationFactor_community()) 
                        * (1.0 - agents[agent_ids[i]]->getNeighbourhoodProbabilityOfInfection() * agents[agent_ids[i]]->getIsolationFactor_community()) 
                        * (1.0 - agents[agent_ids[i]]->getCommunityProbabilityOfInfection() * agents[agent_ids[i]]->getIsolationFactor_community());
        updateAgentStateAllContexts(agents[agent_ids[i]], P_g, false, day);

        // bookkeeping
        auto my_sla = agents[agent_ids[i]]->getSLA()->getID();
        if (!(agents[agent_ids[i]]->isSusceptible()))
        {
            #pragma omp critical(simulateNighttime_SD_2)
            {
                if (agents[agent_ids[i]]->isRemoved())
                {
                    g_number_removed++;
                    g_sla_number_removed_map.at(my_sla)++;
                    g_ag_sla_number_removed_map.at(agents[agent_ids[i]]->getAgeGroup()).at(my_sla)++;
                }
                else if (agents[agent_ids[i]]->isInfected())
                {
                    g_infected_prevalence++;
                    g_ag_infection_prevalence_map.at(agents[agent_ids[i]]->getAgeGroup())++;

                    if (agents[agent_ids[i]]->isIll())
                    {
                        g_illness_prevalence++;
                        g_ag_illness_prevalence_map.at(agents[agent_ids[i]]->getAgeGroup())++;
                    }
                    if (agents[agent_ids[i]]->isIllnessDetected())
                    {
                        g_detected_prevalence++;
                        g_ag_detection_prevalence_map.at(agents[agent_ids[i]]->getAgeGroup())++;
                        g_sla_prevalence_map.at(my_sla)++;
                        g_ag_sla_prevalence_map.at(agents[agent_ids[i]]->getAgeGroup()).at(my_sla)++;
                    }
                }
            }
        }
    }

}


void EpidemicSimulator::simulateDaytime(unsigned day)
{
    /* Update transmissibility at work */
    #pragma omp parallel for
    for (unsigned long i = 0; i < agent_ids.size(); i++)
    {
        bool detected_agent = false;

        agents[agent_ids[i]]->step(day);

        if (agents[agent_ids[i]]->isInfected())
        {
            if ((agents[agent_ids[i]]->getState() == DiseaseStates::StateSymptomatic) && (agents[agent_ids[i]]->isIllnessDetected() == false))
            {
                // If the agent is symptomatic and not detected check for detection
                detected_agent = agents[agent_ids[i]]->Illness_detection(Detection_);
            }
            else
            {
                if ((agents[agent_ids[i]]->getState() == DiseaseStates::StateAsymptomatic) && (agents[agent_ids[i]]->isIllnessDetected() == false))
                {
                    // if agent is asymptomatic
                    // remember that the state 'asymptomatic' applies to both pre-symptomatic and asymptomatic individuals who are otherwise
                    // differentiated by the sign of the incubating_period_ (this value is set to -1 for asymptomatic agents)
                    detected_agent = agents[agent_ids[i]]->Illness_detection(Detection_asymp);
                }
            }
            
            agents[agent_ids[i]]->addToDaytimeProbabilityOfInfection(day);
        }
    }

    // decide who is now infected at work
    #pragma omp parallel for
    for (unsigned long i = 0; i < agent_ids.size(); i++) {
        // This agent is then sampled to know whether it becomes infectious.
        if (agents[agent_ids[i]]->isStudent())
        {
            float P_g = 1.0 - (1.0 - agents[agent_ids[i]]->getDaytimeProbabilityOfInfection(2)) 
                            * (1.0 - agents[agent_ids[i]]->getDaytimeProbabilityOfInfection(1)) 
                            * (1.0 - agents[agent_ids[i]]->getDaytimeProbabilityOfInfection(0));
            updateAgentStateAllContexts(agents[agent_ids[i]], P_g, true, day);
        }
        else
        {
            float P_g = 1.0 - (1.0 - agents[agent_ids[i]]->getDaytimeProbabilityOfInfection());
            updateAgentStateAllContexts(agents[agent_ids[i]], P_g, true, day);
        }
    }
}

// ------------------------------

void EpidemicSimulator::simulateNighttime(unsigned day)
{
    #pragma omp parallel for
    for (unsigned long i = 0; i < agent_ids.size(); i++) {
        agents[agent_ids[i]]->step(day);

        if (agents[agent_ids[i]]->isInfected())
        {
            agents[agent_ids[i]]->addToNighttimeProbabilityOfInfection(day);
        }
    }

    // Decide who is now infected
    #pragma omp parallel for
    for (unsigned long i = 0; i < agent_ids.size(); i++) {
        float P_g = 1.0 - (1.0 - agents[agent_ids[i]]->getHouseholdProbabilityOfInfection()) 
                        * (1.0 - agents[agent_ids[i]]->getHouseholdClusterProbabilityOfInfection()) 
                        * (1.0 - agents[agent_ids[i]]->getNeighbourhoodProbabilityOfInfection()) 
                        * (1.0 - agents[agent_ids[i]]->getCommunityProbabilityOfInfection());
        updateAgentStateAllContexts(agents[agent_ids[i]], P_g, false, day);

        // bookkeeping
        auto my_sla = agents[agent_ids[i]]->getSLA()->getID();
        if (!(agents[agent_ids[i]]->isSusceptible()))
        {
            if (agents[agent_ids[i]]->isRemoved())
            {
                #pragma omp critical(simulateNighttime_1)
                {
                    g_number_removed++;
                    g_sla_number_removed_map.at(my_sla)++;
                    g_ag_sla_number_removed_map.at(agents[agent_ids[i]]->getAgeGroup()).at(my_sla)++;
                }
            }
            else if (agents[agent_ids[i]]->isInfected())
            {
                #pragma omp critical(simulateNighttime_2)
                {
                    g_infected_prevalence++;
                    g_ag_infection_prevalence_map.at(agents[agent_ids[i]]->getAgeGroup())++;

                    if (agents[agent_ids[i]]->isIll())
                    {
                        g_illness_prevalence++;
                        g_ag_illness_prevalence_map.at(agents[agent_ids[i]]->getAgeGroup())++;
                    }

                    if (agents[agent_ids[i]]->isIllnessDetected())
                    {
                        g_detected_prevalence++;
                        g_ag_detection_prevalence_map.at(agents[agent_ids[i]]->getAgeGroup())++;
                        g_sla_prevalence_map.at(my_sla)++;
                        g_ag_sla_prevalence_map.at(agents[agent_ids[i]]->getAgeGroup()).at(my_sla)++;
                    }
                }
            }
        }
    }
}

// ------------------------------

void EpidemicSimulator::simulateDaytimeWeekend(unsigned day)
{
    #pragma omp parallel for
    for (unsigned long i = 0; i < agent_ids.size(); i++) {
        bool detected_agent = false;
        agents[agent_ids[i]]->step(day);
        
        if (agents[agent_ids[i]]->isInfected())
        {
            // detection occurs in the daytime
            if ((agents[agent_ids[i]]->getState() == DiseaseStates::StateSymptomatic) && (agents[agent_ids[i]]->isIllnessDetected() == false)) // If the agent is symptomatic and not detected check for detection
            {
                detected_agent = agents[agent_ids[i]]->Illness_detection(Detection_);
            }
            else
            {
                if ((agents[agent_ids[i]]->getState() == DiseaseStates::StateAsymptomatic) && (agents[agent_ids[i]]->isIllnessDetected() == false))
                {
                    detected_agent = agents[agent_ids[i]]->Illness_detection(Detection_asymp);
                }
            }

            // Every weekend day now is consistently set up with 2 cycles, but only the nighttime cycle is used to update agents' susceptibility.
            // The probability used in this infection check, however, accounts for the possibility of transmitted the virus in both cycles in a day.
            agents[agent_ids[i]]->addToNighttimeProbabilityOfInfection(day);
        }
    }
}

// ------------------------------
// ------------------------------
void EpidemicSimulator::updateAgentStateAllContexts(std::shared_ptr<Agent> c_agent, float P_E, bool in_day_time_contexts, unsigned day)
{
    if (c_agent->isSusceptible() && (P_E > 0))
    {
        if ((rand_r(&seed_thread[omp_get_thread_num()]) / (float)RAND_MAX) < P_E) {
            // A bag of weights containining p_g (j->i) is formed to sample sources of this potential infection. This bag includes potential source IDs and their corresponding p_g (j->i).
            vector<unsigned> potential_source_IDs, source_IDs;
            vector<float> potential_source_weights;

            potential_source_IDs.clear();
            potential_source_weights.clear();
            source_IDs.clear();

            // Build the bag of potential sources - New infections in this cycle may alter the list but their p_g (j->i) are zero so their effects can be neglegible. Very low chance that a source is a new case. Tested with sampling and observed that if weight=0, that item will never be selected.
            // This would fasten the computation, otherwise lists of potential sources and weights need to be built even for the case that no infection happens.
            if (in_day_time_contexts) {
                c_agent->getDaytimePotentialSourcesOfInfection(potential_source_IDs, potential_source_weights);
            } else {
                c_agent->getNighttimePotentialSourcesOfInfection(potential_source_IDs, potential_source_weights);
            }
            // Source(s) will be sampled and stored in source_IDs. Since the raw P > 0, the source list must not be empty. However, if it is, an error will be raised and simulation will be stopped.

            sampleSourcesForAnInfection(potential_source_IDs, potential_source_weights, source_IDs);

            // P_I will be set to depend on the virus transmitted from the source and the agent's immunisation history.
            float P_I = 1.0 - c_agent->get_VEs_from_immunisation_history(day);
            if (rand_r(&seed_thread[omp_get_thread_num()]) / (float)RAND_MAX < P_I)
            {
                if (source_IDs.size() == 1)
                {
                    c_agent->initiate_infection_info(agents.at(source_IDs[0])->infection_info);
                }
                else if (source_IDs.size() == 2)
                {
                    // TO_DO: Need to change to address co-infection. Currently, the function initiate_infection_info only copies the parameters belonging to the first source.
                    c_agent->initiate_infection_info(agents.at(source_IDs[0])->infection_info, agents.at(source_IDs[1])->infection_info);
                }
                else
                {
                    std::cout << "ERROR: More than 2 sources are sampled for the agent " << c_agent->getID() << ". They are: " << std::endl;
                    for (int i = 0; i < source_IDs.size(); i++)
                    {
                        std::cout << " -  Infectious agent " << source_IDs[i] << std::endl;
                    }
                    exit(1);
                }

                // VEc and VEt of expected immunity from natural recovery are passed to the agent,
                // so that when this agent is recovered from this infection, the immunisation from this recovery
                // is saved to the agent's immunisation of history.
                c_agent->infect(day, VEc_NR, VEt_NR, daily_waning_VEc_NR, daily_waning_VEt_NR);
                #pragma omp critical(updateAgentStateAllContexts_1)
                {
                    g_infected_incidence++; // counting the total infected incidence
                    g_ag_infection_incidence_map.at(c_agent->getAgeGroup())++;
                }
            }
        }
    } else if (c_agent->isInfected()) {
        // Bookkeeping
        #pragma omp critical(updateAgentStateAllContexts_2)
        {
            if (c_agent->isNewIllness())
            {
                g_illness_incidence++;
                g_ag_illness_incidence_map.at(c_agent->getAgeGroup())++;
                c_agent->resetNewIllness();
            }

            if (c_agent->isNewlyDetected()) // includes all detected agents, regardless of symptom status
            {
                auto my_sla = c_agent->getSLA()->getID();

                g_detected_incidence++;
                g_ag_detection_incidence_map.at(c_agent->getAgeGroup())++;
                g_sla_incidence_map.at(my_sla)++;
                g_sla_attack_rate_map.at(my_sla)++;
                g_ag_sla_incidence_map.at(c_agent->getAgeGroup()).at(my_sla)++;
                c_agent->resetNewDetection();
            }
        }
    }
}


void EpidemicSimulator::update_HQ()
{
    #pragma omp parallel for
    for (unsigned long i = 0; i < agent_ids.size(); i++) {
        agents[agent_ids[i]]->step_HQ(duration_HQ);
    }
}


void EpidemicSimulator::progressive_vaccination_priority(
    std::stack<unsigned long> &unvaccinated_senior_IDs_s,
    std::stack<unsigned long> &unvaccinated_adult_IDs_s,
    std::stack<unsigned long> &unvaccinated_child_IDs_s,
    unsigned day)
{
    // age-stratified progressive-vaccination
    int num_vac_today = 0;
    int num_vac_65_plus_today = 0;
    int num_vac_18_to_64_today = 0;
    int num_vac_children_today = 0;

    unsigned long n_vac_daily = (unsigned long)round(n_vac_per_week_PV / 7);
    unsigned long n_unvaccinated = unvaccinated_senior_IDs_s.size() + unvaccinated_adult_IDs_s.size();

    while (num_vac_today < n_vac_daily && n_unvaccinated > 0)
    {
        if (unvaccinated_senior_IDs_s.size() > 0)
        {
            for (int i = 1; i < p_vac_65_and_over + 1; i++)
            {
                if (unvaccinated_senior_IDs_s.size() > 0 && num_vac_today < n_vac_daily)
                {
                    auto c_agent = agents.at(unvaccinated_senior_IDs_s.top());
                    if (c_agent->wasIllnessDetected())
                    {
                        unvaccinated_senior_IDs_s.pop();
                    }
                    else
                    {
                        // Add an immunisation to this agent's immunisation history after this agent is vaccinated
                        c_agent->add_an_immunisation_to_history(day, VEc_PV, VEt_PV, daily_waning_VEc_PV, daily_waning_VEt_PV, Immunisation_Types::vaccinated_PV);
                        unvaccinated_senior_IDs_s.pop();
                        num_vac++;
                        num_vac_PV++;
                        num_vac_today++;
                        num_vac_65_plus++;
                        num_vac_65_plus_PV++;
                        num_vac_65_plus_today++;
                    }
                }
            }
        }

        if (unvaccinated_adult_IDs_s.size() > 0)
        {
            for (int i = 1; i < p_vac_18_to_64 + 1; i++)
            {
                if (unvaccinated_adult_IDs_s.size() > 0 && num_vac_today < n_vac_daily)
                {
                    auto c_agent = agents.at(unvaccinated_adult_IDs_s.top());
                    if (c_agent->wasIllnessDetected())
                    {
                        unvaccinated_adult_IDs_s.pop();
                    }
                    else
                    {
                        // Add an immunisation to this agent's immunisation history after this agent is vaccinated
                        c_agent->add_an_immunisation_to_history(day, VEc_PV, VEt_PV, daily_waning_VEc_PV, daily_waning_VEt_PV, Immunisation_Types::vaccinated_PV);
                        unvaccinated_adult_IDs_s.pop();
                        num_vac++;
                        num_vac_PV++;
                        num_vac_today++;
                        num_vac_18_to_64++;
                        num_vac_18_to_64_PV++;
                        num_vac_18_to_64_today++;
                    }
                }
            }
        }

        if (unvaccinated_child_IDs_s.size() > 0)
        {
            for (int i = 0; i < p_vac_children; i++)
            {
                if (unvaccinated_child_IDs_s.size() > 0 && num_vac_today < n_vac_daily)
                {
                    auto c_agent = agents.at(unvaccinated_child_IDs_s.top());
                    if (c_agent->wasIllnessDetected())
                    {
                        unvaccinated_child_IDs_s.pop();
                    }
                    else
                    {
                        // Add an immunisation to this agent's immunisation history after this agent is vaccinated
                        c_agent->add_an_immunisation_to_history(day, VEc_PV, VEt_PV, daily_waning_VEc_PV, daily_waning_VEt_PV, Immunisation_Types::vaccinated_PV);
                        unvaccinated_child_IDs_s.pop();
                        num_vac++;
                        num_vac_PV++;
                        num_vac_today++;
                        num_vac_children++;
                        num_vac_children_PV++;
                        num_vac_children_today++;
                    }
                }
            }
        }

        n_unvaccinated = unvaccinated_senior_IDs_s.size() + unvaccinated_adult_IDs_s.size();
    }

    if (num_vac_today > 0)
    {
        std::cout << "\n\tVaccination at the end of the day with priority vaccine: " << std::endl;
        std::cout << "\t - Number of agents vaccinated today, priority (total): "
                  << num_vac_today << " (" << num_vac_PV << ")" << std::endl;
        std::cout << "\t - Number of children vaccinated today, priority (total): "
                  << num_vac_children_today << " (" << num_vac_children_PV << ")" << std::endl;
        std::cout << "\t - Number of agents aged 18-64 vaccinated today, priority (total): "
                  << num_vac_18_to_64_today << " (" << num_vac_18_to_64_PV << ")" << std::endl;
        std::cout << "\t - Number of agents aged 65+ vaccinated today, priority (total): "
                  << num_vac_65_plus_today << " (" << num_vac_65_plus_PV << ")" << std::endl;
    }
}

// ------------------------------

void EpidemicSimulator::progressive_vaccination_general(
    std::stack<unsigned long> &unvaccinated_senior_IDs_s,
    std::stack<unsigned long> &unvaccinated_adult_IDs_s,
    std::stack<unsigned long> &unvaccinated_child_IDs_s,
    unsigned day)
{
    // age-stratified progressive-vaccination
    int num_vac_today = 0;
    int num_vac_65_plus_today = 0;
    int num_vac_18_to_64_today = 0;
    int num_vac_children_today = 0;

    unsigned long n_vac_daily = (unsigned long)round(n_vac_per_week_GV / 7);
    unsigned long n_unvaccinated = unvaccinated_senior_IDs_s.size() + unvaccinated_adult_IDs_s.size();

    while (num_vac_today < n_vac_daily && n_unvaccinated > 0)
    {
        if (unvaccinated_senior_IDs_s.size() > 0)
        {
            for (int i = 1; i < p_vac_65_and_over + 1; i++)
            {
                if (unvaccinated_senior_IDs_s.size() > 0 && num_vac_today < n_vac_daily)
                {
                    auto c_agent = agents.at(unvaccinated_senior_IDs_s.top());
                    if (c_agent->wasIllnessDetected())
                    {
                        unvaccinated_senior_IDs_s.pop();
                    }
                    else
                    {
                        // Add an immunisation to this agent's immunisation history after this agent is vaccinated
                        c_agent->add_an_immunisation_to_history(day, VEc_GV, VEt_GV, daily_waning_VEc_GV, daily_waning_VEt_GV, Immunisation_Types::vaccinated_GV);
                        unvaccinated_senior_IDs_s.pop();
                        num_vac++;
                        num_vac_GV++;
                        num_vac_today++;
                        num_vac_65_plus++;
                        num_vac_65_plus_GV++;
                        num_vac_65_plus_today++;
                    }
                }
            }
        }

        if (unvaccinated_adult_IDs_s.size() > 0)
        {
            for (int i = 1; i < p_vac_18_to_64 + 1; i++)
            {
                if (unvaccinated_adult_IDs_s.size() > 0 && num_vac_today < n_vac_daily)
                {
                    auto c_agent = agents.at(unvaccinated_adult_IDs_s.top());
                    if (c_agent->wasIllnessDetected())
                    {
                        unvaccinated_adult_IDs_s.pop();
                    }
                    else
                    {
                        // Add an immunisation to this agent's immunisation history after this agent is vaccinated
                        c_agent->add_an_immunisation_to_history(day, VEc_GV, VEt_GV, daily_waning_VEc_GV, daily_waning_VEt_GV, Immunisation_Types::vaccinated_GV);
                        unvaccinated_adult_IDs_s.pop();
                        num_vac++;
                        num_vac_GV++;
                        num_vac_today++;
                        num_vac_18_to_64++;
                        num_vac_18_to_64_GV++;
                        num_vac_18_to_64_today++;
                    }
                }
            }
        }

        if (unvaccinated_child_IDs_s.size() > 0)
        {
            for (int i = 0; i < p_vac_children; i++)
            {
                if (unvaccinated_child_IDs_s.size() > 0 && num_vac_today < n_vac_daily)
                {
                    auto c_agent = agents.at(unvaccinated_child_IDs_s.top());
                    if (c_agent->wasIllnessDetected())
                    {
                        unvaccinated_child_IDs_s.pop();
                    }
                    else
                    {
                        // Add an immunisation to this agent's immunisation history after this agent is vaccinated
                        c_agent->add_an_immunisation_to_history(day, VEc_GV, VEt_GV, daily_waning_VEc_GV, daily_waning_VEt_GV, Immunisation_Types::vaccinated_GV);
                        unvaccinated_child_IDs_s.pop();
                        num_vac++;
                        num_vac_GV++;
                        num_vac_today++;
                        num_vac_children++;
                        num_vac_children_GV++;
                        num_vac_children_today++;
                    }
                }
            }
        }

        n_unvaccinated = unvaccinated_senior_IDs_s.size() + unvaccinated_adult_IDs_s.size();
    }

    if (num_vac_today > 0)
    {
        std::cout << "\n\tVaccination at the end of the day with general vaccine: " << std::endl;
        std::cout << "\t - Number of agents vaccinated today, general (total): "
                  << num_vac_today << " (" << num_vac_GV << ")" << std::endl;
        std::cout << "\t - Number of children vaccinated today, general (total): "
                  << num_vac_children_today << " (" << num_vac_children_GV << ")" << std::endl;
        std::cout << "\t - Number of agents aged 18-64 vaccinated today, general (total): "
                  << num_vac_18_to_64_today << " (" << num_vac_18_to_64_GV << ")" << std::endl;
        std::cout << "\t - Number of agents aged 65+ vaccinated today, general (total): "
                  << num_vac_65_plus_today << " (" << num_vac_65_plus_GV << ")" << std::endl;
    }
}

// ------------------------------
// ------------------------------
// ------------------------------

void EpidemicSimulator::initiate_bookkeeping_everyday(unsigned run_id)
{
    gettimeofday(&tv, NULL);
    struct tm *tm_info = localtime(&tv.tv_sec);
    strftime(buffer, 26, "%Y-%m-%d_%H-%M-%S", tm_info);

    std::stringstream run_savefile,
        run_savefile_incidence,
        run_savefile_prevalence,
        run_savefile_population,
        run_savefile_breakdown,
        run_savefile_CI_HQ,
        run_savefile_infection_I_ag,
        run_savefile_illness_I_ag,
        run_savefile_detection_I_ag,
        run_savefile_infection_P_ag,
        run_savefile_illness_P_ag,
        run_savefile_detection_P_ag,
        run_savefile_P_vaccinations,
        run_savefile_G_vaccinations,
        run_savefile_health_states,
        run_savefile_immunisation_summary,
        run_savefile_recovered_susceptibles;

    run_savefile << base_dir << "/R" << run_id << "_D" << buffer;
    run_savefile_breakdown << run_savefile.str() << "/Iinf_Pinf_Iill_Pill_Idet_Pdet_CDI_CDP.dat";

    run_savefile_incidence << run_savefile.str() << "/I_SLA";
    run_savefile_prevalence << run_savefile.str() << "/P_SLA";
    run_savefile_CI_HQ << run_savefile.str() << "/CI_HQ.dat";
    run_savefile_infection_I_ag << run_savefile.str() << "/I_infection_Ag.dat";
    run_savefile_illness_I_ag << run_savefile.str() << "/I_illness_Ag.dat";
    run_savefile_detection_I_ag << run_savefile.str() << "/I_detection_Ag.dat";
    run_savefile_infection_P_ag << run_savefile.str() << "/P_infection_Ag.dat";
    run_savefile_illness_P_ag << run_savefile.str() << "/P_illness_Ag.dat";
    run_savefile_detection_P_ag << run_savefile.str() << "/P_detection_Ag.dat";
    run_savefile_P_vaccinations << run_savefile.str() << "/Vaccinations_priority_Ag.dat";
    run_savefile_G_vaccinations << run_savefile.str() << "/Vaccinations_general_Ag.dat";

    run_savefile_health_states << run_savefile.str() << "/N_health_states.dat";
    run_savefile_immunisation_summary << run_savefile.str() << "/N_immunity_states.dat";
    run_savefile_recovered_susceptibles << run_savefile.str() << "/N_recovered_susceptibles.dat";

    if (makeDir(run_savefile.str()) != 0)
    {
        std::cerr << "Making directory: " << run_savefile.str() << " failed. Exiting." << std::endl;
        exit(1);
    }
    std::cout << "Created folder " << run_savefile.str() << " for results." << std::endl;

    std::stringstream run_savefile_Detected_incidence_SD_compliance;
    run_savefile_Detected_incidence_SD_compliance << run_savefile.str() << "/I_SD_compliance.dat";
    outfile_stream_Detected_incidence_SD_compliance.open(run_savefile_Detected_incidence_SD_compliance.str());

    run_savefile << "/IP.dat";
    outfile_stream.open(run_savefile.str());
    outfile_stream_breakdown.open(run_savefile_breakdown.str());
    outfile_stream_incidence.open(run_savefile_incidence.str() + ".dat");
    outfile_stream_prevalence.open(run_savefile_prevalence.str() + ".dat");
    outfile_stream_CI_HQ.open(run_savefile_CI_HQ.str());
    outfile_stream_infection_I_ag.open(run_savefile_infection_I_ag.str());
    outfile_stream_illness_I_ag.open(run_savefile_illness_I_ag.str());
    outfile_stream_detection_I_ag.open(run_savefile_detection_I_ag.str());
    outfile_stream_infection_P_ag.open(run_savefile_infection_P_ag.str());
    outfile_stream_illness_P_ag.open(run_savefile_illness_P_ag.str());
    outfile_stream_detection_P_ag.open(run_savefile_detection_P_ag.str());
    outfile_stream_P_vaccination.open(run_savefile_P_vaccinations.str());
    outfile_stream_G_vaccination.open(run_savefile_G_vaccinations.str());

    outfile_stream_N_health_states.open(run_savefile_health_states.str());
    outfile_stream_N_immunity_states.open(run_savefile_immunisation_summary.str());
    outfile_stream_N_recovered_susceptibles.open(run_savefile_recovered_susceptibles.str());

    for (AgeGroup ag = AgeGroups::Age0to4; ag <= AgeGroups::Age65Plus; ag = AgeGroup(ag + 1))
    {
        std::string inc_fname(run_savefile_incidence.str() + "-" + std::to_string(ag) + ".dat");
        std::string prev_fname(run_savefile_prevalence.str() + "-" + std::to_string(ag) + ".dat");
        ag_outfile_stream_incidence.emplace(ag, std::make_shared<std::ofstream>(inc_fname, std::ios_base::app));
        ag_outfile_stream_prevalence.emplace(ag, std::make_shared<std::ofstream>(prev_fname, std::ios_base::app));
    }
}

// ------------------------------
// ------------------------------
void EpidemicSimulator::bookkeeping_everyday(unsigned run_id, unsigned day)
{

    int num_CD_incidence = count_CD_incidence();   // counts number of CD regions newly infected (0 -> at least one new case)
    int num_CD_prevalence = count_CD_prevalence(); // counts number of CD regions with at least one acive case (detected infection)

    float child_symptomatic_fraction =
        float(g_ag_illness_prevalence_map.at(AgeGroups::Age0to4) +
              g_ag_illness_prevalence_map.at(AgeGroups::Age5to18)) /
        float(g_ag_infection_prevalence_map.at(AgeGroups::Age0to4) +
              g_ag_infection_prevalence_map.at(AgeGroups::Age5to18));

    float adult_symptomatic_fraction =
        float(g_ag_illness_prevalence_map.at(AgeGroups::Age19to29) +
              g_ag_illness_prevalence_map.at(AgeGroups::Age30to64) +
              g_ag_illness_prevalence_map.at(AgeGroups::Age65Plus)) /
        float(g_ag_infection_prevalence_map.at(AgeGroups::Age19to29) +
              g_ag_infection_prevalence_map.at(AgeGroups::Age30to64) +
              g_ag_infection_prevalence_map.at(AgeGroups::Age65Plus));

    std::cout << "\n\t----- End-of-day analysis -----" << std::endl;
    std::cout << "\tDetection Incidence (Prop.): " << g_detected_incidence << " (" << float(g_detected_incidence) / float(agents.size()) << ")\n\t"
              << "Infection Incidence (Prop.): " << g_infected_incidence << " (" << float(g_infected_incidence) / float(agents.size()) << ")\n\t"
              << "Illness Incidence (Prop.): " << g_illness_incidence << " (" << float(g_illness_incidence) / float(agents.size()) << ")\n\t"
              << "\n\t"
              << "Detected Case Prevalence (Prop.): " << g_detected_prevalence << " (" << float(g_detected_prevalence) / float(agents.size()) << ")\n\t"
              << "Infection Prevalence (Prop.): " << g_infected_prevalence << " (" << float(g_infected_prevalence) / float(agents.size()) << ")\n\t"
              << "Illness Prevalence (Prop.): " << g_illness_prevalence << " (" << float(g_illness_prevalence) / float(agents.size()) << ")\n\t"
              << "\n\t"
              << "Cumulative Detected (Prop.): " << g_detected_cumulative << " (" << float(g_detected_cumulative) / float(agents.size()) << ")\n\t"
              << "Cumulative Infected (Prop.): " << g_infected_cumulative << " (" << float(g_infected_cumulative) / float(agents.size()) << ")\n\t"
              << "Cumulative Ill (Prop.): " << g_illness_cumulative << " (" << float(g_illness_cumulative) / float(agents.size()) << ")\n\t"
              << "Number removed (Prop.): " << g_number_removed << " (" << float(g_number_removed) / float(agents.size()) << ")\n\t"
              << "\n\t"
              << "Symptomatic Fraction in Children: " << child_symptomatic_fraction << "\n\t"
              << "Symptomatic Fraction in Adults: " << adult_symptomatic_fraction << "\n\t"
              << "\n\t"
              << "Number of airport infections: " << n_airport_infections << std::endl;

    // Print incidence and prevalence for each SLA
    outfile_stream << g_detected_incidence << "," << g_detected_prevalence << std::endl;

    outfile_stream_breakdown << g_infected_incidence
                             << "," << g_infected_prevalence
                             << "," << g_illness_incidence
                             << "," << g_illness_prevalence
                             << "," << g_detected_incidence
                             << "," << g_detected_prevalence
                             << "," << num_CD_incidence
                             << "," << num_CD_prevalence
                             << std::endl;

    outfile_stream_P_vaccination << num_vac_PV
                                 << "," << num_vac_children_PV
                                 << "," << num_vac_18_to_64_PV
                                 << "," << num_vac_65_plus_PV
                                 << std::endl;

    outfile_stream_G_vaccination << num_vac_GV
                                 << "," << num_vac_children_GV
                                 << "," << num_vac_18_to_64_GV
                                 << "," << num_vac_65_plus_GV
                                 << std::endl;

    for (auto sla_entry : statistical_local_areas)
    {
        outfile_stream_incidence << float(g_sla_incidence_map.at(sla_entry.first)) / float(sla_entry.second->getAgentIdxs().size()) << " ";
        outfile_stream_prevalence << float(g_sla_prevalence_map.at(sla_entry.first)) / float(sla_entry.second->getAgentIdxs().size()) << " ";

        for (AgeGroup ag = AgeGroups::Age0to4; ag <= AgeGroups::Age65Plus; ag = AgeGroup(ag + 1))
        {
            auto inc_outfile = ag_outfile_stream_incidence.at(ag);
            auto prev_outfile = ag_outfile_stream_prevalence.at(ag);

            *inc_outfile << float(g_ag_sla_incidence_map.at(ag).at(sla_entry.first)) / float(sla_entry.second->getAgentIdxs().size()) << " ";
            *prev_outfile << float(g_ag_sla_prevalence_map.at(ag).at(sla_entry.first)) / float(sla_entry.second->getAgentIdxs().size()) << " ";
        }
    }
    outfile_stream_incidence << std::endl;
    outfile_stream_prevalence << std::endl;

    for (AgeGroup ag = AgeGroups::Age0to4; ag <= AgeGroups::Age65Plus; ag = AgeGroup(ag + 1))
    {
        *ag_outfile_stream_incidence.at(ag) << std::endl;
        *ag_outfile_stream_prevalence.at(ag) << std::endl;

        outfile_stream_infection_I_ag << g_ag_infection_incidence_map.at(ag) << " ";
        outfile_stream_infection_P_ag << g_ag_infection_prevalence_map.at(ag) << " ";

        outfile_stream_illness_I_ag << g_ag_illness_incidence_map.at(ag) << " ";
        outfile_stream_illness_P_ag << g_ag_illness_prevalence_map.at(ag) << " ";

        outfile_stream_detection_I_ag << g_ag_detection_incidence_map.at(ag) << " ";
        outfile_stream_detection_P_ag << g_ag_detection_prevalence_map.at(ag) << " ";
    }

    outfile_stream_infection_I_ag << std::endl;
    outfile_stream_infection_P_ag << std::endl;
    outfile_stream_illness_I_ag << std::endl;
    outfile_stream_illness_P_ag << std::endl;
    outfile_stream_detection_I_ag << std::endl;
    outfile_stream_detection_P_ag << std::endl;

    // save Detected_incidence and SD_compliance
    outfile_stream_Detected_incidence_SD_compliance << g_detected_incidence << " " << SD_compliance_level << std::endl;

    // Keep track of number of agents in each health states and their immunisation
    std::vector<unsigned> health_states_summary_thread, health_states_summary;
    std::vector<unsigned> immunisation_summary_thread, immunisation_summary;
    unsigned n_recovered_susceptibles_thread, n_recovered_susceptibles;
    unsigned n_recovered_susceptibles_recovered_once_thread, n_recovered_susceptibles_recovered_once, n_recovered_susceptibles_recovered_twice_or_more_thread, n_recovered_susceptibles_recovered_twice_or_more;

    // Not consider the last state: DiseaseState::StateWithdrawn
    for (unsigned state_i=DiseaseState::StateSusceptible; state_i < DiseaseState::StateWithdrawn; state_i++) {
        health_states_summary_thread.emplace_back(0);
        health_states_summary.emplace_back(0);
    }

    for (unsigned state_i=Immunisation_Summary_Type::no_immunisation; state_i <= Immunisation_Summary_Type::vaccinated_twice_or_more_and_recovered_twice_or_more; state_i++) {
        immunisation_summary_thread.emplace_back(0);
        immunisation_summary.emplace_back(0);
    }

    n_recovered_susceptibles_thread = 0;
    n_recovered_susceptibles = 0;

    n_recovered_susceptibles_recovered_once_thread = 0;
    n_recovered_susceptibles_recovered_once = 0;

    n_recovered_susceptibles_recovered_twice_or_more_thread = 0;
    n_recovered_susceptibles_recovered_twice_or_more = 0;

    #pragma omp parallel firstprivate(health_states_summary_thread, immunisation_summary_thread, n_recovered_susceptibles_thread, n_recovered_susceptibles_recovered_once_thread, n_recovered_susceptibles_recovered_twice_or_more_thread)
    {   
        #pragma omp for
        for (unsigned agent_idx=0; agent_idx < agent_ids.size(); agent_idx++) {
            health_states_summary_thread[agents[agent_ids[agent_idx]]->getState()]++;
            immunisation_summary_thread[agents[agent_ids[agent_idx]]->get_immunisation_summary_type()]++;
            if ((agents[agent_ids[agent_idx]]->getState() == DiseaseState::StateSusceptible) && (agents[agent_ids[agent_idx]]->get_immunisation_summary_type() >= Immunisation_Summary_Type::recovered_once)) {
                n_recovered_susceptibles_thread++;

                if ((agents[agent_ids[agent_idx]]->get_immunisation_summary_type() == Immunisation_Summary_Type::recovered_once) || (agents[agent_ids[agent_idx]]->get_immunisation_summary_type() == Immunisation_Summary_Type::vaccinated_once_and_recovered_once) || (agents[agent_ids[agent_idx]]->get_immunisation_summary_type() == Immunisation_Summary_Type::vaccinated_twice_or_more_and_recovered_once)) {
                    n_recovered_susceptibles_recovered_once_thread++;
                } else if ((agents[agent_ids[agent_idx]]->get_immunisation_summary_type() == Immunisation_Summary_Type::recovered_twice_or_more) || (agents[agent_ids[agent_idx]]->get_immunisation_summary_type() == Immunisation_Summary_Type::vaccinated_once_and_recovered_twice_or_more) || (agents[agent_ids[agent_idx]]->get_immunisation_summary_type() == Immunisation_Summary_Type::vaccinated_twice_or_more_and_recovered_twice_or_more)) {
                    n_recovered_susceptibles_recovered_twice_or_more_thread++;
                }
            }
        }

        #pragma omp critical(health_immunisation_summary_)
        {
            for (unsigned state_i=DiseaseState::StateSusceptible; state_i < DiseaseState::StateWithdrawn; state_i++) {
                health_states_summary[state_i] += health_states_summary_thread[state_i];
            }

            for (unsigned state_i=Immunisation_Summary_Type::no_immunisation; state_i <= Immunisation_Summary_Type::vaccinated_twice_or_more_and_recovered_twice_or_more; state_i++) {
                immunisation_summary[state_i] += immunisation_summary_thread[state_i];
            }

            n_recovered_susceptibles += n_recovered_susceptibles_thread;
            n_recovered_susceptibles_recovered_once += n_recovered_susceptibles_recovered_once_thread;
            n_recovered_susceptibles_recovered_twice_or_more += n_recovered_susceptibles_recovered_twice_or_more_thread;

        }
    }

    for (unsigned state_i=DiseaseState::StateSusceptible; state_i < DiseaseState::StateWithdrawn; state_i++) {
        outfile_stream_N_health_states << health_states_summary[state_i] << " ";
    }
    outfile_stream_N_health_states << std::endl;

    for (unsigned state_i=Immunisation_Summary_Type::no_immunisation; state_i <= Immunisation_Summary_Type::vaccinated_twice_or_more_and_recovered_twice_or_more; state_i++) {
        outfile_stream_N_immunity_states << immunisation_summary[state_i] << " ";
    }
    outfile_stream_N_immunity_states << std::endl;

    outfile_stream_N_recovered_susceptibles << n_recovered_susceptibles << " " << n_recovered_susceptibles_recovered_once << " " << n_recovered_susceptibles_recovered_twice_or_more << std::endl;

}

// ------------------------------
// ------------------------------

void EpidemicSimulator::close_bookkeeping_everyday()
{
    outfile_stream.close();
    outfile_stream_breakdown.close();
    outfile_stream_CI_HQ.close();
    outfile_stream_infection_I_ag.close();
    outfile_stream_infection_P_ag.close();
    outfile_stream_illness_I_ag.close();
    outfile_stream_illness_P_ag.close();
    outfile_stream_detection_I_ag.close();
    outfile_stream_detection_P_ag.close();
    outfile_stream_P_vaccination.close();
    outfile_stream_G_vaccination.close();

    for (AgeGroup ag = AgeGroups::Age0to4; ag <= AgeGroups::Age65Plus; ag = AgeGroup(ag + 1))
    {
        (ag_outfile_stream_incidence[ag])->close();
        (ag_outfile_stream_prevalence[ag])->close();
    }

    outfile_stream_N_health_states.close();
    outfile_stream_N_immunity_states.close();
    outfile_stream_N_recovered_susceptibles.close();

    outfile_stream_Detected_incidence_SD_compliance.close();
}

// ------------------------------

int EpidemicSimulator::count_CD_incidence()
{
    int num_CD_incidence_pp = 0; // counts number of CDs with newly infected agents
    int num_CD_incidence_thread = 0;
    #pragma omp parallel firstprivate(num_CD_incidence_thread)
    {
        #pragma omp for
        for (unsigned i = 0; i < cd_ids.size(); i++)
        {
            if (census_districts[cd_ids[i]]->is_CD_newly_infected())
            {
                num_CD_incidence_thread++;
                census_districts[cd_ids[i]]->is_not_newly_infected();
            }
        }
        #pragma omp critical(count_CD_incidence_)
        {
            num_CD_incidence_pp += num_CD_incidence_thread;
        }
    }
    return (num_CD_incidence_pp);
}

// ------------------------------
// ------------------------------

int EpidemicSimulator::count_CD_prevalence()
{

    int num_CD_prevalence_pp = 0;
    int num_CD_prevalence_thread = 0;
    #pragma omp parallel firstprivate(num_CD_prevalence_thread)
    {
        #pragma omp for
        for (unsigned i = 0; i < cd_ids.size(); i++)
        {
            if (census_districts[cd_ids[i]]->get_n_infecteds_tot() > 0)
            {
                num_CD_prevalence_thread++;
            }
        }

        #pragma omp critical(count_CD_prevalence_)
        {
            num_CD_prevalence_pp += num_CD_prevalence_thread;
        }
    }
    return (num_CD_prevalence_pp);
}

// ------------------------------

int main(int argc, char *argv[])
{
    EpidemicSimulator COVID19(argc, argv);
    COVID19.print_summary();
    COVID19.simulate_all_runs();
    return 0;
}
