/*
   *Copyright:  

   ACEMod (2018--2021), The University of Sydney's invention disclosure CDIP Ref. 2019-123.
   Copyright ©  Oliver Cliff, Cameron Zachreson, Nathan Harding, Kristopher Fair, Mikhail Prokopenko, Mahendarajah Piraveenan

   AMTraC-19 (2020-2021), The University of Sydney's invention disclosure CDIP 2020-018 (2020).
   Copyright ©  Sheryl Chang, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   AMTraC-19 (2022-2023), The University of Sydney.
   Copyright ©  Sheryl Chang, Quang Dang Nguyen, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   AMTraC-19 (2024), The University of Sydney.
   Copyright ©  Quang Dang Nguyen, Sheryl Chang, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   *EndCopyright:
*/

#ifndef __STATISTICAL_AREAS_H__
#define __STATISTICAL_AREAS_H__

#include "enums.h"
#include "debug.h"

#include <iostream>
#include <vector>
#include <map>
#include <memory>

#include <iomanip>
#include <random>


/*
 * Base class - all statistical areas (SAs) have a number of agents, an ID (of type T), and a transmission probability
 */
template <class T> class StatisticalAreaBase
{
protected:
   const float P_ZERO = 1.0;

   /* ID associated with this SA */
   const T ID_;

   /* Agents in this SA */
   std::vector <unsigned long> agent_idxs_;

   /* Joint probability of transmission in this SA (dynamic) */
   std::vector <float> joint_transmission_probability_;

   bool Symptomatic_infection_;   // An indicator variable for detection in the area

   unsigned long n_infecteds_tot; // A counter for the total infected prevalence (symptomatic and asymptomatic)

   unsigned long n_infecteds_;

   std::vector<unsigned int> infected_agent_IDs;
   std::vector<float> infected_agent_weights;

public:

   /* Construct/default destructor */
   StatisticalAreaBase(T ID) : ID_(ID), n_infecteds_(0)
   {
   }

   virtual ~StatisticalAreaBase()
   {
   };

   /* Get ID of this SA */
   inline T getID(void)
   {
      return(ID_);
   }

   /* Get vector of agents' (IDs) in this SA */
   inline std::vector <unsigned long>& getAgentIdxs(void)
   {
      return(agent_idxs_);
   }

   /* Get probability of infection in this SA */
   inline float getProbabilityOfTransmission(unsigned short idx)
   {
      return(joint_transmission_probability_.at(idx));
   }

   /* Add an agent (ID) to this SA */
   inline void addAgentIdx(unsigned long idx)
   {
      return(agent_idxs_.push_back(idx));
   }

   /* Re-zero the transmission probabilities */

   inline void reZero(void)
   {
      for (auto& prob : joint_transmission_probability_)
       {
          prob = P_ZERO;
       }

      Symptomatic_infection_ = false;

      n_infecteds_ = 0;

   }


   void clear_infection_lists(void) {
      infected_agent_IDs.clear();
      infected_agent_weights.clear();
   }

   // FOR DEBUGGING PURPOSE
   inline void output_list_of_infected_agents() {
      std::cout << "**** In the saved list:" << std::endl;
      for (int i=0; i < infected_agent_IDs.size(); i++) {
         std::cout << " -  Agent " << infected_agent_IDs[i] << " with weight = " << infected_agent_weights[i] << std::endl;
      }
   }

   // QMOD: Add potential sources to the bag. Note that this function is for every statistical areas, except household_cluster. 
   // For household cluster, we need to know whether the destination and the source is an adult to know the exact weights.
   inline void getPotentialSourcesOfInfection(std::vector<unsigned>& potential_source_IDs, std::vector<float>& potential_source_weights) {
      // Copy (duplicate) entire vector.
      potential_source_IDs.reserve(potential_source_IDs.size() + infected_agent_IDs.size());
      potential_source_weights.reserve(potential_source_weights.size() + infected_agent_weights.size());
      potential_source_IDs.insert(potential_source_IDs.end(), infected_agent_IDs.begin(), infected_agent_IDs.end());
      potential_source_weights.insert(potential_source_weights.end(), infected_agent_weights.begin(), infected_agent_weights.end());
   }

   inline unsigned long getNumberOfInfects(void)
   {
      return(n_infecteds_);
   }
};

/*
 * Statistical local area data structure
 */
class StatisticalLocalArea : public StatisticalAreaBase <unsigned long>
{
   /* Independent probability of transmission in this SLA (constant) */
   static std::vector <float> independent_transmission_probability_;

   static constexpr float ORIG_CR_SIZE_ = 4300.0;
   static constexpr float ORIG_CR_AREA_ = 150518872.0;
   static constexpr float OUR_CR_AREA_  = 5409282700.0;
   float density_scaling_;

public:

   ~StatisticalLocalArea()
   {
   }

   StatisticalLocalArea(unsigned long ID) : StatisticalAreaBase <unsigned long>(ID)
   {
      joint_transmission_probability_ = std::vector <float>(1);
   }

   inline static void addTransmissionRate(float rate)
   {
      independent_transmission_probability_.push_back(rate);
   }

   inline void computeDensity(void)
   {
      density_scaling_ = ORIG_CR_SIZE_ / float(agent_idxs_.size());

#ifdef DEBUG_MODE_
      std::cout << "density scaling SLA:" << this->getID() << " (" << agent_idxs_.size() << " agents): " << density_scaling_ << std::endl;
#endif
   }

   // Update transmission probability as well as the list of infected agent pointers and their weights.
   inline void addInfectedAgent(float trans_scaler, AgeGroup ag, float factor, unsigned int infected_agent_ID)
   {
      // trans_scaler from this infected agent is now multiplied to P_inf.
      float P_inf = trans_scaler * independent_transmission_probability_.at(ag) * factor * density_scaling_;
      joint_transmission_probability_.at(0) *= 1.0 - P_inf;
      n_infecteds_++;
      
      // Appending agent ID and its weight to the lists
      infected_agent_IDs.push_back(infected_agent_ID);
      infected_agent_weights.push_back(P_inf);
   }

   inline float getProbabilityOfTransmission(void)
   {
      return(joint_transmission_probability_.at(0));
   }


};

/*
 * Census district data structure
 */
class CensusDistrict : public StatisticalAreaBase <unsigned long>
{
private:
   const std::shared_ptr <StatisticalLocalArea> sla_;
   const float latitude_;
   const float longitude_;

   static constexpr float ORIG_CR_SIZE_ = 2000.0;      // people
   static constexpr float ORIG_CR_AREA_ = 69992883.0;  // m^2
   static constexpr float OUR_CR_AREA_  = 198816201.0; // m^2
   float density_scaling_;
   bool infected_CD       = false;
   bool newly_infected_CD = false;

   static std::vector <float> independent_transmission_probability_;

public:

   ~CensusDistrict()
   {
   }

   CensusDistrict(const unsigned long ID,
      const std::shared_ptr <StatisticalLocalArea>& sla,
      const float latitude,
      const float longitude)
      : StatisticalAreaBase <unsigned long>(ID),
      sla_(sla), latitude_(latitude), longitude_(longitude)
   {
      joint_transmission_probability_ = std::vector <float>(1);
      Symptomatic_infection_          = false;
      n_infecteds_tot                 = 0;
   }

   inline const std::shared_ptr <StatisticalLocalArea>& getSLA(void)
   {
      return(sla_);
   }

   inline static void addTransmissionRate(float rate)
   {
      independent_transmission_probability_.push_back(rate);
   }

   inline void computeDensity(void)
   {
      density_scaling_ = ORIG_CR_SIZE_ / float(agent_idxs_.size());
   }

   /* overloaded method */
   inline void addAgentIdx(unsigned long idx)
   {
      sla_->addAgentIdx(idx);
      agent_idxs_.push_back(idx);
   }

   // Update transmission probability as well as the list of infected agent pointers and their weights.
   inline void addInfectedAgent(float trans_scaler, AgeGroup ag, float factor, unsigned int infected_agent_ID)
   {
      // trans_scaler from this infected agent is now multiplied to P_inf.
      float P_inf = trans_scaler * independent_transmission_probability_.at(ag) * factor * density_scaling_;
      joint_transmission_probability_.at(0) *= 1.0 - P_inf;
      n_infecteds_++;

      // Appending agent ID and its weight to the lists
      infected_agent_IDs.push_back(infected_agent_ID);
      infected_agent_weights.push_back(P_inf);
   }


   inline float getProbabilityOfTransmission(void)
   {
      return(joint_transmission_probability_.at(0));
   }

   // This is for the intervention. It adds 1 to the number of sypmtomatic infected for a given location
   inline void addSymptomaticAgent(void)
   {
      Symptomatic_infection_ = true;
   }

   inline bool getSymptomaticInfection(void)
   {
      return(Symptomatic_infection_);
   }

   inline void add_n_infecteds_tot(void)
   {
      n_infecteds_tot += 1.0;
   }

   inline void remove_n_infecteds_tot(void)
   {
      n_infecteds_tot -= 1.0;
   }

   inline void rezero_n_infecteds_tot(void)
   {
      n_infecteds_tot = 0;
   }

   inline unsigned long get_n_infecteds_tot(void)
   {
      return(n_infecteds_tot);
   }

   inline void infect_CD(void)
   {
      newly_infected_CD = true;
      infected_CD       = true;
   }

   inline void is_not_newly_infected(void)
   {
      newly_infected_CD = false;
   }

   inline void uninfect_CD(void)
   {
      infected_CD = false;
   }

   inline bool is_CD_newly_infected(void)
   {
      return(newly_infected_CD);
   }

};

/*
 * Destination zone data structure
 */
class DestinationZone : public StatisticalAreaBase <unsigned long long>
{
private:

   const std::shared_ptr <StatisticalLocalArea> sla_;

public:

   ~DestinationZone()
   {
   }

   DestinationZone(const unsigned long long ID,
      const std::shared_ptr <StatisticalLocalArea>& sla)
      : StatisticalAreaBase <unsigned long long>(ID), sla_(sla)
   {
   }

   inline const std::shared_ptr <StatisticalLocalArea>& getSLA(void)
   {
      return(sla_);
   }

};

/*
 * Working group data structure
 */
class WorkingGroup : public StatisticalAreaBase <unsigned long>
{
private:

   const std::shared_ptr <DestinationZone> dz_;

   float density_scaling_;
   const int type_;

   static const std::vector <float> ORIG_TR_SIZE_; // people

   static std::vector <float> independent_transmission_probability_;

public:

   ~WorkingGroup()
   {
   }

   WorkingGroup(const unsigned long ID,
      const int type,
      const std::shared_ptr <DestinationZone>& dz)
      : StatisticalAreaBase <unsigned long>(ID), dz_(dz), type_(type)
   {
      joint_transmission_probability_ = std::vector <float>(1);
      Symptomatic_infection_          = false;
   }

   inline void computeDensity(void)
   {
      density_scaling_ = ORIG_TR_SIZE_.at(type_) / float( agent_idxs_.size());
   }

   /* Getter methods */
   inline const std::shared_ptr <DestinationZone>& getDZ(void)
   {
      return(dz_);
   }

   inline int getType(void)
   {
      return(type_);
   }

   inline float getProbabilityOfTransmission(void)
   {
      /* Only one prob of transmission in a WG */
      return(joint_transmission_probability_.at(0));
   }

   /* Overloaded methods */
   inline void addAgentIdx(unsigned long idx)
   {
      dz_->addAgentIdx(idx);
      agent_idxs_.push_back(idx);
   }

   inline static void addTransmissionRate(float rate)
   {
      independent_transmission_probability_.push_back(rate);
   }

   // QMOD: Update transmission probability as well as the list of infected agent pointers and their weights.
   inline void addInfectedAgent(float trans_scaler, float factor, unsigned int infected_agent_ID)
   {  
      // QMOD: trans_scaler from this infected agent is now multiplied to P_inf.
      float P_inf = trans_scaler * independent_transmission_probability_.at(type_) * factor * density_scaling_;
      joint_transmission_probability_.at(0) *= 1.0 - P_inf;
      n_infecteds_++;

      // QMOD: Appending agent ID and its weight to the lists
      infected_agent_IDs.push_back(infected_agent_ID);
      infected_agent_weights.push_back(P_inf);
   }



   inline void addSymptomaticAgent(void)
   {
      Symptomatic_infection_ = true;
   }

   inline bool getSymptomaticInfection(void)
   {
      return(Symptomatic_infection_);
   }

};

/*
 * Household cluster data structure
 */
class HouseholdCluster : public StatisticalAreaBase <unsigned long>
{
private:

   const static unsigned short CHILD_IDX = 0, ADULT_IDX = 1;

   const std::shared_ptr <CensusDistrict> cd_;

   /* Special cases for household where we have different transmission rates/probs depending on who it's coming from */
   static std::vector <std::vector <float> > independent_transmission_probability_matrix_;
   std::vector <std::vector <float> > joint_transmission_probability_matrix_;
   
   std::vector<bool> infected_agent_is_an_adult;

public:

   ~HouseholdCluster()
   {
   }

   HouseholdCluster(const unsigned long ID,
      const std::shared_ptr <CensusDistrict>& cd)
      : StatisticalAreaBase <unsigned long>(ID), cd_(cd)
   {
      joint_transmission_probability_ = std::vector <float>(5);
      for (unsigned i = 0; i < 5; ++i)
       {
          joint_transmission_probability_matrix_.push_back(std::vector <float>(5));
       }
   }

   inline const std::shared_ptr <CensusDistrict>& getCD(void)
   {
      return(cd_);
   };

   // Update transmission probability as well as the list of infected agent pointers and their weights.
   inline void addInfectedAgent(float trans_scaler, bool is_adult, float factor, unsigned int infected_agent_ID)
   {
      float P_adult_to_child = independent_transmission_probability_matrix_.at(is_adult).at(CHILD_IDX);
      float P_adult_to_adult = independent_transmission_probability_matrix_.at(is_adult).at(ADULT_IDX);

      // trans_scaler from this infected agent is now multiplied to P_inf.
      joint_transmission_probability_matrix_.at(is_adult).at(CHILD_IDX) *= 1.0 - trans_scaler * P_adult_to_child * factor;
      joint_transmission_probability_matrix_.at(is_adult).at(ADULT_IDX) *= 1.0 - trans_scaler * P_adult_to_adult * factor;
      n_infecteds_++;

      // Appending agent ID and its weight to the lists
      infected_agent_IDs.push_back(infected_agent_ID);
      // This is a specical case for HC when only (trans_scaler * factor) is saved. The weight will be then processed based on the is-adult check of both agent j and agent i
      infected_agent_weights.push_back(trans_scaler * factor);
      // Save whether this agent is an adult
      infected_agent_is_an_adult.push_back(is_adult);

   }

   // Add potential sources to the bag. Note that this function is for every statistical areas, except household_cluster. 
   // For household cluster, we need to know whether the dest and the source is adult to know exact weights.
   inline void getPotentialSourcesOfInfection(bool is_agent_i_an_adult, std::vector<unsigned>& potential_source_IDs, std::vector<float>& potential_source_weights) {
      // Copy (duplicate) entire vector.
      potential_source_IDs.reserve(potential_source_IDs.size() + infected_agent_IDs.size());
      potential_source_IDs.insert(potential_source_IDs.end(), infected_agent_IDs.begin(), infected_agent_IDs.end());

      for (int i=0; i < infected_agent_IDs.size(); i++) {
         potential_source_weights.push_back(
            (float) independent_transmission_probability_matrix_.at(infected_agent_is_an_adult[i]).at(is_agent_i_an_adult) * infected_agent_weights[i]
         );
      }
   }
   
   inline float getProbabilityOfTransmission(bool is_adult)
   {
      return(joint_transmission_probability_matrix_.at(CHILD_IDX).at(is_adult)
             * joint_transmission_probability_matrix_.at(ADULT_IDX).at(is_adult));
   }

   inline static void addTransmissionRate(bool is_adult, float rate)
   {
      if (independent_transmission_probability_matrix_.size() <= is_adult)
       {
          independent_transmission_probability_matrix_.push_back(std::vector <float>());
       }
      independent_transmission_probability_matrix_.at(is_adult).push_back(rate);
   }

   inline void addAgentIdx(unsigned long idx)
   {
      cd_->addAgentIdx(idx);
      agent_idxs_.push_back(idx);
   }

   inline void reZero(void)
   {
      for (auto& tp : joint_transmission_probability_)
       {
          tp = P_ZERO;
       }


      for (auto& tp_row : joint_transmission_probability_matrix_)
       {
          for (auto& tp : tp_row)
           {
              tp = P_ZERO;
           }
       }
      n_infecteds_ = 0;
      
   }

   void clear_infection_lists(void) {
      infected_agent_IDs.clear();
      infected_agent_weights.clear();
      infected_agent_is_an_adult.clear();
   }

};

/*
 * Household data structure
 */
class Household : public StatisticalAreaBase <unsigned long>
{
private:
   const std::shared_ptr <HouseholdCluster> hc_;

   static std::vector <std::vector <float> > independent_transmission_probability_matrix_;

   // a flag indicating that a household has a case-isolated illness. this will be used as a condition for home quarantine.
   bool CI_in_household = false;


public:

   ~Household()
   {
   }

   Household(const unsigned long ID,
      const std::shared_ptr <HouseholdCluster>& hc)
      : StatisticalAreaBase <unsigned long>(ID), hc_(hc)
   {
      joint_transmission_probability_ = std::vector <float>(1),
      Symptomatic_infection_          = false;
   }

   inline const std::shared_ptr <HouseholdCluster>& getHC(void)
   {
      return(hc_);
   }

   inline unsigned getHouseholdSize(void)
   {
      return(this->agent_idxs_.size());
   }

   inline HouseholdSizeGroup getHouseholdSizeGroup(void)
   {
      HouseholdSizeGroup my_size_group;

      switch (this->getHouseholdSize())
       {
       case 1:
       case 2:
          my_size_group = HouseholdSizeGroups::TwoOrLessPeople;
          break;

       case 3:
          my_size_group = HouseholdSizeGroups::ThreePeople;
          break;

       case 4:
          my_size_group = HouseholdSizeGroups::FourPeople;
          break;

       case 5:
          my_size_group = HouseholdSizeGroups::FivePeople;
          break;

       default:
          my_size_group = HouseholdSizeGroups::SixOrMorePeople;
       }
      return(my_size_group);
   }

   /* Get probability of infection in this household */
   inline float getProbabilityOfTransmission(void)
   {
      return(joint_transmission_probability_.at(0));
   }

   inline static void addTransmissionRate(unsigned short hh_size, float rate)
   {
      if (independent_transmission_probability_matrix_.size() <= hh_size)
       {
          independent_transmission_probability_matrix_.push_back(std::vector <float>());
       }
      independent_transmission_probability_matrix_.at(hh_size).push_back(rate);
   }

   // Update transmission probability as well as the list of infected agent pointers and their weights.
   inline void addInfectedAgent(float trans_scaler, AgeGroup ag, float factor, unsigned int infected_agent_ID)
   {
      // trans_scaler from this infected agent is now multiplied to P_inf.
      float P_inf = trans_scaler * independent_transmission_probability_matrix_.at(this->getHouseholdSizeGroup()).at(ag) * factor;
      joint_transmission_probability_.at(0) *= 1.0 - P_inf;
      n_infecteds_++;
      
      // Appending agent ID and its weight to the lists
      infected_agent_IDs.push_back(infected_agent_ID);
      infected_agent_weights.push_back(P_inf);
   }


   inline void addAgentIdx(unsigned long idx)
   {
      hc_->addAgentIdx(idx);
      agent_idxs_.push_back(idx);
   }

   inline void reZero(void)
   {
      for (auto& tp : joint_transmission_probability_)
       {
          tp = P_ZERO;
       }

      Symptomatic_infection_ = false;

      n_infecteds_ = 0;

      CI_in_household = false;

   }

   inline void addSymptomaticAgent(void)
   {
      Symptomatic_infection_ = true;
   }

   inline bool getSymptomaticInfection(void)
   {
      return(Symptomatic_infection_);
   }

   inline bool getCIinHousehold(void)
   {
      return CI_in_household;
   }

   inline void setCIinHousehold(bool CI_flag)
   {
      CI_in_household = CI_flag;
   }


};


typedef std::map <unsigned long, std::shared_ptr <StatisticalLocalArea> >   sla_map_type;
typedef std::map <unsigned long, std::shared_ptr <CensusDistrict> >         cd_map_type;
typedef std::map <unsigned long long, std::shared_ptr <DestinationZone> >   dz_map_type;
typedef std::map <signed long, std::shared_ptr <WorkingGroup> >             wg_map_type;
typedef std::map <signed long, std::shared_ptr <WorkingGroup> >             school_map_type;
typedef std::map <unsigned long, std::shared_ptr <HouseholdCluster> >       hc_map_type;
typedef std::map <unsigned long, std::shared_ptr <Household> >              hh_map_type;

#endif // __STATISTICAL_AREAS_H__
