/*
   *Copyright:  

   ACEMod (2018--2021), The University of Sydney's invention disclosure CDIP Ref. 2019-123.
   Copyright ©  Oliver Cliff, Cameron Zachreson, Nathan Harding, Kristopher Fair, Mikhail Prokopenko, Mahendarajah Piraveenan

   AMTraC-19 (2020-2021), The University of Sydney's invention disclosure CDIP 2020-018 (2020).
   Copyright ©  Sheryl Chang, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   AMTraC-19 (2022-2023), The University of Sydney.
   Copyright ©  Sheryl Chang, Quang Dang Nguyen, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   AMTraC-19 (2024), The University of Sydney.
   Copyright ©  Quang Dang Nguyen, Sheryl Chang, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   *EndCopyright:
*/

#include "virus.h"
#include <random>
#include <iostream>
#include <iomanip>

// A virus can be initialised with all information or can be assigned as a placeholder 
// so that when a real infection happens, information of the specific infected virus 
// will be transfered from the source to the destination.
// This constructor is for the former case.
Virus::Virus( 
    float trans_scaler_mean, float trans_scaler_std,
    float latent_period_mean, float latent_period_std,
    float incubation_period_mu, float incubation_period_sig, 
    float recovery_period_hi, float recovery_period_low, 
    float after_recovery_period_mean, float after_recovery_period_std,
    float adult_symptomatic_fraction, float child_symptomatic_fraction, 
    float asymptomatic_infectivity,
    unsigned origin_code) {

    set_trans_scaler_dist(trans_scaler_mean, trans_scaler_std);
    set_latent_period_dist(latent_period_mean, latent_period_std);
    set_incubation_period_dist(incubation_period_mu, incubation_period_sig);
    set_recovery_period_dist(recovery_period_hi, recovery_period_low);
    set_after_recovery_period_dist(after_recovery_period_mean, after_recovery_period_std);
    set_adult_symptomatic_fraction(adult_symptomatic_fraction);
    set_child_symptomatic_fraction(child_symptomatic_fraction);
    set_asymptomatic_infectivity(asymptomatic_infectivity);
    set_origin_code(origin_code);

}

// This constructor is for the latter case.
Virus::Virus(void) {    
}

// Each agent will call this function and sample a value representing the effect of the virus to itself.
float Virus::sample_trans_scaler(void) {
    std::normal_distribution<float> trans_scaler_dist(trans_scaler_mean, trans_scaler_std);
    std::random_device rd;
    std::default_random_engine generator(rd());
    return std::max(trans_scaler_dist(generator), 0.0f);
}

// Each agent will call this function and sample a value representing the effect of the virus to itself.
float Virus::sample_latent_period(void) {
    std::normal_distribution<float> latent_period_dist(latent_period_mean, latent_period_std);
    std::random_device rd;
    std::default_random_engine generator(rd());
    return std::max(latent_period_dist(generator), 0.0f);
}

// Each agent will call this function and sample a value representing the effect of the virus to itself.
float Virus::sample_incubation_period(void) {
    std::lognormal_distribution<float> incubation_period_dist(incubation_period_mu, incubation_period_sig);
    std::random_device rd;
    std::default_random_engine generator(rd());
    return std::max(incubation_period_dist(generator), 0.0f);
}

// Each agent will call this function and sample a value representing the effect of the virus to itself.
int Virus::sample_recovery_period(void) {
    std::uniform_int_distribution<int>  recovery_period_dist(recovery_period_low, recovery_period_hi);
    std::random_device rd;
    std::default_random_engine generator(rd());
    return std::max(recovery_period_dist(generator), 0);
}

// Each agent will call this function and sample a value representing the effect of the virus to itself.
float Virus::sample_after_recovery_period(void) {
    std::normal_distribution<float> after_recovery_period_dist(after_recovery_period_mean, after_recovery_period_std);
    std::random_device rd;
    std::default_random_engine generator(rd());
    return std::max(after_recovery_period_dist(generator), 0.0f);
}

// ---------------------------------------------------------

float Virus::get_trans_scaler_mean(void) {
    return this->trans_scaler_mean;
}

float Virus::get_trans_scaler_std(void) {
    return this->trans_scaler_std;
}

float Virus::get_latent_period_mean(void) {
    return this->latent_period_mean;
}

float Virus::get_latent_period_std(void) {
    return this->latent_period_std;
}

float Virus::get_adult_symptomatic_fraction(void) {
    return this->adult_symptomatic_fraction;
}

float Virus::get_child_symptomatic_fraction(void) {
    return this->child_symptomatic_fraction;
}

float Virus::get_asymptomatic_infectivity(void) {
    return this->asymptomatic_infectivity;
}

float Virus::get_incubation_period_dist_mu(void) {
    return this->incubation_period_mu;
}

float Virus::get_incubation_period_dist_sig(void) {
    return this->incubation_period_sig;
}

float Virus::get_recovery_period_dist_high(void) {
    return this->recovery_period_hi;
}

float Virus::get_recovery_period_dist_low(void) {
    return this->recovery_period_low;
}

float Virus::get_after_recovery_period_mean(void) {
    return this->after_recovery_period_mean;
}

float Virus::get_after_recovery_period_std(void) {
    return this->after_recovery_period_std;
}
unsigned Virus::get_origin_code(void) {
    return this->origin_code;
}

// ---------------------------------------------------------

void Virus::set_trans_scaler_dist(float trans_scaler_mean, float trans_scaler_std) {
    this->trans_scaler_mean = trans_scaler_mean;
    this->trans_scaler_std = trans_scaler_std;
}

void Virus::set_latent_period_dist(float latent_period_mean, float latent_period_std) {
    this->latent_period_mean = latent_period_mean;
    this->latent_period_std = latent_period_std;
}

void Virus::set_incubation_period_dist(float incubation_period_mu, float incubation_period_sig) {
    this->incubation_period_mu = incubation_period_mu;
    this->incubation_period_sig = incubation_period_sig;
}

void Virus::set_recovery_period_dist(float recovery_period_hi, float recovery_period_low) {
    this->recovery_period_hi = recovery_period_hi;
    this->recovery_period_low = recovery_period_low;
}

void Virus::set_after_recovery_period_dist(float after_recovery_period_mean, float after_recovery_period_std) {
    this->after_recovery_period_mean = after_recovery_period_mean;
    this->after_recovery_period_std = after_recovery_period_std;
}

void Virus::set_adult_symptomatic_fraction(float adult_symptomatic_fraction) {
    this->adult_symptomatic_fraction = adult_symptomatic_fraction;
}

void Virus::set_child_symptomatic_fraction(float child_symptomatic_fraction) {
    this->child_symptomatic_fraction = child_symptomatic_fraction;
}
void Virus::set_asymptomatic_infectivity(float asymptomatic_infectivity) {
    this->asymptomatic_infectivity = asymptomatic_infectivity;
}

void Virus::set_origin_code(unsigned origin_code) {
    this->origin_code = origin_code;
}

// ----------------------------------------

void Virus::output_viral_info() {
    std::cout << "Viral info: " << std::endl;
    std::cout << " - origin name code = " << origin_code << std::endl;
    std::cout << " - trans_scaler_mean = " << trans_scaler_mean << std::endl;
    std::cout << " - trans_scaler_std = " << trans_scaler_std << std::endl;
    std::cout << " - incubation_period_mu = " << incubation_period_mu << std::endl;
    std::cout << " - incubation_period_sig = " << incubation_period_sig << std::endl;
    std::cout << " - recovery_period_hi = " << recovery_period_hi << std::endl;
    std::cout << " - recovery_period_low = " << recovery_period_low << std::endl;
    std::cout << " - after_recovery_period_mean = " << after_recovery_period_mean << std::endl;
    std::cout << " - after_recovery_period_std = " << after_recovery_period_std << std::endl;
    std::cout << " - adult_symptomatic_fraction = " << adult_symptomatic_fraction << std::endl;
    std::cout << " - child_symptomatic_fraction = " << child_symptomatic_fraction << std::endl;
    std::cout << " - asymptomatic_infectivity = " << asymptomatic_infectivity << std::endl;
}
